/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Brutus e-book backend class. The Groupwise backend has been used
 *    as skeleton code for this class. The Brutus and Groupwise classes
 *    are therefore structurally very similar. Some code is even taken
 *    directly from the Groupwise class implementation.
 *
 *    Authors:
 *        Jules Colding <colding@42tools.com>
 *
 *    Copyright (C) 2007 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <errno.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <glib/gi18n-lib.h>

#include <libedataserver/e-sexp.h>

#if (EDS_EPOCH >= 224)
#include <libebackend/e-db3-utils.h>
#else
#include <libedataserver/e-db3-utils.h>
#endif

#include <libebook/e-contact.h>
#include <libedata-book/e-data-book.h>
#include <libedata-book/e-data-book-view.h>

#include <idl_output/BrutusLogOn.h>
#include <idl_output/IMailUser.h>
#include <idl_output/return_codes.h>
#include <idl_output/uid.h>

#include <session/brutusd.h>
#include <session/brutusd-util.h>
#include <server/brutus_corba.h>
#include <server/brutus_mapi.h>
#include <server/brutus_util.h>
#include <server/brutus_conv.h>
#include <server/brutus_jpeg.h>
#include <server/brutus_e-flag.h>
#include <server/brutus.h>
#include <servant_impl/BrutusCheck_impl.c>

#include "e-book-backend-brutus.h"

#define SHOW_ALL_SEARCH "(contains \"x-evolution-any-field\" \"\")"
#define G_OBJECT_BRUTUS_BOOK_VIEW "EBookBackendBrutus.BookView::closure"
#define SUMMARY_FLUSH_TIMEOUT 5000

/* auto-check timeout interval in minutes */
#ifdef AUTO_CHECK_TIMEOUT_DEFAULT
#undef AUTO_CHECK_TIMEOUT_DEFAULT
#endif
#define AUTO_CHECK_TIMEOUT_DEFAULT 60

static EBookBackendClass *e_book_backend_brutus_parent_class;

typedef struct {
	EBookBackendBrutus *brutus_backend;
	GThread *thread;
	EFlag *running;
} BrutusBackendSearchClosure;

#include <servant_impl/types_impl.c>

static void
brutus_kill__defined_but_not_used__compile_warnings(void)
{
	BRUTUS_FUZZY_TYPE fuzzy_level = BRUTUS_BRUTUS_FL_FULLSTRING;
	BRUTUS_BDEFINE prop_tag = 0;
	BRUTUS_SPropValue prop_value = { 0 };
	BRUTUS_RELOP_TYPE relop = BRUTUS_BRUTUS_RELOP_LT;
	BRUTUS_BMR_TYPE bmr = BRUTUS_BRUTUS_BMR_EQZ;

	// don't ever execute this function!
	g_assert(FALSE);

	brutus_kill__defined_but_not_used__compile_warnings();

	impl_BRUTUS_SRestrictionContainer__create(CORBA_OBJECT_NIL, CORBA_OBJECT_NIL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SAndRestriction__create(NULL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SOrRestriction__create(NULL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SNotRestriction__create(0, NULL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SContentRestriction__create(fuzzy_level, prop_tag, prop_value, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SPropertyRestriction__create(relop, prop_tag, prop_value, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SComparePropsRestriction__create(relop, prop_tag, prop_tag, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SBitMaskRestriction__create(bmr, prop_tag, prop_tag, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SSizeRestriction__create(relop, prop_tag, 0, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SExistRestriction__create(0, prop_tag, 0, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SSubRestriction__create(prop_tag, CORBA_OBJECT_NIL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SCommentRestriction__create(CORBA_OBJECT_NIL, NULL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_IUnknown__create(CORBA_OBJECT_NIL, NULL);
}

/*
 * GAL search mapping...
 */
static BRUTUS_BDEFINE GAL_content_tags[] =
{
	BRUTUS_BRUTUS_PR_ENTRYID, // *must* be the first entry
	BRUTUS_BRUTUS_PR_DISPLAY_NAME,
	BRUTUS_BRUTUS_PR_SURNAME,
	BRUTUS_BRUTUS_PR_GIVEN_NAME,
	BRUTUS_BRUTUS_PR_TITLE,
	BRUTUS_BRUTUS_PR_COMPANY_NAME,
	BRUTUS_BRUTUS_PR_DEPARTMENT_NAME,
	BRUTUS_BRUTUS_PR_OFFICE_LOCATION,
	BRUTUS_BRUTUS_PR_EMS_AB_DESCRIPTION,
	BRUTUS_BRUTUS_PR_COMMENT,
	BRUTUS_BRUTUS_PR_SMTP_ADDRESS,
	BRUTUS_BRUTUS_PR_BUSINESS_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_STREET_ADDRESS,
	BRUTUS_BRUTUS_PR_LOCALITY,
	BRUTUS_BRUTUS_PR_STATE_OR_PROVINCE
};
static BrutusSearchDef*
brutus_GAL_search_mapping()
{
	int i;
	BrutusSearchDef *retv = NULL;
	const size_t len = 13;

	if (0)
		brutus_kill__defined_but_not_used__compile_warnings(); // to avoid compiler warning

	retv = g_new0(BrutusSearchDef, 1);
	if (!retv)
		return NULL;

	retv->mapping = g_new0(BrutusEBookSearchDef, len);
	if (!retv->mapping) {
		g_free(retv);
		return NULL;
	}
	retv->len = len;

	/*
	 * 'file_as' is special as I use the PR_ANR pseudo
	 * property for this, and only this, search_token.
	 */
	i = 0;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("file_as");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_ANR;

	i++;
	retv->mapping[i].tag_count = 3;
	retv->mapping[i].search_token = g_strdup("full_name");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_SURNAME;
	retv->mapping[i].search_tags[1] = BRUTUS_BRUTUS_PR_GIVEN_NAME;
	retv->mapping[i].search_tags[2] = BRUTUS_BRUTUS_PR_TITLE;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("given_name");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_GIVEN_NAME;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("family_name");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_SURNAME;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("org");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_COMPANY_NAME;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("org_unit");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_DEPARTMENT_NAME;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("office");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_OFFICE_LOCATION;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("title");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_TITLE;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("role");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_EMS_AB_DESCRIPTION;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("note");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_COMMENT;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("email");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_SMTP_ADDRESS;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("phone");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_BUSINESS_TELEPHONE_NUMBER;

	i++;
	retv->mapping[i].tag_count = 4;
	retv->mapping[i].search_token = g_strdup("address");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_OFFICE_LOCATION;
	retv->mapping[i].search_tags[1] = BRUTUS_BRUTUS_PR_STREET_ADDRESS;
	retv->mapping[i].search_tags[2] = BRUTUS_BRUTUS_PR_LOCALITY;
	retv->mapping[i].search_tags[3] = BRUTUS_BRUTUS_PR_STATE_OR_PROVINCE;

	return retv;
}

/*
 * PAB search mapping...
 */
static BRUTUS_BDEFINE PAB_content_tags[] =
{
	BRUTUS_BRUTUS_PR_ENTRYID, // *must* be the first entry
	BRUTUS_BRUTUS_PR_ASSISTANT,
	BRUTUS_BRUTUS_PR_ASSISTANT_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_BUSINESS2_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_BUSINESS_HOME_PAGE,
	BRUTUS_BRUTUS_PR_BUSINESS_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_CALLBACK_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_CAR_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_COMMENT,
	BRUTUS_BRUTUS_PR_COMPANY_MAIN_PHONE_NUMBER,
	BRUTUS_BRUTUS_PR_COMPANY_NAME,
	BRUTUS_BRUTUS_PR_COUNTRY,
	BRUTUS_BRUTUS_PR_DEPARTMENT_NAME,
	BRUTUS_BRUTUS_PR_DISPLAY_NAME,
	BRUTUS_BRUTUS_PR_GIVEN_NAME,
	BRUTUS_BRUTUS_PR_HOME2_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_HOME_ADDRESS_CITY,
	BRUTUS_BRUTUS_PR_HOME_ADDRESS_COUNTRY,
	BRUTUS_BRUTUS_PR_HOME_ADDRESS_POSTAL_CODE,
	BRUTUS_BRUTUS_PR_HOME_ADDRESS_POST_OFFICE_BOX,
	BRUTUS_BRUTUS_PR_HOME_ADDRESS_STATE_OR_PROVINCE,
	BRUTUS_BRUTUS_PR_HOME_ADDRESS_STREET,
	BRUTUS_BRUTUS_PR_HOME_FAX_NUMBER,
	BRUTUS_BRUTUS_PR_HOME_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_ISDN_NUMBER,
	BRUTUS_BRUTUS_PR_LOCALITY,
	BRUTUS_BRUTUS_PR_LOCATION,
	BRUTUS_BRUTUS_PR_MANAGER_NAME,
	BRUTUS_BRUTUS_PR_MOBILE_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_NICKNAME,
	BRUTUS_BRUTUS_PR_OFFICE_LOCATION,
	BRUTUS_BRUTUS_PR_OTHER_ADDRESS_CITY,
	BRUTUS_BRUTUS_PR_OTHER_ADDRESS_COUNTRY,
	BRUTUS_BRUTUS_PR_OTHER_ADDRESS_POSTAL_CODE,
	BRUTUS_BRUTUS_PR_OTHER_ADDRESS_POST_OFFICE_BOX,
	BRUTUS_BRUTUS_PR_OTHER_ADDRESS_STATE_OR_PROVINCE,
	BRUTUS_BRUTUS_PR_OTHER_ADDRESS_STREET,
	BRUTUS_BRUTUS_PR_OTHER_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_PAGER_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_PERSONAL_HOME_PAGE,
	BRUTUS_BRUTUS_PR_POSTAL_CODE,
	BRUTUS_BRUTUS_PR_POST_OFFICE_BOX,
	BRUTUS_BRUTUS_PR_PRIMARY_FAX_NUMBER,
	BRUTUS_BRUTUS_PR_PRIMARY_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_RADIO_TELEPHONE_NUMBER,
	BRUTUS_BRUTUS_PR_SPOUSE_NAME,
	BRUTUS_BRUTUS_PR_STATE_OR_PROVINCE,
	BRUTUS_BRUTUS_PR_STREET_ADDRESS,
	BRUTUS_BRUTUS_PR_SURNAME,
	BRUTUS_BRUTUS_PR_TELEX_NUMBER,
	BRUTUS_BRUTUS_PR_TITLE,
	BRUTUS_BRUTUS_PR_TTYTDD_PHONE_NUMBER
};
static BrutusSearchDef*
brutus_PAB_search_mapping()
{
	int i;
	BrutusSearchDef *retv = NULL;
	const size_t len = 16;

	retv = g_new0(BrutusSearchDef, 1);
	if (!retv)
		return NULL;

	retv->mapping = g_new0(BrutusEBookSearchDef, len);
	if (!retv->mapping) {
		g_free(retv);
		return NULL;
	}
	retv->len = len;

	/*
	 * 'file_as' is special as I use the PR_ANR pseudo
	 * property for this, and only this, search_token.
	 */
	i = 0;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("file_as");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_ANR;

	i++;
	retv->mapping[i].tag_count = 3;
	retv->mapping[i].search_token = g_strdup("full_name");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_SURNAME;
	retv->mapping[i].search_tags[1] = BRUTUS_BRUTUS_PR_GIVEN_NAME;
	retv->mapping[i].search_tags[2] = BRUTUS_BRUTUS_PR_TITLE;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("given_name");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_GIVEN_NAME;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("family_name");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_SURNAME;

	i++;
	retv->mapping[i].tag_count = 2;
	retv->mapping[i].search_token = g_strdup("url");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_BUSINESS_HOME_PAGE;
	retv->mapping[i].search_tags[1] = BRUTUS_BRUTUS_PR_PERSONAL_HOME_PAGE;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("org");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_COMPANY_NAME;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("org_unit");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_DEPARTMENT_NAME;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("office");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_OFFICE_LOCATION;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("title");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_TITLE;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("manager");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_MANAGER_NAME;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("assistant");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_ASSISTANT;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("nickname");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_NICKNAME;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("spouse");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_SPOUSE_NAME;

	i++;
	retv->mapping[i].tag_count = 1;
	retv->mapping[i].search_token = g_strdup("note");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_COMMENT;

	i++;
	retv->mapping[i].tag_count = 18;
	retv->mapping[i].search_token = g_strdup("phone");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_ASSISTANT_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[1] = BRUTUS_BRUTUS_PR_BUSINESS_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[2] = BRUTUS_BRUTUS_PR_BUSINESS2_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[3] = BRUTUS_BRUTUS_PR_PRIMARY_FAX_NUMBER;
	retv->mapping[i].search_tags[4] = BRUTUS_BRUTUS_PR_CALLBACK_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[5] = BRUTUS_BRUTUS_PR_CAR_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[6] = BRUTUS_BRUTUS_PR_COMPANY_MAIN_PHONE_NUMBER;
	retv->mapping[i].search_tags[7] = BRUTUS_BRUTUS_PR_HOME_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[8] = BRUTUS_BRUTUS_PR_HOME2_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[9] = BRUTUS_BRUTUS_PR_HOME_FAX_NUMBER;
	retv->mapping[i].search_tags[10] = BRUTUS_BRUTUS_PR_ISDN_NUMBER;
	retv->mapping[i].search_tags[11] = BRUTUS_BRUTUS_PR_MOBILE_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[12] = BRUTUS_BRUTUS_PR_OTHER_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[13] = BRUTUS_BRUTUS_PR_PAGER_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[14] = BRUTUS_BRUTUS_PR_PRIMARY_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[15] = BRUTUS_BRUTUS_PR_RADIO_TELEPHONE_NUMBER;
	retv->mapping[i].search_tags[16] = BRUTUS_BRUTUS_PR_TELEX_NUMBER;
	retv->mapping[i].search_tags[17] = BRUTUS_BRUTUS_PR_TTYTDD_PHONE_NUMBER;

	i++;
	retv->mapping[i].tag_count = 20;
	retv->mapping[i].search_token = g_strdup("address");
	retv->mapping[i].search_tags = g_malloc(sizeof(BRUTUS_BDEFINE) * retv->mapping[i].tag_count);
	retv->mapping[i].search_tags[0] = BRUTUS_BRUTUS_PR_LOCATION;
	retv->mapping[i].search_tags[1] = BRUTUS_BRUTUS_PR_HOME_ADDRESS_STREET;
	retv->mapping[i].search_tags[2] = BRUTUS_BRUTUS_PR_HOME_ADDRESS_CITY;
	retv->mapping[i].search_tags[3] = BRUTUS_BRUTUS_PR_HOME_ADDRESS_STATE_OR_PROVINCE;
	retv->mapping[i].search_tags[4] = BRUTUS_BRUTUS_PR_HOME_ADDRESS_POSTAL_CODE;
	retv->mapping[i].search_tags[5] = BRUTUS_BRUTUS_PR_HOME_ADDRESS_COUNTRY;
	retv->mapping[i].search_tags[6] = BRUTUS_BRUTUS_PR_HOME_ADDRESS_POST_OFFICE_BOX;
	retv->mapping[i].search_tags[7] = BRUTUS_BRUTUS_PR_OFFICE_LOCATION;
	retv->mapping[i].search_tags[8] = BRUTUS_BRUTUS_PR_STREET_ADDRESS;
	retv->mapping[i].search_tags[9] = BRUTUS_BRUTUS_PR_LOCALITY;
	retv->mapping[i].search_tags[10] = BRUTUS_BRUTUS_PR_STATE_OR_PROVINCE;
	retv->mapping[i].search_tags[11] = BRUTUS_BRUTUS_PR_POSTAL_CODE;
	retv->mapping[i].search_tags[12] = BRUTUS_BRUTUS_PR_COUNTRY;
	retv->mapping[i].search_tags[13] = BRUTUS_BRUTUS_PR_POST_OFFICE_BOX;
	retv->mapping[i].search_tags[14] = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_STREET;
	retv->mapping[i].search_tags[15] = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_CITY;
	retv->mapping[i].search_tags[16] = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_STATE_OR_PROVINCE;
	retv->mapping[i].search_tags[17] = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_POSTAL_CODE;
	retv->mapping[i].search_tags[18] = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_COUNTRY;
	retv->mapping[i].search_tags[19] = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_POST_OFFICE_BOX;

	return retv;
}

static void
brutus_free_search_mapping(BrutusSearchDef *search_def)
{
	int i;

	if (!search_def)
		return;

	if (!search_def->mapping)
		goto out;

	for (i = 0; i < search_def->len; i++) {
		if (search_def->mapping[i].search_token)
			g_free(search_def->mapping[i].search_token);

		if (search_def->mapping[i].search_tags)
			g_free(search_def->mapping[i].search_tags);
	}
	g_free(search_def->mapping);

out:
	g_free(search_def);
}

static const BrutusEBookSearchDef*
brutus_find_search_mapping(const char *search_token,
			   BrutusSearchDef *search_def)
{
	guint i;

	for (i = 0; i < search_def->len; i++) {
		if (!strcmp(search_token, search_def->mapping[i].search_token))
			return &(search_def->mapping[i]);
	}

	return NULL;
}


static void
brutus_closure_destroy(BrutusBackendSearchClosure *closure)
{
	if (!closure)
		return;

	e_flag_free(closure->running);
	g_free(closure);
}

static BrutusBackendSearchClosure*
brutus_init_closure(EDataBookView *book_view,
		    EBookBackendBrutus *brutus_backend)
{
	BrutusBackendSearchClosure *closure = NULL;

	if (!book_view || !brutus_backend)
		return NULL;

	closure = g_new(BrutusBackendSearchClosure, 1);
	if (!closure)
		return NULL;

	closure->running = e_flag_new();
	if (!closure->running) {
		g_free(closure);
		return NULL;
	}
	closure->brutus_backend = brutus_backend;
	closure->thread = NULL;

	g_object_set_data_full(G_OBJECT(book_view),
			       G_OBJECT_BRUTUS_BOOK_VIEW,
			       closure,
			       (GDestroyNotify)brutus_closure_destroy);

	return closure;
}

static BrutusBackendSearchClosure*
brutus_get_closure(EDataBookView *book_view)
{
	return g_object_get_data(G_OBJECT(book_view),
				 G_OBJECT_BRUTUS_BOOK_VIEW);
}

static gchar*
brutus_get_book_name(ADDRESS_BOOK_KIND ab_type)
{
	gchar *name = NULL;

	switch (ab_type) {
	case ADDRESS_BOOK_PAB:
		name = g_strdup("PAB");
		break;
	case ADDRESS_BOOK_GAL:
		name = g_strdup("GAL");
		break;
	default:
		g_assert_not_reached();
	}

	return name;
}

/* only invoked from brutus_(dis)connect() and from finalize() when the backend is locked */
static void
brutus_mapi_shutdown(EBookBackendBrutus *brutus_backend)
{
	EBookBackendBrutusPrivate *priv = brutus_backend->priv;
	BrutusBaseClass *brutus_base = BRUTUS_BASE_CLASS(priv);
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	d("");
	// the server-side MAPI address books
	brutus_release_object(BRUTUS_IABContainer_tc,
			      (CORBA_Object*) &priv->GAL_book,
			      CORBA_OBJECT_NIL,
			      ev);
	brutus_release_object(BRUTUS_IMAPIFolder_tc,
			      (CORBA_Object*) &priv->PAB_book,
			      CORBA_OBJECT_NIL,
			      ev);

	// the server-side MAPI address containers
	brutus_release_object(BRUTUS_IAddrBook_tc,
			      (CORBA_Object*) &priv->addrbook,
			      CORBA_OBJECT_NIL,
			      ev);

	brutus_release_object(BRUTUS_IMsgStore_tc,
			      (CORBA_Object*) &priv->msg_store,
			      CORBA_OBJECT_NIL,
			      ev);

	/* mapi session and instance ID */
	if (brutus_is_connected(brutus_base->brutus_server,
				brutus_base->mapi_session))
		BRUTUS_IMAPISession_Destroy(brutus_base->mapi_session, brutus_base->instance_id, ev);
	cb = CORBA_Object_is_nil(brutus_base->mapi_session, ev);
	if (!ORBIT2_EX(ev) && !cb)
		CORBA_Object_release(brutus_base->mapi_session, ev);
	brutus_base->mapi_session = CORBA_OBJECT_NIL;
	brutus_base->instance_id = 0;

	CORBA_exception_free(ev);
}

struct BrutusConnectData {
	EBookBackendBrutus *brutus_backend;
	ESource *source;
};

static gboolean
brutus_connect(gpointer data)
{
	struct BrutusConnectData *cnx_data = NULL;
	EBookBackendBrutus *brutus_backend = NULL;
	ESource *source = NULL;
	char *prompt = NULL;
	char *err_prompt = NULL;
	char *windows_domain = NULL;
	char *username = NULL;
	char *mailbox = NULL;
	char *mailbox_dn = NULL;
	char *mailbox_email = NULL;
	gboolean retv = FALSE;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	BRUTUS_BrutusLogOn brutus_logon = CORBA_OBJECT_NIL;
	BRUTUS_BrutusLogOn_BrutusVersion brutus_server_version = { 0 };
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];

	d("");
	if (!data)
		return FALSE;
	cnx_data = (struct BrutusConnectData*)data;
	source = cnx_data->source;
	brutus_backend = cnx_data->brutus_backend;
	g_free(data);

	if (!brutus_backend)
		return FALSE;
	priv = brutus_backend->priv;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	if (!g_mutex_trylock(brutus_base->connect_mutex))
		goto exit;
	brutus_set_likely_connected(brutus_base, FALSE);

	CORBA_exception_init(ev);

	/* check connection status */
	retv = brutus_is_connected(brutus_base->brutus_server, brutus_base->mapi_session);
	if (retv)
		goto out; // connected
	else
		brutus_mapi_shutdown(brutus_backend); // not connected - resetting MAPI

#ifndef HAVE_ORBIT2_TIMEOUT
	if (!brutus_server_ping(brutus_base->brutus_server))
		goto out; // no reason to even try...
#endif

	windows_domain = brutus_strip_trailing_and_leading_whitespace(e_source_get_property(source, "windows_domain"));
	if (!windows_domain)
		windows_domain = g_strdup("");
	if (!strlen(brutus_base->effective_mailbox)) 
		goto out; // no way we can connect
	mailbox = brutus_strip_trailing_and_leading_whitespace(e_source_get_property(source, "exchange_mailbox"));
	mailbox_dn = brutus_strip_trailing_and_leading_whitespace(e_source_get_property(source, "exchange_mailbox_dn"));
	mailbox_email = brutus_strip_trailing_and_leading_whitespace(e_source_get_property(source, "exchange_mailbox_email")); 	

	// construct unique username
	username = g_strconcat(windows_domain, "\\", e_source_get_property(source, "windows_user"), NULL);
	if (!username)
		goto out;

	d("Attempting to query a session from the proxy");
	{
		BRUTUS_BrutusProxy_QueryState query_state = BRUTUS_BrutusProxy_offerInProgress;

		d(brutus_base->exchange_server);
		do {
			query_state = BRUTUS_BrutusProxy_querySession(brutus_base->brutus_proxy,
								      username,
								      brutus_base->exchange_server,
								      &brutus_base->mapi_session,
								      ev);
			if (ORBIT2_EX(ev)) {
				if (!strcmp(CORBA_exception_id(ev), ex_CORBA_NO_PERMISSION)) {
					g_usleep(100000); // sleep for 0.1 second
					continue;
				}
				PRINT_ORBIT2_EX(ev);
				goto proceed;
			}
			switch (query_state) {
			case BRUTUS_BrutusProxy_mustOffer :
				goto proceed;
			case BRUTUS_BrutusProxy_offerInProgress :
				g_usleep(200000); // sleep for 0.2 seconds
				break;
			case BRUTUS_BrutusProxy_queryOK : // got session
				break;
			default:
				d("Unknown query response");
				goto proceed;
			}
		} while (BRUTUS_BrutusProxy_queryOK != query_state);

		// check session for validity
		cb = CORBA_Object_is_nil((CORBA_Object)brutus_base->mapi_session, ev);
		if (!ORBIT2_EX(ev) && !cb) {
			brutus_base->instance_id = BRUTUS_IMAPISession_addLifeline(brutus_base->mapi_session, brutus_base->lifeline.wrapper, ev);
			if (ORBIT2_EX(ev)) {
				d("Could not add lifeline");
				goto proceed;
			}

			d("Query successful");
			retv = TRUE; // we are connected, no query lock to release
			goto out_get_book;
		}
	}

proceed:
	d("Could not get session from proxy, must log on explicitly");
	do { 
		BRUTUS_BrutusProxy_getProfile(brutus_base->brutus_proxy,
					      username,
					      brutus_base->exchange_server,
					      &brutus_base->effective_mailbox,
					      &brutus_base->mapi_profile,
					      ev);
		if (ORBIT2_EX(ev)) {
			brutus_base->effective_mailbox = NULL;
			brutus_base->mapi_profile = NULL;
			if (!strcmp(CORBA_exception_id(ev), ex_CORBA_NO_PERMISSION)) {
				g_usleep(100000); // sleep for 0.1 second
				continue;
			}
			PRINT_ORBIT2_EX(ev);
		} else if (!strlen(brutus_base->mapi_profile)) {
			d("No stored MAPI Profile - generating a new one");
			uuid_str_t uuid_str;

			CORBA_free(brutus_base->mapi_profile);
			brutus_get_random_uuid(uuid_str);
			brutus_base->mapi_profile = CORBA_string_dup(uuid_str);
		}
		break;
	} while (TRUE);
	dfmt("Using this MAPI Profile: \"%s\"", brutus_base->mapi_profile);

	d("Getting reference to BrutusLogOn object");
	{
		brutus_logon = (BRUTUS_BrutusLogOn)CORBA_ORB_string_to_object(brutus_base->orb, brutus_base->brutus_logon_corbaloc, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_release_lock;
		}

		cb = CORBA_Object_is_nil(brutus_logon, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			cb = TRUE;
		}
		if (cb) {
			d("Could not resolve (string_to_object) BrutusLogOn, NIL object returned");
			goto out_release_lock;
		} else
			d("BrutusLogOn resolved");
		d("BrutusLogOn reference has been acquired");

		BRUTUS_BrutusLogOn_GetVersion(brutus_logon, &brutus_server_version, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_release_lock;
		}
		if (!brutus_server_version_OK((int)brutus_server_version.major,
					      (int)brutus_server_version.minor,
					      (int)brutus_server_version.micro,
					      (int)brutus_server_version.release_tag)) {
			d("Unsupported version of Brutus Server. You must have Brutus Server version "BRUTUS_SERVER_REQ_VERSION_STRING);
			goto out_release_lock;
		}
		d("Brutus Server version check completed");
	}
	prompt = g_strdup_printf(_("Please enter Windows domain password for \"%s\\%s\""),
				 windows_domain,
				 e_source_get_property(source, "windows_user"));
	err_prompt = g_strdup_printf(_("Authentication attempt failed - Please reenter Windows domain password for \"%s\\%s\""),
				     windows_domain,
				     e_source_get_property(source, "windows_user"));
	if (!prompt || !err_prompt)
		d("No memory");
	else {
		retv = brutus_logon_loop(brutus_logon,
					 brutus_base->lifeline.wrapper,
					 &brutus_base->mapi_session,
					 &brutus_base->instance_id,
					 brutus_base->brutus_logon_corbaloc,
					 prompt,
					 err_prompt,
					 e_source_get_property(source, "windows_user"),
					 windows_domain,
					 mailbox,
					 mailbox_dn,
					 mailbox_email,
					 brutus_base->exchange_server,
					 &brutus_base->mapi_profile,
					 &brutus_base->effective_mailbox);
	}

out_release_lock:
	if (retv) {
		d("Offering IMAPISession to Brutus proxy");
		do {
			BRUTUS_BrutusProxy_offerSession(brutus_base->brutus_proxy,
							brutus_base->mapi_session,
							username,
							brutus_base->exchange_server,
							brutus_base->effective_mailbox,
							brutus_base->mapi_profile,
							ev);
			if (ORBIT2_EX(ev) && !strcmp(CORBA_exception_id(ev), ex_CORBA_NO_PERMISSION)) {
				g_usleep(100000); // sleep for 0.1 second
				continue;
			}
			break;
		} while (TRUE);
		PRINT_IF_ORBIT2_EX(ev);
	} else {
		d("No IMAPISession offer - releasing query lock");
		do {
			BRUTUS_BrutusProxy_offerSession(brutus_base->brutus_proxy,
							CORBA_OBJECT_NIL,
							"",
							"",
							"",
							"",
							ev);
			if (ORBIT2_EX(ev) && !strcmp(CORBA_exception_id(ev), ex_CORBA_NO_PERMISSION)) {
				g_usleep(100000); // sleep for 0.1 second
				continue;
			}
			break;
		} while (TRUE);
		PRINT_IF_ORBIT2_EX(ev);
	}

out_get_book:
	// get the right mapi object - GAL or PAB
	if (retv) {
		retv = FALSE; // need to check for book availability before setting likely connected status
		switch (priv->kind) {
		case ADDRESS_BOOK_PAB :
			br = brutus_get_default_message_store(brutus_base->mapi_session,
							      &priv->msg_store,
							      ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto out;
			}
			if (BRUTUS_BRUTUS_S_OK != br) {
				d(brutus_bresult_to_str(br));
				goto out;
			}
			cb = CORBA_Object_is_nil(priv->msg_store, ev);
			if (ORBIT2_EX(ev) || cb) {
				if (!cb) {
					CORBA_Object_release(priv->msg_store, ev);
					priv->msg_store = CORBA_OBJECT_NIL;
				}
				goto out;
			}

			br = BRUTUS_IMAPISession_GetOutlookFolder(brutus_base->mapi_session,
								  BRUTUS_OUTLOOK_CONTACTS,
								  priv->msg_store,
								  &priv->PAB_book,
								  ev);
			break;
		case ADDRESS_BOOK_GAL :
			br = BRUTUS_IMAPISession_OpenAddressBook(brutus_base->mapi_session,
								 "",
								 0,
								 &priv->addrbook,
								 ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto out;
			}
			if (BRUTUS_BRUTUS_S_OK != br) {
				d(brutus_bresult_to_str(br));
				goto out;
			}
			cb = CORBA_Object_is_nil(priv->addrbook, ev);
			if (ORBIT2_EX(ev) || cb) {
				if (!cb) {
					CORBA_Object_release(priv->addrbook, ev);
					priv->addrbook = CORBA_OBJECT_NIL;
				}
				goto out;
			}
			br = BRUTUS_IAddrBook_GAL(priv->addrbook, &priv->GAL_book, ev);
			break;
		default :
			g_assert_not_reached();
		}
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			goto out;
		}
		switch (priv->kind) {
		case ADDRESS_BOOK_PAB :
			cb = CORBA_Object_is_nil(priv->PAB_book, ev);
			if (ORBIT2_EX(ev) || cb) {
				if (!cb) {
					CORBA_Object_release(priv->PAB_book, ev);
					priv->PAB_book = CORBA_OBJECT_NIL;
				}
			}
			break;
		case ADDRESS_BOOK_GAL :
			cb = CORBA_Object_is_nil(priv->GAL_book, ev);
			if (ORBIT2_EX(ev) || cb) {
				if (!cb) {
					CORBA_Object_release(priv->GAL_book, ev);
					priv->GAL_book = CORBA_OBJECT_NIL;
				}
			}
			break;
		default :
			g_assert_not_reached();
		}
		retv = TRUE; // all is well

		d("acquired Address Book");
	}
out:
	g_free(prompt);
	g_free(err_prompt);
	g_free(windows_domain);
	g_free(username);
	g_free(mailbox);
	g_free(mailbox_dn);
	g_free(mailbox_email);

	cb = CORBA_Object_is_nil(brutus_logon, ev);
	if (!ORBIT2_EX(ev) && !cb)
		CORBA_Object_release(brutus_logon, ev);

	CORBA_exception_free(ev);

	if (source)
		g_object_unref(G_OBJECT(source));

	brutus_set_likely_connected(brutus_base, retv);

	g_mutex_unlock(brutus_base->connect_mutex);
exit:
	return FALSE;
}

static gboolean
brutus_orb_init(EBookBackendBrutusPrivate *priv)
{
	static guint orb_num = 0;
	BrutusBaseClass *brutus_base = BRUTUS_BASE_CLASS(priv);
	gboolean retv = FALSE;
	char *orb_name = NULL;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	// Hmm... I am unsure as to whether the EBook sub system will invoke the same backend
	// object more than once, so I put this here to avoid identical ORB names for certainty.
	orb_num++;
	orb_name = g_strdup_printf("%d-%s-%s", orb_num, priv->book_name, "Brutus_EBook_ORB_orbit-io-thread");
	if (!orb_name) {
		d("No memory");
		goto out;
	}

	d("Invoking ORB_Init()");
	brutus_base->orb = brutus_create_orb(0,
					     NULL,
					     "BrutusLogOn",
					     BRUTUS_BrutusLogOn_IOR_TABLE_KEY,
					     brutus_base->brutus_server,
					     brutus_base->brutus_port,
					     orb_name,
					     ev);
	g_free(orb_name);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	cb = CORBA_Object_is_nil((CORBA_Object)brutus_base->orb, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (cb) {
		d("Could not initialize the ORB");
		goto out;
	}

	d("Getting root POA");
	brutus_base->root_poa = brutus_get_root_poa(brutus_base->orb, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	cb = CORBA_Object_is_nil((CORBA_Object)brutus_base->root_poa, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (cb) {
		d("Could not get the root POA");
		goto out;
	}
	retv = TRUE;

out:
	CORBA_exception_free(ev);

	return retv;
}

static gboolean
brutus_backend_construct(EBookBackendBrutus *brutus_backend,
			 ESource *source)
{
	gboolean retv = FALSE;
	EBookBackendBrutusPrivate *priv = brutus_backend->priv;
	BrutusBaseClass *brutus_base = BRUTUS_BASE_CLASS(priv);
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];
	const char *tmp = NULL;
	enum E_SOURCE_BRUTUS_TYPE e_type;

	d("");
	CORBA_exception_init(ev);

	if (!brutus_backend || !source)
		goto out;

	if (!e_source_get_property(source, "mapi_profile")
	    || !e_source_get_property(source, "exchange_server")
	    || !e_source_get_property(source, "windows_user")
	    || !e_source_get_property(source, "effective_mailbox")
	    || !e_source_get_property(source, "brutus_server")
	    || !e_source_get_property(source, "brutus_kind")
	    || !e_source_get_property(source, "brutus_port")) {
		d("Account not configured");
		goto out;
	}

	tmp = e_source_get_property(source, "brutus_kind");
	e_type = brutus_int_to_e_type(atoi(tmp));
	switch (e_type) {
	case E_SOURCE_BRUTUS_ADDRESSBOOK_PAB :
		priv->kind = ADDRESS_BOOK_PAB;
		priv->search_def = brutus_PAB_search_mapping();
		priv->content_tags = PAB_content_tags;
		break;
	case E_SOURCE_BRUTUS_ADDRESSBOOK_GAL :
		priv->kind = ADDRESS_BOOK_GAL;
		priv->search_def = brutus_GAL_search_mapping();
		priv->content_tags = GAL_content_tags;
		break;
	default :
		g_assert_not_reached();
	}
	if (!priv->search_def) {
		d("No memory");
		goto out;
	}

	priv->book_name = brutus_get_book_name(priv->kind);
	if (!priv->book_name) {
		d("No memory");
		goto out;
	}

	brutus_base->expect_other_clients = e_source_get_property(source, "expect_other_clients") ? TRUE : FALSE;

	brutus_base->brutus_server = brutus_strip_trailing_and_leading_whitespace(e_source_get_property(source, "brutus_server"));
	if (!brutus_base->brutus_server) {
		d("No memory");
		goto out;
	}

	brutus_base->exchange_server = brutus_strip_trailing_and_leading_whitespace(e_source_get_property(source, "exchange_server"));
	if (!brutus_base->exchange_server) {
		d("No memory");
		goto out;
	}

	// construct the name of the addressbook cache directory
	{
		gchar *mangled_uri = NULL;

		if (!priv->uri)
			goto out;

		mangled_uri = brutus_mangle_uri(priv->uri);
		d(mangled_uri);
		brutus_base->cache_dir = g_build_filename(g_get_home_dir(),
							  ".evolution/cache/addressbook",
							  mangled_uri,
							  "/",
							  NULL);
		g_free(mangled_uri);
	}
	d(brutus_base->cache_dir);

	d("Initializing CORBA");
	{
		gboolean orb_ok = FALSE;
		guint64 brutus_port = 951;

		if (!brutus_get_xbit_number(16, e_source_get_property(source, "brutus_port"), &brutus_port)) {
			d("Invalid Brutus port in url");
			goto out;
		}
		brutus_base->brutus_port = brutus_port;

		brutus_base->brutus_logon_corbaloc = brutus_build_corbaloc_string(brutus_base->brutus_server,
										  BRUTUS_BrutusLogOn_IOR_TABLE_KEY,
										  brutus_base->brutus_port);
		orb_ok = brutus_orb_init(priv);
		if (!orb_ok) {
			d("Could not initiate ORB");
			goto out;
		}
	}

	d("Getting root POAManager, activating the RootPOA and instantiating the lifeline");
	{
		PortableServer_POAManager root_poa_manager = CORBA_OBJECT_NIL;
		gboolean use_local_proxy = !(ORBit_proto_use("IPv4") || ORBit_proto_use("IPv6"));

		root_poa_manager = PortableServer_POA__get_the_POAManager(brutus_base->root_poa, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_ex;
		}

		cb = CORBA_Object_is_nil((CORBA_Object)root_poa_manager, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_ex;
		}
		if (cb) {
			d("Could not get the root POAManager");
			goto out_ex;
		}
		PortableServer_POAManager_activate(root_poa_manager, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_ex;
		}

		CORBA_Object_release((CORBA_Object)root_poa_manager, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_ex;
		}

		// get brutus proxy
		{
			gchar *ref_path = g_strconcat(getenv("HOME"), "/", BRUTUSD_REF_FILE_NAME, NULL);

			brutus_base->brutus_proxy = (BRUTUS_BrutusProxy)brutus_object_ref_from_file(brutus_base->orb, ref_path);
			g_free(ref_path);
			cb = CORBA_Object_is_nil((CORBA_Object)brutus_base->brutus_proxy, ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto out_ex;
			}
			if (cb) {
				d("Could not get the Brutus proxy reference");
				goto out_ex;
			}
		}

		// the lifeline
		brutus_base->lifeline.filling = impl_BRUTUS_BrutusCheck__create(brutus_base->root_poa, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_ex;
		} else
			d("brutus_base->lifeline.filling created");

		if (use_local_proxy) {
			d("Using local network capable Brutus proxy");
			brutus_base->lifeline.key = BRUTUS_BrutusProxy_wrapObject(brutus_base->brutus_proxy,
										  BRUTUS_BrutusCheck_tc,
										  brutus_base->lifeline.filling,
										  &brutus_base->lifeline.wrapper,
										  ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto out_ex;
			}
			d("Local network capable Brutus proxy wrapped OK");
		} else {
			/* this will enable us to use "wrapper" unconditionally */
			d("Local network capable Brutus proxy not used");
			brutus_base->lifeline.wrapper = CORBA_Object_duplicate(brutus_base->lifeline.filling, ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto out_ex;
			}
		}

		goto no_err; // all is fine

	out_ex:
		d("Could not construct the EBook backend - CORBA exception caught");
		goto out;
	}

no_err:
	retv = TRUE;

out:
	CORBA_exception_free(ev);

	return retv;
}


/**************************************************************
 *                                                            *
 *        Conversion functions: EContact <==> IMessage        *
 *                                                            *
 *************************************************************/

static void
brutus_econtact_to_GAL_proplist(EContact *contact,
				BRUTUS_seq_SPropValue *propval_list)
{
	CORBA_unsigned_long i = 0;
	char *str_name = NULL;
	char *tmp = NULL;
	const char *inp = NULL;
	EContactName *contact_name = NULL;
	EContactAddress *contact_addr = NULL;

	// the length is known
	propval_list->_maximum = 24;
	propval_list->_length = propval_list->_maximum;
	propval_list->_buffer = BRUTUS_seq_SPropValue_allocbuf(propval_list->_maximum);
	if (!propval_list->_buffer) {
		propval_list->_maximum = 0;
		propval_list->_length = 0;
		return;
	}
	CORBA_sequence_set_release(propval_list, CORBA_FALSE);

	i = 0;
	inp = e_contact_get_const(contact, E_CONTACT_FILE_AS) ? e_contact_get_const(contact, E_CONTACT_FILE_AS) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_DISPLAY_NAME;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	i = 1;
	inp = e_contact_get_const(contact, E_CONTACT_FILE_AS) ? e_contact_get_const(contact, E_CONTACT_FILE_AS) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_7BIT_DISPLAY_NAME;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	{ // get, and set, all name details
		str_name = e_contact_get(contact, E_CONTACT_FULL_NAME);
		if (str_name)
			contact_name = e_contact_name_from_string(str_name);
		else
			contact_name = NULL;

		i = 2;
		tmp = g_strdup_printf("%s %s",
				      (const char*)((contact_name && contact_name->family) ? contact_name->family : ""),
				      (const char*)((contact_name && contact_name->suffixes) ? contact_name->suffixes : ""));
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_SURNAME;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup((const char*)(tmp ? tmp : ""));
		if (tmp)
			g_free(tmp);

		i = 3;
		tmp = g_strdup_printf("%s %s",
				      (const char*)((contact_name && contact_name->given) ? contact_name->given : ""),
				      (const char*)((contact_name && contact_name->additional) ? contact_name->additional : ""));
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_GIVEN_NAME;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup((const char*)(tmp ? tmp : ""));
		if (tmp)
			g_free(tmp);

		i = 4;
		if (contact_name)
			inp = contact_name->prefixes;
		else
			inp = NULL;
		if (!inp || !strcmp(inp, ""))
			inp = e_contact_get_const(contact, E_CONTACT_TITLE) ? e_contact_get_const(contact, E_CONTACT_TITLE) : "";
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_TITLE;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

		if (contact_name)
			e_contact_name_free(contact_name);
	}

	{ // get, and set, the most significant email address, if any
		inp = e_contact_get_const(contact, E_CONTACT_EMAIL_1);
		while (!inp) {
			inp = e_contact_get_const(contact, E_CONTACT_EMAIL_2);
			if (inp)
				break;

			inp = e_contact_get_const(contact, E_CONTACT_EMAIL_3);
			if (inp)
				break;

			inp = e_contact_get_const(contact, E_CONTACT_EMAIL_4);
			break;
		}
		i = 5;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_SMTP_ADDRESS;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");
	}

	i = 6;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_BUSINESS) ? e_contact_get_const(contact, E_CONTACT_PHONE_BUSINESS) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_BUSINESS_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	i = 7;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_PAGER) ? e_contact_get_const(contact, E_CONTACT_PHONE_PAGER) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PAGER_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	i = 8;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_HOME) ? e_contact_get_const(contact, E_CONTACT_PHONE_HOME) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_HOME_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	i = 9;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_MOBILE) ? e_contact_get_const(contact, E_CONTACT_PHONE_MOBILE) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_MOBILE_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	i = 10;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_BUSINESS_FAX) ? e_contact_get_const(contact, E_CONTACT_PHONE_BUSINESS_FAX) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PRIMARY_FAX_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	i = 11;
	inp = e_contact_get_const(contact, E_CONTACT_ORG) ? e_contact_get_const(contact, E_CONTACT_ORG) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_COMPANY_NAME;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	i = 12;
	inp = e_contact_get_const(contact, E_CONTACT_ORG_UNIT) ? e_contact_get_const(contact, E_CONTACT_ORG_UNIT) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_DEPARTMENT_NAME;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	i = 13;
	inp = e_contact_get_const(contact, E_CONTACT_OFFICE) ? e_contact_get_const(contact, E_CONTACT_OFFICE) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_OFFICE_LOCATION;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	i = 14;
	inp = e_contact_get_const(contact, E_CONTACT_ROLE) ? e_contact_get_const(contact, E_CONTACT_ROLE) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMS_AB_DESCRIPTION;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	i = 15;
	inp = e_contact_get_const(contact, E_CONTACT_NOTE) ? e_contact_get_const(contact, E_CONTACT_NOTE) : "";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_COMMENT;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

	{ // null terminated strings instead of NULLs
		contact_addr = e_contact_get(contact, E_CONTACT_ADDRESS_HOME);

		i = 16;
		inp = (contact_addr && contact_addr->po && *contact_addr->po) ? contact_addr->po : "";
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_POST_OFFICE_BOXES;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_MV_STRING8;
		propval_list->_buffer[i].Value._u.MVszA._maximum = 1;
		propval_list->_buffer[i].Value._u.MVszA._length = propval_list->_buffer[i].Value._u.MVszA._maximum;
		propval_list->_buffer[i].Value._u.MVszA._buffer = BRUTUS_SLPSTRArray_allocbuf(propval_list->_buffer[i].Value._u.MVszA._maximum);
		propval_list->_buffer[i].Value._u.MVszA._buffer[0] = CORBA_string_dup(inp);

		i = 17;
		inp = (contact_addr && contact_addr->ext && *contact_addr->ext) ? contact_addr->ext : "";
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_OFFICE_LOCATION;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

		i = 18;
		inp = (contact_addr && contact_addr->street && *contact_addr->street) ? contact_addr->street : "";
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_STREET_ADDRESS;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

		i = 19;
		inp = (contact_addr && contact_addr->locality && *contact_addr->locality) ? contact_addr->locality : "";
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_LOCALITY;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

		i = 20;
		inp = (contact_addr && contact_addr->region && *contact_addr->region) ? contact_addr->region : "";
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_STATE_OR_PROVINCE;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

		i = 21;
		inp = (contact_addr && contact_addr->code && *contact_addr->code) ? contact_addr->code : "";
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_POSTAL_CODE;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

		i = 22;
		inp = (contact_addr && contact_addr->country && *contact_addr->country) ? contact_addr->country : "";
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_COUNTRY;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp);

		if (contact_addr)
			e_contact_address_free(contact_addr);
	}

	i = 23;
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_DISPLAY_TYPE;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_LONG;
	propval_list->_buffer[i].Value._u.l = e_contact_get(contact, E_CONTACT_IS_LIST) ? BRUTUS_BRUTUS_DT_DISTLIST : BRUTUS_BRUTUS_DT_MAILUSER;
}

/*
 * 'i' points to the last entry in the property list.
 * The first address data will be added at entry (i).
 * i will point, upon return, at the next available entry.
 * Return: Always the number of entries to be added.
 */
static void
brutus_econtact_extract_address(const EContactField addr_type,
				EContact *contact,
				CORBA_char **office_location,
				CORBA_char **street,
				CORBA_char **city,
				CORBA_char **state,
				CORBA_char **postal_code,
				CORBA_char **country,
				CORBA_char **po_box)
{
	EContactAddress *contact_addr = NULL;
	const char *inp = NULL;

	// contact_address has null terminated strings instead of NULLs
	if (contact)
		contact_addr = e_contact_get(contact, addr_type);

	inp = (contact_addr && contact_addr->ext && *contact_addr->ext) ? contact_addr->ext : "";
	*office_location = CORBA_string_dup(inp);

	inp = (contact_addr && contact_addr->street && *contact_addr->street) ? contact_addr->street : "";
	*street = CORBA_string_dup(inp);

	inp = (contact_addr && contact_addr->locality && *contact_addr->locality) ? contact_addr->locality : "";
	*city = CORBA_string_dup(inp);

	inp = (contact_addr && contact_addr->region && *contact_addr->region) ? contact_addr->region : "";
	*state = CORBA_string_dup(inp);

	inp = (contact_addr && contact_addr->code && *contact_addr->code) ? contact_addr->code : "";
	*postal_code = CORBA_string_dup(inp);

	inp = (contact_addr && contact_addr->country && *contact_addr->country) ? contact_addr->country : "";
	*country = CORBA_string_dup(inp);

	inp = (contact_addr && contact_addr->po && *contact_addr->po) ? contact_addr->po : "";
	*po_box = CORBA_string_dup(inp);

	if (contact_addr)
		e_contact_address_free(contact_addr);
}

static gboolean
brutus_econtact_to_PAB_proplist(EContact *contact,
				BRUTUS_seq_SPropValue       *propval_list,
				BRUTUS_seq_SPropValue       *named_pubstr_propval_list,
				BRUTUS_seq_NamedPropertyTag *named_pubstr_prop_tags,
				BRUTUS_seq_SPropValue       *named_contact_propval_list,
				BRUTUS_seq_NamedPropertyTag *named_contact_prop_tags,
				BRUTUS_seq_SPropValue       *named_common_propval_list,
				BRUTUS_seq_NamedPropertyTag *named_common_prop_tags)
{
	gboolean retv = FALSE;
	CORBA_unsigned_long i = 0;
	char *str_name = NULL;
	char *tmp = NULL;
	const char *name_title = NULL;
	const char *given_name = NULL;
	const char *middle_name = NULL;
	const char *last_name = NULL;
	const char *generation = NULL;
	const char *inp = NULL;
	EContactName *contact_name = NULL;
	EContactDate *ec_date_birth = NULL;
	EContactDate *ec_date_anniversary = NULL;
	struct BrutusGregorianTime bgt;

	propval_list->_buffer = NULL;
	named_pubstr_prop_tags->_buffer = NULL;
	named_pubstr_propval_list->_buffer = NULL;
	named_contact_prop_tags->_buffer = NULL;
	named_contact_propval_list->_buffer = NULL;
	named_common_prop_tags->_buffer = NULL;
	named_common_propval_list->_buffer = NULL;

	ec_date_birth = e_contact_get(contact, E_CONTACT_BIRTH_DATE);
	ec_date_anniversary = e_contact_get(contact, E_CONTACT_ANNIVERSARY);

	// standard properties
	propval_list->_maximum = 56 + !!ec_date_birth + !!ec_date_anniversary;
	propval_list->_length = propval_list->_maximum;
	propval_list->_buffer = BRUTUS_seq_SPropValue_allocbuf(propval_list->_maximum);
	if (!propval_list->_buffer)
		goto out;
	CORBA_sequence_set_release(propval_list, CORBA_FALSE);

	//
	// named properties
	//

	// initiate the named property tags - PS_PROPSET_CONTACT
	named_contact_prop_tags->_maximum = 20;
	named_contact_prop_tags->_length = named_contact_prop_tags->_maximum;
	named_contact_prop_tags->_buffer = BRUTUS_seq_NamedPropertyTag_allocbuf(named_contact_prop_tags->_maximum);
	if (!named_contact_prop_tags->_buffer)
		goto out;
	CORBA_sequence_set_release(named_contact_prop_tags, CORBA_FALSE);

	// initiate the named property values
	named_contact_propval_list->_maximum = named_contact_prop_tags->_maximum;
	named_contact_propval_list->_length = named_contact_propval_list->_maximum;
	named_contact_propval_list->_buffer = BRUTUS_seq_SPropValue_allocbuf(named_contact_propval_list->_maximum);
	if (!named_contact_propval_list->_buffer)
		goto out;
	CORBA_sequence_set_release(named_contact_propval_list, CORBA_FALSE);

	// initiate the named property tags - PS_PROPSET_COMMON
	named_common_prop_tags->_maximum = 1;
	named_common_prop_tags->_length = named_common_prop_tags->_maximum;
	named_common_prop_tags->_buffer = BRUTUS_seq_NamedPropertyTag_allocbuf(named_common_prop_tags->_maximum);
	if (!named_common_prop_tags->_buffer)
		goto out;
	CORBA_sequence_set_release(named_common_prop_tags, CORBA_FALSE);

	// initiate the named property values
	named_common_propval_list->_maximum = named_common_prop_tags->_maximum;
	named_common_propval_list->_length = named_common_propval_list->_maximum;
	named_common_propval_list->_buffer = BRUTUS_seq_SPropValue_allocbuf(named_common_propval_list->_maximum);
	if (!named_common_propval_list->_buffer)
		goto out;
	CORBA_sequence_set_release(named_common_propval_list, CORBA_FALSE);

	// initiate the named property tags - PS_PUBLIC_STRINGS
	named_pubstr_prop_tags->_maximum = 1;
	named_pubstr_prop_tags->_length = named_pubstr_prop_tags->_maximum;
	named_pubstr_prop_tags->_buffer = BRUTUS_seq_NamedPropertyTag_allocbuf(named_pubstr_prop_tags->_maximum);
	if (!named_pubstr_prop_tags->_buffer)
		goto out;
	CORBA_sequence_set_release(named_pubstr_prop_tags, CORBA_FALSE);

	// initiate the named property values
	named_pubstr_propval_list->_maximum = named_pubstr_prop_tags->_maximum;
	named_pubstr_propval_list->_length = named_pubstr_propval_list->_maximum;
	named_pubstr_propval_list->_buffer = BRUTUS_seq_SPropValue_allocbuf(named_pubstr_propval_list->_maximum);
	if (!named_pubstr_propval_list->_buffer)
		goto out;
	CORBA_sequence_set_release(named_pubstr_propval_list, CORBA_FALSE);

	/********************************************
	 *           Standard Properties            *
	 ********************************************/

	{ // get all name details
		str_name = e_contact_get(contact, E_CONTACT_FULL_NAME);
		if (str_name)
			contact_name = e_contact_name_from_string(str_name);
		else
			contact_name = NULL;

		name_title = e_contact_get_const(contact, E_CONTACT_TITLE) ? e_contact_get_const(contact, E_CONTACT_TITLE) : "";
		given_name = (contact_name && contact_name->given) ? contact_name->given : "";
		middle_name = (contact_name && contact_name->additional) ? contact_name->additional : "";
		last_name = (contact_name && contact_name->family) ? contact_name->family : "";
		generation = (contact_name && contact_name->suffixes) ? contact_name->suffixes : "";

		i = 0;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_DISPLAY_NAME_PREFIX;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(name_title);

		i = 1;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_GIVEN_NAME;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(given_name);

		i = 2;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_MIDDLE_NAME;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(middle_name);

		i = 3;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_SURNAME;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(last_name);

		i = 4;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_GENERATION;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(generation);

		i = 5;
		tmp = g_strdup_printf("%s %s %s %s %s", name_title, given_name, middle_name, last_name, generation);
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_DISPLAY_NAME;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup((const char*)(tmp ? tmp : ""));
		if (tmp)
			g_free(tmp);
	}

	i = 6;
	inp = e_contact_get_const(contact, E_CONTACT_NICKNAME);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_NICKNAME;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	{ // Home address
		i = 7;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_LOCATION;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 8;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_HOME_ADDRESS_STREET;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 9;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_HOME_ADDRESS_CITY;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 10;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_HOME_ADDRESS_STATE_OR_PROVINCE;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 11;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_HOME_ADDRESS_POSTAL_CODE;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 12;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_HOME_ADDRESS_COUNTRY;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 13;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_HOME_ADDRESS_POST_OFFICE_BOX;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		brutus_econtact_extract_address(E_CONTACT_ADDRESS_HOME,
						contact,
						&(propval_list->_buffer[7].Value._u.lpszA),   // location
						&(propval_list->_buffer[8].Value._u.lpszA),   // street,
						&(propval_list->_buffer[9].Value._u.lpszA),   // city
						&(propval_list->_buffer[10].Value._u.lpszA),  // state or province
						&(propval_list->_buffer[11].Value._u.lpszA),  // postal code
						&(propval_list->_buffer[12].Value._u.lpszA),  // country
						&(propval_list->_buffer[13].Value._u.lpszA)); // PO box
	}

	{ // Business address
		i = 14;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_OFFICE_LOCATION;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 15;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_STREET_ADDRESS;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 16;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_LOCALITY;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 17;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_STATE_OR_PROVINCE;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 18;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_POSTAL_CODE;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 19;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_COUNTRY;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 20;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_POST_OFFICE_BOX;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		brutus_econtact_extract_address(E_CONTACT_ADDRESS_WORK,
						contact,
						&(propval_list->_buffer[14].Value._u.lpszA),  // office location
						&(propval_list->_buffer[15].Value._u.lpszA),  // street,
						&(propval_list->_buffer[16].Value._u.lpszA),  // city
						&(propval_list->_buffer[17].Value._u.lpszA),  // state or province
						&(propval_list->_buffer[18].Value._u.lpszA),  // postal code
						&(propval_list->_buffer[19].Value._u.lpszA),  // country
						&(propval_list->_buffer[20].Value._u.lpszA)); // PO box
	}

	{ // Other address
		CORBA_char *loc = NULL;

		i = 21;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_STREET;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 22;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_CITY;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 23;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_STATE_OR_PROVINCE;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 24;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_POSTAL_CODE;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 25;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_COUNTRY;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 26;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_OTHER_ADDRESS_POST_OFFICE_BOX;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		brutus_econtact_extract_address(E_CONTACT_ADDRESS_OTHER,
						contact,
						&loc,    // office location
						&(propval_list->_buffer[21].Value._u.lpszA),  // street,
						&(propval_list->_buffer[22].Value._u.lpszA),  // city
						&(propval_list->_buffer[23].Value._u.lpszA),  // state or province
						&(propval_list->_buffer[24].Value._u.lpszA),  // postal code
						&(propval_list->_buffer[25].Value._u.lpszA),  // country
						&(propval_list->_buffer[26].Value._u.lpszA)); // PO box
		if (loc)
			CORBA_free(loc);
	}

	i = 27;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_ASSISTANT);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_ASSISTANT_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 28;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_BUSINESS);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_BUSINESS_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 29;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_BUSINESS_2);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_BUSINESS2_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 30;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_BUSINESS_FAX);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PRIMARY_FAX_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 31;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_CALLBACK);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_CALLBACK_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 32;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_CAR);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_CAR_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 33;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_COMPANY);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_COMPANY_MAIN_PHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 34;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_HOME);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_HOME_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 35;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_HOME_2);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_HOME2_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 36;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_HOME_FAX);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_HOME_FAX_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 37;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_ISDN);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_ISDN_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 38;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_MOBILE);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_MOBILE_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 39;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_OTHER);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_OTHER_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 40;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_PAGER);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PAGER_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 41;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_PRIMARY);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PRIMARY_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 42;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_RADIO);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_RADIO_TELEPHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 43;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_TELEX);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_TELEX_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 44;
	inp = e_contact_get_const(contact, E_CONTACT_PHONE_TTYTDD);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_TTYTDD_PHONE_NUMBER;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 45;
	inp = e_contact_get_const(contact, E_CONTACT_ORG);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_COMPANY_NAME;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 46;
	inp = e_contact_get_const(contact, E_CONTACT_ORG_UNIT);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_DEPARTMENT_NAME;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 47;
	inp = e_contact_get_const(contact, E_CONTACT_OFFICE);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_OFFICE_LOCATION;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 48;
	inp = e_contact_get_const(contact, E_CONTACT_TITLE);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_TITLE;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 49;
	inp = e_contact_get_const(contact, E_CONTACT_MANAGER);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_MANAGER_NAME;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 50;
	inp = e_contact_get_const(contact, E_CONTACT_ASSISTANT);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_ASSISTANT;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 51;
	inp = e_contact_get_const(contact, E_CONTACT_HOMEPAGE_URL);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PERSONAL_HOME_PAGE;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 52;
	inp = e_contact_get_const(contact, E_CONTACT_SPOUSE);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_SPOUSE_NAME;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 53;
	inp = e_contact_get_const(contact, E_CONTACT_NOTE);
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_COMMENT;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 54;
	inp = e_contact_get(contact, E_CONTACT_IS_LIST) ? "IPM.DistList" : "IPM.Contact";
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_MESSAGE_CLASS;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");

	i = 55;
	propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_DISPLAY_TYPE;
	propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_LONG;
	propval_list->_buffer[i].Value._u.l = e_contact_get(contact, E_CONTACT_IS_LIST) ? BRUTUS_BRUTUS_DT_PRIVATE_DISTLIST : BRUTUS_BRUTUS_DT_MAILUSER;

	if (ec_date_birth) { // Birth date
		bgt.bt_seconds = 0;
		bgt.bt_minutes = 0;
		bgt.bt_hour = 0;
		bgt.bt_day = ec_date_birth->day;
		bgt.bt_month = ec_date_birth->month;
		bgt.bt_year = ec_date_birth->year;

		e_contact_date_free(ec_date_birth);

		i++;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_BIRTHDAY;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_SYSTIME;
		brutus_bgt_to_filetime(&bgt, &(propval_list->_buffer[i].Value._u.ft));
	}

	if (ec_date_anniversary) { // Anniversary
		bgt.bt_seconds = 0;
		bgt.bt_minutes = 0;
		bgt.bt_hour = 0;
		bgt.bt_day = ec_date_anniversary->day;
		bgt.bt_month = ec_date_anniversary->month;
		bgt.bt_year = ec_date_anniversary->year;

		e_contact_date_free(ec_date_anniversary);

		i++;
		propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_WEDDING_ANNIVERSARY;
		propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_SYSTIME;
		brutus_bgt_to_filetime(&bgt, &(propval_list->_buffer[i].Value._u.ft));
	}


	/***************************************************************
	 *          Named Properties from PS_PROPSET_CONTACT           *
	 ***************************************************************/

	{ // Business address
		CORBA_char *loc = NULL;

		i = 0;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_STREET);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_STREET);
		named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_STREET;
		named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 1;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_CITY);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_CITY);
		named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_CITY;
		named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 2;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_STATE);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_STATE);
		named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_STATE;
		named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 3;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_POSTAL_CODE);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_POSTAL_CODE);
		named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_POSTAL_CODE;
		named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 4;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_COUNTRY);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_COUNTRY);
		named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_COUNTRY;
		named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		i = 5;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_POST_OFFICE_BOX);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_POST_OFFICE_BOX);
		named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_PROPSET_CONTACT_BUSINESS_ADDRESS_POST_OFFICE_BOX;
		named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		brutus_econtact_extract_address(E_CONTACT_ADDRESS_WORK,
						contact,
						&loc,    // office location
						&(named_contact_propval_list->_buffer[0].Value._u.lpszA),  // street,
						&(named_contact_propval_list->_buffer[1].Value._u.lpszA),  // city
						&(named_contact_propval_list->_buffer[2].Value._u.lpszA),  // state or province
						&(named_contact_propval_list->_buffer[3].Value._u.lpszA),  // postal code
						&(named_contact_propval_list->_buffer[4].Value._u.lpszA),  // country
						&(named_contact_propval_list->_buffer[5].Value._u.lpszA)); // PO box
		if (loc)
			CORBA_free(loc);
	}

	i = 6;
	named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_FILE_AS);
	named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
	named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_FILE_AS);
	{
		named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_FILE_AS;
		named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		inp = e_contact_get_const(contact, E_CONTACT_FILE_AS);
		named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");
	}

	{ // EMail 1
		i = 7;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_1_DISPLAY_AS);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_1_DISPLAY_AS);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_1_DISPLAY_AS;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

			inp = e_contact_get_const(contact, E_CONTACT_EMAIL_1);
			tmp = g_strdup_printf("<%s %s> %s", given_name, last_name, inp);
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup((const char*)(tmp ? tmp : ""));
			if (tmp)
				g_free(tmp);
		}

		i = 8;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_1_ADDRESS_TYPE);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_1_ADDRESS_TYPE);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_1_ADDRESS_TYPE;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup("SMTP");
		}

		i = 9;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_1_ADDRESS);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_1_ADDRESS);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_1_ADDRESS;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

			inp = e_contact_get_const(contact, E_CONTACT_EMAIL_1);
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");
		}

		i = 10;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_1_DISPLAY_NAME);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_1_DISPLAY_NAME);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_1_DISPLAY_NAME;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

			tmp = g_strdup_printf("%s %s", given_name, last_name);
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup((const char*)(tmp ? tmp : ""));
			if (tmp)
				g_free(tmp);
		}
	}

	{ // EMail 2
		i = 11;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_2_DISPLAY_AS);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_2_DISPLAY_AS);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_2_DISPLAY_AS;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

			inp = e_contact_get_const(contact, E_CONTACT_EMAIL_2);
			tmp = g_strdup_printf("<%s %s> %s", given_name, last_name, inp);
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup((const char*)(tmp ? tmp : ""));
			if (tmp)
				g_free(tmp);
		}

		i = 12;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_2_ADDRESS_TYPE);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_2_ADDRESS_TYPE);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_2_ADDRESS_TYPE;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup("SMTP");
		}

		i = 13;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_2_ADDRESS);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_2_ADDRESS);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_2_ADDRESS;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

			inp = e_contact_get_const(contact, E_CONTACT_EMAIL_2);
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");
		}

		i = 14;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_2_DISPLAY_NAME);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_2_DISPLAY_NAME);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_2_DISPLAY_NAME;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

			tmp = g_strdup_printf("%s %s", given_name, last_name);
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup((const char*)(tmp ? tmp : ""));
			if (tmp)
				g_free(tmp);
		}
	}

	{ // EMail 3
		i = 15;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_3_DISPLAY_AS);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_3_DISPLAY_AS);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_3_DISPLAY_AS;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

			inp = e_contact_get_const(contact, E_CONTACT_EMAIL_3);
			tmp = g_strdup_printf("<%s %s> %s", given_name, last_name, inp);
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup((const char*)(tmp ? tmp : ""));
			if (tmp)
				g_free(tmp);
		}

		i = 16;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_3_ADDRESS_TYPE);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_3_ADDRESS_TYPE);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_3_ADDRESS_TYPE;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup("SMTP");
		}

		i = 17;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_3_ADDRESS);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_3_ADDRESS);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_3_ADDRESS;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

			inp = e_contact_get_const(contact, E_CONTACT_EMAIL_3);
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");
		}

		i = 18;
		named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_3_DISPLAY_NAME);
		named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_3_DISPLAY_NAME);
		{
			named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_3_DISPLAY_NAME;
			named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

			tmp = g_strdup_printf("%s %s", given_name, last_name);
			named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup((const char*)(tmp ? tmp : ""));
			if (tmp)
				g_free(tmp);
		}
	}

	i = 19;
	named_contact_prop_tags->_buffer[i].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_INTERNET_FREEBUSY_ADDRESS);
	named_contact_prop_tags->_buffer[i].name._d = BRUTUS_BRUTUS_ID_INTEGER;
	named_contact_prop_tags->_buffer[i].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_INTERNET_FREEBUSY_ADDRESS);
	{
		named_contact_propval_list->_buffer[i].ulPropTag = BRUTUS_BRUTUS_PR_INTERNET_FREEBUSY_ADDRESS;
		named_contact_propval_list->_buffer[i].Value._d = BRUTUS_BRUTUS_PT_STRING8;

		inp = e_contact_get_const(contact, E_CONTACT_FREEBUSY_URL);
		named_contact_propval_list->_buffer[i].Value._u.lpszA = CORBA_string_dup(inp ? inp : "");
	}

	/***************************************************************
	 *                   Other named properties                    *
	 ***************************************************************/

	{ // PS_PUBLIC_STRINGS
		GList *category_list = NULL;

		category_list = e_contact_get(contact, E_CONTACT_CATEGORY_LIST);
		if (!category_list || !g_list_length(category_list)) {
			CORBA_free(named_pubstr_prop_tags->_buffer);
			named_pubstr_prop_tags->_length = 0;
			named_pubstr_prop_tags->_maximum = 0;
			named_pubstr_prop_tags->_buffer = NULL;

			CORBA_free(named_pubstr_propval_list->_buffer);
			named_pubstr_propval_list->_length = 0;
			named_pubstr_propval_list->_maximum = 0;
			named_pubstr_propval_list->_buffer = NULL;
		} else {
			named_pubstr_prop_tags->_buffer[0].type = BRUTUS_BRUTUS_PT_MV_STRING8;
			named_pubstr_prop_tags->_buffer[0].name._d = BRUTUS_BRUTUS_ID_STRING;
			named_pubstr_prop_tags->_buffer[0].name._u.str_name = CORBA_string_dup(BRUTUS_BRUTUS_PR_MNID_STRING_KEYWORDS);

			named_pubstr_propval_list->_buffer[0].ulPropTag = BRUTUS_BRUTUS_PR_POST_OFFICE_BOXES;
			named_pubstr_propval_list->_buffer[0].Value._d = BRUTUS_BRUTUS_PT_MV_STRING8;
			named_pubstr_propval_list->_buffer[0].Value._u.MVszA._maximum = g_list_length(category_list);
			named_pubstr_propval_list->_buffer[0].Value._u.MVszA._length = named_pubstr_propval_list->_buffer[0].Value._u.MVszA._maximum;
			named_pubstr_propval_list->_buffer[0].Value._u.MVszA._buffer = BRUTUS_SLPSTRArray_allocbuf(named_pubstr_propval_list->_buffer[0].Value._u.MVszA._maximum);

			for (i = 0; i < g_list_length(category_list); i++) {
				inp = g_list_nth_data(category_list, i);
				named_pubstr_propval_list->_buffer[0].Value._u.MVszA._buffer[i] = CORBA_string_dup(inp ? inp : "");
			}

			g_list_foreach(category_list, (GFunc)g_free, NULL);
			g_list_free(category_list);
		}
	}

	{ // PS_PROPSET_COMMON
		named_common_prop_tags->_buffer[0].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_HIDE_ATTACHMENTS);
		named_common_prop_tags->_buffer[0].name._d = BRUTUS_BRUTUS_ID_INTEGER;
		named_common_prop_tags->_buffer[0].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_HIDE_ATTACHMENTS);
		named_common_propval_list->_buffer[0].ulPropTag = BRUTUS_BRUTUS_PR_HIDE_ATTACHMENTS;
		named_common_propval_list->_buffer[0].Value._d = BRUTUS_BRUTUS_PT_BOOLEAN;
		named_common_propval_list->_buffer[0].Value._u.b = CORBA_TRUE;
	}

	retv = TRUE;

out:
	if (contact_name)
		e_contact_name_free(contact_name);

	return retv;
}

/*
 * Caller *must* call SaveChanges() on mapi_prop
 */
static void
brutus_econtact_to_GAL_imailuser(EContact *contact,
				 BRUTUS_IMAPIProp mapi_prop)
{
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_SPropProblemArray *mapi_problems = NULL;
	BRUTUS_seq_SPropValue propval_list;
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	// get all property values
	brutus_econtact_to_GAL_proplist(contact,
					&propval_list);

	// allocate pseudo SPropProblemArray
	mapi_problems = BRUTUS_seq_SPropProblem__alloc();
	mapi_problems->_maximum = 0;
	mapi_problems->_length = 0;
	mapi_problems->_buffer = NULL;

	// set standard properties
	br = BRUTUS_IMAPIProp_SetProps(mapi_prop,
				       &propval_list,
				       FALSE,
				       &mapi_problems,
				       ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}
out:
	CORBA_free(mapi_problems);
	CORBA_exception_free(ev);

	brutus_free_seq_spropvalue_content(&propval_list, CORBA_OBJECT_NIL);
}

/*
 * Caller *must* call SaveChanges() on mapi_message
 */
static void
brutus_econtact_to_PAB_imessage(EContact *contact,
				BRUTUS_IMessage mapi_message)
{
	EContactPhoto *photo = NULL;
	BRUTUS_SPropProblemArray *mapi_problems = NULL;
	BRUTUS_seq_SPropValue propval_list;
	BRUTUS_seq_SPropValue photo_propval_list;
	BRUTUS_seq_SPropValue named_pubstr_propval_list;
	BRUTUS_seq_NamedPropertyTag named_pubstr_prop_tags;
	BRUTUS_seq_SPropValue named_contact_propval_list;
	BRUTUS_seq_NamedPropertyTag named_contact_prop_tags;
	BRUTUS_seq_SPropValue named_common_propval_list;
	BRUTUS_seq_NamedPropertyTag named_common_prop_tags;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	// contact photo properties
	photo_propval_list._maximum = 11;
	photo_propval_list._length = photo_propval_list._maximum;
	photo_propval_list._buffer = BRUTUS_seq_SPropValue_allocbuf(photo_propval_list._maximum);
	if (!photo_propval_list._buffer)
		goto out;
	CORBA_sequence_set_release(&photo_propval_list, CORBA_FALSE);

	photo_propval_list._buffer[0].ulPropTag = BRUTUS_BRUTUS_PR_ACCESS_LEVEL;
	photo_propval_list._buffer[0].Value._d = BRUTUS_BRUTUS_PT_LONG;
	photo_propval_list._buffer[0].Value._u.l = 0;

	photo_propval_list._buffer[1].ulPropTag = BRUTUS_BRUTUS_PR_ATTACHMENT_CONTACTPHOTO;
	photo_propval_list._buffer[1].Value._d = BRUTUS_BRUTUS_PT_BOOLEAN;
	photo_propval_list._buffer[1].Value._u.b = CORBA_TRUE;

	photo_propval_list._buffer[2].ulPropTag = BRUTUS_BRUTUS_PR_ATTACHMENT_FLAGS;
	photo_propval_list._buffer[2].Value._d = BRUTUS_BRUTUS_PT_LONG;
	photo_propval_list._buffer[2].Value._u.l = 0;

	photo_propval_list._buffer[3].ulPropTag = BRUTUS_BRUTUS_PR_ATTACHMENT_HIDDEN;
	photo_propval_list._buffer[3].Value._d = BRUTUS_BRUTUS_PT_BOOLEAN;
	photo_propval_list._buffer[3].Value._u.b = CORBA_FALSE;

	photo_propval_list._buffer[4].ulPropTag = BRUTUS_BRUTUS_PR_ATTACHMENT_LINKID;
	photo_propval_list._buffer[4].Value._d = BRUTUS_BRUTUS_PT_LONG;
	photo_propval_list._buffer[4].Value._u.l = 0;

	photo_propval_list._buffer[5].ulPropTag = BRUTUS_BRUTUS_PR_DISPLAY_NAME;
	photo_propval_list._buffer[5].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	photo_propval_list._buffer[5].Value._u.lpszA = "ContactPicture.jpg";

	photo_propval_list._buffer[6].ulPropTag = BRUTUS_BRUTUS_PR_ATTACH_FILENAME;
	photo_propval_list._buffer[6].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	photo_propval_list._buffer[6].Value._u.lpszA = "ContactPicture.jpg";

	photo_propval_list._buffer[7].ulPropTag = BRUTUS_BRUTUS_PR_ATTACH_LONG_FILENAME;
	photo_propval_list._buffer[7].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	photo_propval_list._buffer[7].Value._u.lpszA = "ContactPicture.jpg";

	photo_propval_list._buffer[8].ulPropTag = BRUTUS_BRUTUS_PR_URL_COMP_NAME;
	photo_propval_list._buffer[8].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	photo_propval_list._buffer[8].Value._u.lpszA = "ContactPicture.jpg";

	photo_propval_list._buffer[9].ulPropTag = BRUTUS_BRUTUS_PR_ATTACH_EXTENSION;
	photo_propval_list._buffer[9].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	photo_propval_list._buffer[9].Value._u.lpszA = ".jpg";

	photo_propval_list._buffer[10].ulPropTag = BRUTUS_BRUTUS_PR_ATTACH_METHOD;
	photo_propval_list._buffer[10].Value._d = BRUTUS_BRUTUS_PT_LONG;
	photo_propval_list._buffer[10].Value._u.l = BRUTUS_BRUTUS_ATTACH_BY_VALUE;

	// get all standard property values
	if (!brutus_econtact_to_PAB_proplist(contact,
					     &propval_list,
					     &named_pubstr_propval_list,
					     &named_pubstr_prop_tags,
					     &named_contact_propval_list,
					     &named_contact_prop_tags,
					     &named_common_propval_list,
					     &named_common_prop_tags))
		goto out;

	// allocate placeholder SPropProblemArray
	mapi_problems = BRUTUS_seq_SPropProblem__alloc();
	if (!mapi_problems)
		goto out;
	mapi_problems->_maximum = 0;
	mapi_problems->_length = 0;
	mapi_problems->_buffer = NULL;

	// set standard properties
	br = BRUTUS_IMessage_SetProps(mapi_message,
				      &propval_list,
				      FALSE,
				      &mapi_problems,
				      ev);
	brutus_free_seq_spropvalue_content(&propval_list, CORBA_OBJECT_NIL);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}

	if (named_pubstr_prop_tags._length) {// set PS_PUBLIC_STRINGS named properties if any is present
		br = BRUTUS_IMessage_SetNamedProps(mapi_message,
						   BRUTUS_BRUTUS_PS_PUBLIC_STRINGS,
						   &named_pubstr_prop_tags,
						   &named_pubstr_propval_list,
						   CORBA_FALSE,
						   &mapi_problems,
						   ev);
		brutus_free_seq_spropvalue_content(&named_pubstr_propval_list, CORBA_OBJECT_NIL);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			goto out;
		}
	}

	// set PS_PROPSET_CONTACT named properties
	br = BRUTUS_IMessage_SetNamedProps(mapi_message,
					   BRUTUS_BRUTUS_PS_PROPSET_CONTACT,
					   &named_contact_prop_tags,
					   &named_contact_propval_list,
					   CORBA_FALSE,
					   &mapi_problems,
					   ev);
	brutus_free_seq_spropvalue_content(&named_contact_propval_list, CORBA_OBJECT_NIL);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}

	// set PS_PROPSET_COMMON named properties
	br = BRUTUS_IMessage_SetNamedProps(mapi_message,
					   BRUTUS_BRUTUS_PS_PROPSET_COMMON,
					   &named_common_prop_tags,
					   &named_common_propval_list,
					   CORBA_FALSE,
					   &mapi_problems,
					   ev);
	brutus_free_seq_spropvalue_content(&named_common_propval_list, CORBA_OBJECT_NIL);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}

	// set contact photo
	photo = e_contact_get(contact, E_CONTACT_PHOTO);
	if (!photo)
		goto out;
	if (!photo->data.inlined.length)
		goto out;
	if (!photo->data.inlined.data)
		goto out;
	if (E_CONTACT_PHOTO_TYPE_URI == photo->type)
		goto out;

	d(photo->data.inlined.mime_type);
	dfmt("photo->data.inlined.length = %d", photo->data.inlined.length);
	// Outlook needs JPEG
	if (strcmp(photo->data.inlined.mime_type, "image/jpeg")) {
		size_t jpeg_len = 0;
		void *jpeg_buf = NULL;

		jpeg_buf = brutus_any_graphics_to_jpeg(photo->data.inlined.length,
						       photo->data.inlined.data,
						       &jpeg_len);
		e_contact_photo_free(photo);
		if (!jpeg_buf)
			goto out;

		dfmt("jpeg_len = %d", jpeg_len);
		brutus_add_attachment(mapi_message,
				      TRUE,
				      BRUTUS_BRUTUS_OID_NONE,
				      &photo_propval_list,
				      jpeg_len,
				      jpeg_buf);
		free(jpeg_buf);
	} else {
		brutus_add_attachment(mapi_message,
				      TRUE,
				      BRUTUS_BRUTUS_OID_NONE,
				      &photo_propval_list,
				      photo->data.inlined.length,
				      photo->data.inlined.data);
		e_contact_photo_free(photo);
	}

out:
	if (photo_propval_list._buffer)
		CORBA_free(photo_propval_list._buffer);

	if (named_pubstr_prop_tags._buffer)
		CORBA_free(named_pubstr_prop_tags._buffer);

	if (named_contact_prop_tags._buffer)
		CORBA_free(named_contact_prop_tags._buffer);

	if (named_common_prop_tags._buffer)
		CORBA_free(named_common_prop_tags._buffer);

	if (mapi_problems)
		CORBA_free(mapi_problems);

	CORBA_exception_free(ev);

	brutus_free_seq_spropvalue_content(&propval_list, CORBA_OBJECT_NIL);
	brutus_free_seq_spropvalue_content(&named_contact_propval_list, CORBA_OBJECT_NIL);
	brutus_free_seq_spropvalue_content(&named_common_propval_list, CORBA_OBJECT_NIL);
}

static void
brutus_GAL_imailuser_to_econtact(BRUTUS_IMailUser mapi_mailuser,
				 EContact *contact)
{
	CORBA_long n;
	EContactName *contact_name = NULL;
	EContactAddress *contact_addr = NULL;
	char *uid = NULL;
	BRUTUS_ENTRYID *eid = NULL;
	BRUTUS_seq_SPropValue *propval_list = NULL;
	BRUTUS_BRESULT br;
	CORBA_Environment ev[1];
	static BRUTUS_BDEFINE msg_props[22] = {
		/* 0 */  BRUTUS_BRUTUS_PR_POST_OFFICE_BOXES,         // E_CONTACT_ADDRESS_HOME(po)
		/* 1 */  BRUTUS_BRUTUS_PR_OFFICE_LOCATION,           // E_CONTACT_ADDRESS_HOME(ext) E_CONTACT_OFFICE
		/* 2 */  BRUTUS_BRUTUS_PR_STREET_ADDRESS,            // E_CONTACT_ADDRESS_HOME(street)
		/* 3 */  BRUTUS_BRUTUS_PR_LOCALITY,                  // E_CONTACT_ADDRESS_HOME(locality)
		/* 4 */  BRUTUS_BRUTUS_PR_STATE_OR_PROVINCE,         // E_CONTACT_ADDRESS_HOME(region)
		/* 5 */  BRUTUS_BRUTUS_PR_POSTAL_CODE,               // E_CONTACT_ADDRESS_HOME(code)
		/* 6 */  BRUTUS_BRUTUS_PR_COUNTRY,                   // E_CONTACT_ADDRESS_HOME(country)
		/* 7 */  BRUTUS_BRUTUS_PR_SURNAME,                   // E_CONTACT_FULL_NAME(family suffixes)
		/* 8 */  BRUTUS_BRUTUS_PR_GIVEN_NAME,                // E_CONTACT_FULL_NAME(given additional)
		/* 9  */ BRUTUS_BRUTUS_PR_TITLE,                     // E_CONTACT_FULL_NAME(prefixes) E_CONTACT_TITLE
		/* 10 */ BRUTUS_BRUTUS_PR_DISPLAY_NAME,              // E_CONTACT_FILE_AS
		/* 11 */ BRUTUS_BRUTUS_PR_SMTP_ADDRESS,              // E_CONTACT_EMAIL_1
		/* 12 */ BRUTUS_BRUTUS_PR_BUSINESS_TELEPHONE_NUMBER, // E_CONTACT_PHONE_BUSINESS
		/* 13 */ BRUTUS_BRUTUS_PR_PAGER_TELEPHONE_NUMBER,    // E_CONTACT_PHONE_PAGER
		/* 14 */ BRUTUS_BRUTUS_PR_HOME_TELEPHONE_NUMBER,     // E_CONTACT_PHONE_HOME
		/* 15 */ BRUTUS_BRUTUS_PR_MOBILE_TELEPHONE_NUMBER,   // E_CONTACT_PHONE_MOBILE
		/* 16 */ BRUTUS_BRUTUS_PR_PRIMARY_FAX_NUMBER,        // E_CONTACT_PHONE_BUSINESS_FAX
		/* 17 */ BRUTUS_BRUTUS_PR_COMPANY_NAME,              // E_CONTACT_ORG
		/* 18 */ BRUTUS_BRUTUS_PR_DEPARTMENT_NAME,           // E_CONTACT_ORG_UNIT
		/* 19 */ BRUTUS_BRUTUS_PR_EMS_AB_DESCRIPTION,        // E_CONTACT_ROLE
		/* 20 */ BRUTUS_BRUTUS_PR_COMMENT,                   // E_CONTACT_NOTE
		/* 21 */ BRUTUS_BRUTUS_PR_ENTRYID                    // E_CONTACT_UID
	};
	static BRUTUS_SPropTagArray msg_proptag_array = {
		._maximum = 22,
		._length = 22,
		._buffer = msg_props,
	};

	CORBA_exception_init(ev);

	br = BRUTUS_IMessage_GetProps(mapi_mailuser,
				      &msg_proptag_array,
				      0,
				      &propval_list,
				      ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if ((BRUTUS_BRUTUS_MAPI_W_ERRORS_RETURNED != br) && (BRUTUS_BRUTUS_S_OK != br)) {
		d(brutus_bresult_to_str(br));
		goto out;
	}

	contact_addr = g_new0(EContactAddress, 1);
	if (contact_addr) { // Address
		contact_addr->address_format = g_strdup("");

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_POST_OFFICE_BOXES);
		if (-1 != n) {
			do {
				if (!propval_list->_buffer[n].Value._u.MVszA._length)
					break;
				contact_addr->po = g_strdup(propval_list->_buffer[n].Value._u.MVszA._buffer[0]);
			} while (FALSE);
		} else
			contact_addr->po = g_strdup("");

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_OFFICE_LOCATION);
		if (-1 != n)
			contact_addr->ext = g_strdup(propval_list->_buffer[n].Value._u.lpszA);
		else
			contact_addr->ext = g_strdup("");

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_STREET_ADDRESS);
		if (-1 != n)
			contact_addr->street = g_strdup(propval_list->_buffer[n].Value._u.lpszA);
		else
			contact_addr->street = g_strdup("");

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_LOCALITY);
		if (-1 != n)
			contact_addr->locality = g_strdup(propval_list->_buffer[n].Value._u.lpszA);
		else
			contact_addr->locality = g_strdup("");

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_STATE_OR_PROVINCE);
		if (-1 != n)
			contact_addr->region = g_strdup(propval_list->_buffer[n].Value._u.lpszA);
		else
			contact_addr->region = g_strdup("");

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_POSTAL_CODE);
		if (-1 != n)
			contact_addr->code = g_strdup(propval_list->_buffer[n].Value._u.lpszA);
		else
			contact_addr->code = g_strdup("");

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_COUNTRY);
		if (-1 != n)
			contact_addr->country = g_strdup(propval_list->_buffer[n].Value._u.lpszA);
		else
			contact_addr->country = g_strdup("");

		e_contact_set(contact, E_CONTACT_ADDRESS_HOME, contact_addr);
	}

	contact_name = e_contact_name_new();
	if (contact_name) { // Name		
		const char *family = NULL;
		const char *given = NULL;
		const char *additional = NULL;
		const char *suffixes = NULL;
		char *full_name = NULL;

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_SURNAME);
		if (-1 != n) {
			family = propval_list->_buffer[n].Value._u.lpszA;
			suffixes = brutus_next_word(propval_list->_buffer[n].Value._u.lpszA);
		}
		if (!suffixes)
			suffixes = "";
		contact_name->family = g_strdup(family);
		contact_name->suffixes = g_strdup(suffixes);

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_GIVEN_NAME);
		if (-1 != n) {
			given = propval_list->_buffer[n].Value._u.lpszA;
			additional = brutus_next_word(propval_list->_buffer[n].Value._u.lpszA);
		}
		if (!additional)
			additional = "";
		contact_name->given = g_strdup(given);
		contact_name->additional = g_strdup(additional);

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_TITLE);
		if (-1 != n) {
			contact_name->prefixes = g_strdup(propval_list->_buffer[n].Value._u.lpszA);
			e_contact_set(contact, E_CONTACT_TITLE, propval_list->_buffer[n].Value._u.lpszA);
		} else
			contact_name->prefixes = g_strdup("");

		full_name = e_contact_name_to_string(contact_name);
		e_contact_set(contact, E_CONTACT_FULL_NAME, full_name);
		g_free(full_name);
	}

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_DISPLAY_NAME);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_FILE_AS, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_SMTP_ADDRESS);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_EMAIL_1, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_BUSINESS_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_BUSINESS, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_PAGER_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_PAGER, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_HOME_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_HOME, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_MOBILE_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_MOBILE, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_PRIMARY_FAX_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_BUSINESS_FAX, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_COMPANY_NAME);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_ORG, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_DEPARTMENT_NAME);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_ORG_UNIT, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_OFFICE_LOCATION);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_OFFICE, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_EMS_AB_DESCRIPTION);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_ROLE, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_COMMENT);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_NOTE, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_ENTRYID);
	if (-1 == n)
		g_assert_not_reached();
	else {
		brutus_sbinary_to_entryid(&propval_list->_buffer[n].Value._u.bin, &eid);
		uid = brutus_entryid_to_string(eid);
		if (uid) {
			e_contact_set(contact, E_CONTACT_UID, uid);
			free(uid);
		}
	}

out:
	CORBA_exception_free(ev);

	if (propval_list)
		CORBA_free(propval_list);

	if (contact_name)
		e_contact_name_free(contact_name);

	if (contact_addr)
		e_contact_address_free(contact_addr);
}

/*
 * Will scan for and append the contact photo to the EContact
 */
static void
brutus_contact_photo_from_imessage(BRUTUS_IMessage mapi_message,
				   EContact *contact)
{
	BRUTUS_IMAPITable attachment_table = CORBA_OBJECT_NIL;
	BRUTUS_IAttach mapi_attach = CORBA_OBJECT_NIL;
	BRUTUS_seq_octet *attach_data = NULL;
	BRUTUS_BRESULT br;
	CORBA_long attach_num;
	CORBA_long n;
	CORBA_long r;
	CORBA_Environment ev[1];
	BRUTUS_SRowSet *row_set = NULL;            // for the attachments table
	static BRUTUS_SSortOrderSet sort_order = { // for the attachments table
		.cCategories = 0,
		.cExpanded = 0,
		.aSort._maximum = 0,
		.aSort._length = 0,
		.aSort._buffer = NULL,
	};
	EContactPhoto photo = {
		.data.inlined.mime_type = "image/jpeg",
		.type = E_CONTACT_PHOTO_TYPE_INLINED,
		.data.inlined.length = 0,
		.data.inlined.data = NULL,
	};
	static BRUTUS_BDEFINE attach_props[2] = {                  // for the attachment table
		/* 0 */ BRUTUS_BRUTUS_PR_ATTACH_NUM,
		/* 1 */ BRUTUS_BRUTUS_PR_ATTACHMENT_CONTACTPHOTO,  // is this it?
	};
	static BRUTUS_SPropTagArray attach_proptag_array = {
		._maximum = 2,
		._length = 2,
		._buffer = attach_props,
	};

	CORBA_exception_init(ev);

	br = BRUTUS_IMessage_GetAttachmentTable(mapi_message,
						0,
						&attachment_table,
						ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		d("Could not query message object for attachment table");
		goto out;
	}

	br = brutus_QueryAllRows(attachment_table,
				 &attach_proptag_array,
				 CORBA_OBJECT_NIL,
				 &sort_order,
				 0,
				 &row_set,
				 ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}
	if (!row_set->_length)
		goto out;

	// iterate over content (columns)
	// See http://windowssdk.msdn.microsoft.com/library/default.asp?url=/library/en-us/MAPI/html/_mapi1book_opening_an_attachment.asp
	for (r = 0; r < row_set->_length; r++) {
		n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_ATTACHMENT_CONTACTPHOTO);
		if (-1 == n) {
			d("No PR_ATTACHMENT_CONTACTPHOTO");
			continue;
		}
		if (CORBA_FALSE == row_set->_buffer[r]._buffer[n].Value._u.b) // not contact photo?
			continue;

		n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_ATTACH_NUM);
		if (-1 == n)
			break;
		attach_num = row_set->_buffer[r]._buffer[n].Value._u.l;

		br = BRUTUS_IMessage_OpenAttach(mapi_message,
						attach_num,
						"",
						0, //  read-only
						&mapi_attach,
						ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			break;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			break;
		}

		// read attachment data from stream
		attach_data = brutus_read_binary_stream(BRUTUS_BRUTUS_PR_ATTACH_DATA_BIN, mapi_attach);
		if (!attach_data)
			break;
		photo.data.inlined.length = attach_data->_length;
		photo.data.inlined.data = attach_data->_buffer;

		e_contact_set(contact, E_CONTACT_PHOTO, &photo);
	}
out:
	brutus_release_object(BRUTUS_IAttach_tc, &mapi_attach, CORBA_OBJECT_NIL, ev);
	PRINT_IF_ORBIT2_EX(ev);

	brutus_release_object(BRUTUS_IMAPITable_tc, &attachment_table, CORBA_OBJECT_NIL, ev);
	PRINT_IF_ORBIT2_EX(ev);

	CORBA_exception_free(ev);

	CORBA_free(row_set);
	CORBA_free(attach_data);
}

static void
brutus_PAB_imessage_to_econtact(BRUTUS_IMessage mapi_message,
				EContact *contact)
{
	CORBA_long n;
	EContactName contact_name;
	EContactAddress contact_addr;
	char *uid = NULL;
	BRUTUS_ENTRYID *eid = NULL;
	CORBA_Environment ev[1];
	BRUTUS_BRESULT br;
	BRUTUS_seq_SPropValue *propval_list = NULL;
	BRUTUS_seq_SPropValue *contact_named_prop_values = NULL;
	BRUTUS_seq_SPropValue *pubstr_named_prop_values = NULL;
	static BRUTUS_BDEFINE msg_tags[57] = {
		/* 0 */  BRUTUS_BRUTUS_PR_BIRTHDAY,            // E_CONTACT_BIRTH_DATE
		/* 1 */  BRUTUS_BRUTUS_PR_WEDDING_ANNIVERSARY, // E_CONTACT_ANNIVERSARY
		/* 2 */  BRUTUS_BRUTUS_PR_DISPLAY_NAME_PREFIX, // E_CONTACT_TITLE
		/* 3 */  BRUTUS_BRUTUS_PR_NICKNAME,            // E_CONTACT_NICKNAME
		/* 4 */  BRUTUS_BRUTUS_PR_GIVEN_NAME,  // E_CONTACT_FULL_NAME(given)
		/* 5 */  BRUTUS_BRUTUS_PR_MIDDLE_NAME, // E_CONTACT_FULL_NAME(additional)
		/* 6 */  BRUTUS_BRUTUS_PR_SURNAME,     // E_CONTACT_FULL_NAME(family)
		/* 7 */  BRUTUS_BRUTUS_PR_GENERATION,  // E_CONTACT_FULL_NAME(suffixes)
		/* 8 */  BRUTUS_BRUTUS_PR_LOCATION,                       // E_CONTACT_ADDRESS_HOME(ext)
		/* 9  */ BRUTUS_BRUTUS_PR_HOME_ADDRESS_STREET,            // E_CONTACT_ADDRESS_HOME(street)
		/* 10 */ BRUTUS_BRUTUS_PR_HOME_ADDRESS_CITY,              // E_CONTACT_ADDRESS_HOME(locality)
		/* 11 */ BRUTUS_BRUTUS_PR_HOME_ADDRESS_STATE_OR_PROVINCE, // E_CONTACT_ADDRESS_HOME(region)
		/* 12 */ BRUTUS_BRUTUS_PR_HOME_ADDRESS_POSTAL_CODE,       // E_CONTACT_ADDRESS_HOME(code)
		/* 13 */ BRUTUS_BRUTUS_PR_HOME_ADDRESS_COUNTRY,           // E_CONTACT_ADDRESS_HOME(country)
		/* 14 */ BRUTUS_BRUTUS_PR_HOME_ADDRESS_POST_OFFICE_BOX,   // E_CONTACT_ADDRESS_HOME(po)
		/* 15 */ BRUTUS_BRUTUS_PR_OFFICE_LOCATION,   // E_CONTACT_ADDRESS_WORK(ext)
		/* 16 */ BRUTUS_BRUTUS_PR_STREET_ADDRESS,    // E_CONTACT_ADDRESS_WORK(street)
		/* 17 */ BRUTUS_BRUTUS_PR_LOCALITY,          // E_CONTACT_ADDRESS_WORK(locality)
		/* 18 */ BRUTUS_BRUTUS_PR_STATE_OR_PROVINCE, // E_CONTACT_ADDRESS_WORK(region)
		/* 19 */ BRUTUS_BRUTUS_PR_POSTAL_CODE,       // E_CONTACT_ADDRESS_WORK(code)
		/* 20 */ BRUTUS_BRUTUS_PR_COUNTRY,           // E_CONTACT_ADDRESS_WORK(country)
		/* 21 */ BRUTUS_BRUTUS_PR_POST_OFFICE_BOX,   // E_CONTACT_ADDRESS_WORK(po)
		/* 22 */ BRUTUS_BRUTUS_PR_OTHER_ADDRESS_STREET,            // E_CONTACT_ADDRESS_OTHER(street)
		/* 23 */ BRUTUS_BRUTUS_PR_OTHER_ADDRESS_CITY,              // E_CONTACT_ADDRESS_OTHER(locality)
		/* 24 */ BRUTUS_BRUTUS_PR_OTHER_ADDRESS_STATE_OR_PROVINCE, // E_CONTACT_ADDRESS_OTHER(region)
		/* 25 */ BRUTUS_BRUTUS_PR_OTHER_ADDRESS_POSTAL_CODE,       // E_CONTACT_ADDRESS_OTHER(code)
		/* 26 */ BRUTUS_BRUTUS_PR_OTHER_ADDRESS_COUNTRY,           // E_CONTACT_ADDRESS_OTHER(country)
		/* 27 */ BRUTUS_BRUTUS_PR_OTHER_ADDRESS_POST_OFFICE_BOX,   // E_CONTACT_ADDRESS_OTHER(po)
		/* 28 */ BRUTUS_BRUTUS_PR_ASSISTANT_TELEPHONE_NUMBER, // E_CONTACT_PHONE_ASSISTANT
		/* 29 */ BRUTUS_BRUTUS_PR_BUSINESS_TELEPHONE_NUMBER,  // E_CONTACT_PHONE_BUSINESS
		/* 30 */ BRUTUS_BRUTUS_PR_BUSINESS2_TELEPHONE_NUMBER, // E_CONTACT_PHONE_BUSINESS_2
		/* 31 */ BRUTUS_BRUTUS_PR_PRIMARY_FAX_NUMBER,         // E_CONTACT_PHONE_BUSINESS_FAX
		/* 32 */ BRUTUS_BRUTUS_PR_CALLBACK_TELEPHONE_NUMBER,  // E_CONTACT_PHONE_CALLBACK
		/* 33 */ BRUTUS_BRUTUS_PR_CAR_TELEPHONE_NUMBER,       // E_CONTACT_PHONE_CAR
		/* 34 */ BRUTUS_BRUTUS_PR_COMPANY_MAIN_PHONE_NUMBER,  // E_CONTACT_PHONE_COMPANY
		/* 35 */ BRUTUS_BRUTUS_PR_HOME_TELEPHONE_NUMBER,      // E_CONTACT_PHONE_HOME
		/* 36 */ BRUTUS_BRUTUS_PR_HOME2_TELEPHONE_NUMBER,     // E_CONTACT_PHONE_HOME_2
		/* 37 */ BRUTUS_BRUTUS_PR_HOME_FAX_NUMBER,            // E_CONTACT_PHONE_HOME_FAX
		/* 38 */ BRUTUS_BRUTUS_PR_ISDN_NUMBER,                // E_CONTACT_PHONE_ISDN
		/* 39 */ BRUTUS_BRUTUS_PR_MOBILE_TELEPHONE_NUMBER,    // E_CONTACT_PHONE_MOBILE
		/* 40 */ BRUTUS_BRUTUS_PR_OTHER_TELEPHONE_NUMBER,     // E_CONTACT_PHONE_OTHER
		/* 41 */ BRUTUS_BRUTUS_PR_PAGER_TELEPHONE_NUMBER,     // E_CONTACT_PHONE_PAGER
		/* 42 */ BRUTUS_BRUTUS_PR_PRIMARY_TELEPHONE_NUMBER,   // E_CONTACT_PHONE_PRIMARY
		/* 43 */ BRUTUS_BRUTUS_PR_RADIO_TELEPHONE_NUMBER,     // E_CONTACT_PHONE_RADIO
		/* 44 */ BRUTUS_BRUTUS_PR_TELEX_NUMBER,               // E_CONTACT_PHONE_TELEX
		/* 45 */ BRUTUS_BRUTUS_PR_TTYTDD_PHONE_NUMBER,        // E_CONTACT_PHONE_TTYTDD
		/* 46 */ BRUTUS_BRUTUS_PR_COMPANY_NAME,               // E_CONTACT_ORG
		/* 47 */ BRUTUS_BRUTUS_PR_DEPARTMENT_NAME,            // E_CONTACT_ORG_UNIT
		/* 48 */ BRUTUS_BRUTUS_PR_OFFICE_LOCATION,            // E_CONTACT_OFFICE
		/* 49 */ BRUTUS_BRUTUS_PR_TITLE,                      // E_CONTACT_TITLE
		/* 50 */ BRUTUS_BRUTUS_PR_MANAGER_NAME,               // E_CONTACT_MANAGER
		/* 51 */ BRUTUS_BRUTUS_PR_ASSISTANT,                  // E_CONTACT_ASSISTANT
		/* 52 */ BRUTUS_BRUTUS_PR_PERSONAL_HOME_PAGE,         // E_CONTACT_HOMEPAGE_URL
		/* 53 */ BRUTUS_BRUTUS_PR_SPOUSE_NAME,                // E_CONTACT_SPOUSE
		/* 54 */ BRUTUS_BRUTUS_PR_COMMENT,                    // E_CONTACT_NOTE
		/* 55 */ BRUTUS_BRUTUS_PR_ENTRYID,                    // E_CONTACT_UID
		/* 56 */ BRUTUS_BRUTUS_PR_HASATTACH                   // E_CONTACT_PHOTO?
	};
	static BRUTUS_SPropTagArray prop_tags= {
		._maximum = 57,
		._length = 57,
		._buffer = msg_tags,
	};
	BRUTUS_seq_NamedPropertyTag contact_named_prop_tags = {
		._maximum = 0,
		._length = 0,
		._buffer = NULL,
	};
	BRUTUS_seq_NamedPropertyTag pubstr_named_prop_tags = {
		._maximum = 0,
		._length = 0,
		._buffer = NULL,
	};

	CORBA_exception_init(ev);

	contact_named_prop_tags._maximum = 5;
	contact_named_prop_tags._length = contact_named_prop_tags._maximum;
	contact_named_prop_tags._buffer = BRUTUS_seq_NamedPropertyTag_allocbuf(contact_named_prop_tags._maximum);
	CORBA_sequence_set_release(&contact_named_prop_tags, CORBA_FALSE);

	contact_named_prop_tags._buffer[0].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_FILE_AS); // E_CONTACT_FILE_AS
	contact_named_prop_tags._buffer[0].name._d = BRUTUS_BRUTUS_ID_INTEGER;
	contact_named_prop_tags._buffer[0].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_FILE_AS);

	contact_named_prop_tags._buffer[1].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_1_ADDRESS); // E_CONTACT_EMAIL_1
	contact_named_prop_tags._buffer[1].name._d = BRUTUS_BRUTUS_ID_INTEGER;
	contact_named_prop_tags._buffer[1].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_1_ADDRESS);

	contact_named_prop_tags._buffer[2].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_2_ADDRESS); // E_CONTACT_EMAIL_2
	contact_named_prop_tags._buffer[2].name._d = BRUTUS_BRUTUS_ID_INTEGER;
	contact_named_prop_tags._buffer[2].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_2_ADDRESS);

	contact_named_prop_tags._buffer[3].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_EMAIL_3_ADDRESS); // E_CONTACT_EMAIL_3
	contact_named_prop_tags._buffer[3].name._d = BRUTUS_BRUTUS_ID_INTEGER;
	contact_named_prop_tags._buffer[3].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_EMAIL_3_ADDRESS);

	contact_named_prop_tags._buffer[4].type = BRUTUS_PROP_TYPE(BRUTUS_BRUTUS_PR_INTERNET_FREEBUSY_ADDRESS); // E_CONTACT_FREEBUSY_URL
	contact_named_prop_tags._buffer[4].name._d = BRUTUS_BRUTUS_ID_INTEGER;
	contact_named_prop_tags._buffer[4].name._u.int_name = BRUTUS_PROP_ID(BRUTUS_BRUTUS_PR_INTERNET_FREEBUSY_ADDRESS);

	pubstr_named_prop_tags._maximum = 1;
	pubstr_named_prop_tags._length = pubstr_named_prop_tags._maximum;
	pubstr_named_prop_tags._buffer = BRUTUS_seq_NamedPropertyTag_allocbuf(pubstr_named_prop_tags._maximum);
	CORBA_sequence_set_release(&pubstr_named_prop_tags, CORBA_FALSE);

	pubstr_named_prop_tags._buffer[0].type = BRUTUS_BRUTUS_PT_MV_STRING8; // E_CONTACT_CATEGORY_LIST
	pubstr_named_prop_tags._buffer[0].name._d = BRUTUS_BRUTUS_ID_STRING;
	pubstr_named_prop_tags._buffer[0].name._u.str_name = CORBA_string_dup(BRUTUS_BRUTUS_PR_MNID_STRING_KEYWORDS);


	/*
	 * Standard properties
	 */
	br = BRUTUS_IMessage_GetProps(mapi_message,
				      &prop_tags,
				      0,
				      &propval_list,
				      ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if ((BRUTUS_BRUTUS_MAPI_W_ERRORS_RETURNED != br) && (BRUTUS_BRUTUS_S_OK != br)) {
		d(brutus_bresult_to_str(br));
		goto out;
	}

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_BIRTHDAY);
	if (-1 != n) {// Birth date
		EContactDate ec_date;
		struct BrutusGregorianTime bgt;

		brutus_filetime_to_bgt(&(propval_list->_buffer[n].Value._u.ft), &bgt);
		ec_date.day = bgt.bt_day;
		ec_date.month = bgt.bt_month;
		ec_date.year = bgt.bt_year;

		e_contact_set(contact, E_CONTACT_BIRTH_DATE, &ec_date);
	}

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_WEDDING_ANNIVERSARY);
	if (-1 != n) {// Anniversary
		EContactDate ec_date;
		struct BrutusGregorianTime bgt;

		brutus_filetime_to_bgt(&(propval_list->_buffer[n].Value._u.ft), &bgt);
		ec_date.day = bgt.bt_day;
		ec_date.month = bgt.bt_month;
		ec_date.year = bgt.bt_year;

		e_contact_set(contact, E_CONTACT_ANNIVERSARY, &ec_date);
	}

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_DISPLAY_NAME_PREFIX);
	if (-1 != n) // Anniversary
		e_contact_set(contact, E_CONTACT_TITLE, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_NICKNAME);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_NICKNAME, propval_list->_buffer[n].Value._u.lpszA);

	{ // E_CONTACT_FULL_NAME
		char *full_name = NULL;

		contact_name.prefixes = "";
		
		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_GIVEN_NAME);
		if (-1 != n)
			contact_name.given = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_name.given = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_MIDDLE_NAME);
		if (-1 != n)
			contact_name.additional = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_name.additional = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_SURNAME);
		if (-1 != n)
			contact_name.family = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_name.family = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_GENERATION);
		if (-1 != n)
			contact_name.suffixes = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_name.suffixes = "";
		d(contact_name.suffixes);

		full_name = e_contact_name_to_string(&contact_name);
		e_contact_set(contact, E_CONTACT_FULL_NAME, full_name);
		g_free(full_name);
	}

	{ // E_CONTACT_ADDRESS_HOME
		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_LOCATION);
		if (-1 != n)
			contact_addr.ext = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.ext = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_HOME_ADDRESS_STREET);
		if (-1 != n)
			contact_addr.street = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.street = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_HOME_ADDRESS_CITY);
		if (-1 != n)
			contact_addr.locality = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.locality = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_HOME_ADDRESS_STATE_OR_PROVINCE);
		if (-1 != n)
			contact_addr.region = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.region = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_HOME_ADDRESS_POSTAL_CODE);
		if (-1 != n)
			contact_addr.code = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.code = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_HOME_ADDRESS_COUNTRY);
		if (-1 != n)
			contact_addr.country = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.country = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_HOME_ADDRESS_POST_OFFICE_BOX);
		if (-1 != n)
			contact_addr.po = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.po = "";

		e_contact_set(contact, E_CONTACT_ADDRESS_HOME, &contact_addr);
	}

	{ // E_CONTACT_ADDRESS_WORK
		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_OFFICE_LOCATION);
		if (-1 != n)
			contact_addr.ext = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.ext = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_STREET_ADDRESS);
		if (-1 != n)
			contact_addr.street = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.street = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_LOCALITY);
		if (-1 != n)
			contact_addr.locality = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.locality = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_STATE_OR_PROVINCE);
		if (-1 != n)
			contact_addr.region = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.region = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_POSTAL_CODE);
		if (-1 != n)
			contact_addr.code = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.code = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_COUNTRY);
		if (-1 != n)
			contact_addr.country = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.country = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_POST_OFFICE_BOX);
		if (-1 != n)
			contact_addr.po = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.po = "";

		e_contact_set(contact, E_CONTACT_ADDRESS_WORK, &contact_addr);
	}

	{ // E_CONTACT_ADDRESS_OTHER
		contact_addr.ext = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_OTHER_ADDRESS_STREET);
		if (-1 != n)
			contact_addr.street = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.street = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_OTHER_ADDRESS_CITY);
		if (-1 != n)
			contact_addr.locality = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.locality = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_OTHER_ADDRESS_STATE_OR_PROVINCE);
		if (-1 != n)
			contact_addr.region = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.region = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_OTHER_ADDRESS_POSTAL_CODE);
		if (-1 != n)
			contact_addr.code = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.code = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_OTHER_ADDRESS_COUNTRY);
		if (-1 != n)
			contact_addr.country = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.country = "";

		n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_OTHER_ADDRESS_POST_OFFICE_BOX);
		if (-1 != n)
			contact_addr.po = propval_list->_buffer[n].Value._u.lpszA;
		else
			contact_addr.po = "";

		e_contact_set(contact, E_CONTACT_ADDRESS_WORK, &contact_addr);
	}

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_ASSISTANT_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_ASSISTANT, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_BUSINESS_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_BUSINESS, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_BUSINESS2_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_BUSINESS_2, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_PRIMARY_FAX_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_BUSINESS_FAX, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_CALLBACK_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_CALLBACK, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_CAR_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_CAR, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_COMPANY_MAIN_PHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_COMPANY, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_HOME_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_HOME, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_HOME2_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_HOME_2, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_HOME_FAX_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_HOME_FAX, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_ISDN_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_ISDN, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_MOBILE_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_MOBILE, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_OTHER_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_OTHER, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_PAGER_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_PAGER, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_PRIMARY_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_PRIMARY, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_RADIO_TELEPHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_RADIO, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_TELEX_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_TELEX, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_TTYTDD_PHONE_NUMBER);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_PHONE_TTYTDD, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_COMPANY_NAME);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_ORG, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_DEPARTMENT_NAME);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_ORG_UNIT, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_OFFICE_LOCATION);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_OFFICE, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_TITLE);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_TITLE, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_MANAGER_NAME);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_MANAGER, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_ASSISTANT);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_ASSISTANT, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_PERSONAL_HOME_PAGE);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_HOMEPAGE_URL, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_SPOUSE_NAME);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_SPOUSE, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_COMMENT);
	if (-1 != n)
		e_contact_set(contact, E_CONTACT_NOTE, propval_list->_buffer[n].Value._u.lpszA);

	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_ENTRYID);
	if (-1 == n)
		g_assert_not_reached();
	else {
		brutus_sbinary_to_entryid(&propval_list->_buffer[n].Value._u.bin, &eid);
		uid = brutus_entryid_to_string(eid);
		if (uid) {
			e_contact_set(contact, E_CONTACT_UID, uid);
			free(uid);
		}
	}

	// Contact photo
	n = brutus_get_spropvalue_index(propval_list, BRUTUS_BRUTUS_PR_HASATTACH);
	if ((-1 != n) && propval_list->_buffer[n].Value._u.b)
		brutus_contact_photo_from_imessage(mapi_message, contact);


	/*
	 * PS_PROPSET_CONTACT
	 */
	br = BRUTUS_IMessage_GetNamedProps(mapi_message,
					   BRUTUS_BRUTUS_PS_PROPSET_CONTACT,
					   &contact_named_prop_tags,
					   0,
					   &contact_named_prop_values,
					   ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto pub_strings;
	}
	if ((BRUTUS_BRUTUS_MAPI_W_ERRORS_RETURNED != br) && (BRUTUS_BRUTUS_S_OK != br)) {
		d(brutus_bresult_to_str(br));
		goto pub_strings;
	}

	if (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(contact_named_prop_values->_buffer[0].ulPropTag)) // BRUTUS_BRUTUS_PR_FILE_AS
		e_contact_set(contact, E_CONTACT_FILE_AS, contact_named_prop_values->_buffer[0].Value._u.lpszA);

	if (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(contact_named_prop_values->_buffer[1].ulPropTag)) // BRUTUS_BRUTUS_PR_EMAIL_1_ADDRESS
		e_contact_set(contact, E_CONTACT_EMAIL_1, contact_named_prop_values->_buffer[1].Value._u.lpszA);

	if (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(contact_named_prop_values->_buffer[2].ulPropTag)) // BRUTUS_BRUTUS_PR_EMAIL_2_ADDRESS
		e_contact_set(contact, E_CONTACT_EMAIL_2, contact_named_prop_values->_buffer[2].Value._u.lpszA);

	if (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(contact_named_prop_values->_buffer[3].ulPropTag)) // BRUTUS_BRUTUS_PR_EMAIL_3_ADDRESS
		e_contact_set(contact, E_CONTACT_EMAIL_3, contact_named_prop_values->_buffer[3].Value._u.lpszA);

	if (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(contact_named_prop_values->_buffer[4].ulPropTag)) // BRUTUS_BRUTUS_PR_INTERNET_FREEBUSY_ADDRESS
		e_contact_set(contact, E_CONTACT_FREEBUSY_URL, contact_named_prop_values->_buffer[4].Value._u.lpszA);

pub_strings:

	/*
	 * PS_PUBLIC_STRINGS
	 */
	br = BRUTUS_IMessage_GetNamedProps(mapi_message,
					   BRUTUS_BRUTUS_PS_PUBLIC_STRINGS,
					   &pubstr_named_prop_tags,
					   0,
					   &pubstr_named_prop_values,
					   ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if ((BRUTUS_BRUTUS_MAPI_W_ERRORS_RETURNED != br) && (BRUTUS_BRUTUS_S_OK != br)) {
		d(brutus_bresult_to_str(br));
		goto out;
	}

	if (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(pubstr_named_prop_values->_buffer[0].ulPropTag)) { // BRUTUS_BRUTUS_PR_CATEGORY_LIST
		GList *category_names = NULL;

		for (n = 0; n < pubstr_named_prop_values->_buffer[0].Value._u.MVszA._length; n++)
			category_names = g_list_append(category_names, pubstr_named_prop_values->_buffer[0].Value._u.MVszA._buffer[n]);

		if (category_names) {
			e_contact_set(contact, E_CONTACT_CATEGORY_LIST, category_names);
			g_list_free(category_names);
		}
	}
out:
	CORBA_exception_free(ev);

	if (contact_named_prop_tags._buffer)
		CORBA_free(contact_named_prop_tags._buffer);

	if (pubstr_named_prop_tags._buffer)
		CORBA_free(pubstr_named_prop_tags._buffer);

	if (propval_list)
		CORBA_free(propval_list);

	if (contact_named_prop_values)
		CORBA_free(contact_named_prop_values);

	if (pubstr_named_prop_values)
		CORBA_free(pubstr_named_prop_values);
}

static void
brutus_imapiprop_to_econtact(const ADDRESS_BOOK_KIND kind,
			     BRUTUS_IMAPIProp mapi_prop,
			     EContact *contact)
{
	if (!contact || !mapi_prop)
		return;

	switch (kind) {
	case ADDRESS_BOOK_GAL :
		brutus_GAL_imailuser_to_econtact((BRUTUS_IMailUser)mapi_prop, contact);
		break;
	case ADDRESS_BOOK_PAB :
		brutus_PAB_imessage_to_econtact((BRUTUS_IMessage)mapi_prop, contact);
		break;
	default :
		g_assert_not_reached();
	}

	return;
}

static BRUTUS_IMAPIProp
brutus_create_GAL_entry(BRUTUS_BRESULT *bresult,
			EBookBackendBrutusPrivate *priv,
			BrutusBaseClass *brutus_base)
{
	CORBA_long n;
	BRUTUS_ENTRYID *eid = NULL;
	PortableServer_POA content_poa = CORBA_OBJECT_NIL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	gboolean use_local_proxy = !(ORBit_proto_use("IPv4") || ORBit_proto_use("IPv6"));
	CORBA_unsigned_long ref_key = 0;
	CORBA_unsigned_long row_count = 0;
	BRUTUS_SRestrictionContainer res_container = CORBA_OBJECT_NIL;
	BRUTUS_SRestrictionContainer res_container_wrapped = CORBA_OBJECT_NIL;
	BRUTUS_SRestrictionContainer res_container_active = CORBA_OBJECT_NIL;
	BRUTUS_SRestriction res;
	BRUTUS_SPropValue pv;
	BRUTUS_IMAPITable mapi_table = CORBA_OBJECT_NIL;
	BRUTUS_IMAPIProp mapi_prop = CORBA_OBJECT_NIL;
	CORBA_Environment ev[1];
	BRUTUS_SRowSet *row_set = NULL;    // for the templates table
	static BRUTUS_BDEFINE props[1] = { // for the templates table
		/* 0 */ BRUTUS_BRUTUS_PR_ENTRYID,
	};
	static BRUTUS_SPropTagArray proptag_array = {
		._maximum = 1,
		._length = 1,
		._buffer = props,
	};
	static BRUTUS_SSortOrderSet sort_order = { // for the attachments table
		.cCategories = 0,
		.cExpanded = 0,
		.aSort._maximum = 0,
		.aSort._length = 0,
		.aSort._buffer = NULL,
	};

	br = BRUTUS_IABContainer_OpenProperty(priv->GAL_book,
					      BRUTUS_BRUTUS_PR_CREATE_TEMPLATES,
					      BRUTUS_BRUTUS_IID_IMAPITable,
					      0,
					      0,
					      &mapi_table,
					      ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto err;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto err;
	}

	// we are only interested in the entry id
	br = BRUTUS_IMAPITable_SetColumns(mapi_table,
					  &proptag_array,
					  0,
					  ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto err;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto err;
	}

	content_poa = brutus_create_activated_poa(brutus_base->root_poa);
	if (CORBA_OBJECT_NIL == content_poa)
		goto err;

	// initiate SPropValue
	pv.dwAlignPad = 0;
	pv.ulPropTag = BRUTUS_BRUTUS_PR_ADDRTYPE;
	pv.Value._d = BRUTUS_BRUTUS_PT_STRING8;
	pv.Value._u.lpszA = "SMTP";

	// initiate restriction union
	res._d = BRUTUS_BRUTUS_RES_PROPERTY;

	// create SPropertyRestriction
	res._u.resProperty = impl_BRUTUS_SPropertyRestriction__create(BRUTUS_BRUTUS_RELOP_EQ,
								      BRUTUS_BRUTUS_PR_ADDRTYPE,
								      pv,
								      content_poa,
								      ev);
	// Create restriction container.
	res_container = impl_BRUTUS_SRestrictionContainer__create(&res,         // content
								  content_poa,  // content poa
								  brutus_base->root_poa, // poa for restriction container
								  ev);
	// the content poa and the restriction content will
	// be properly destroyed by the container
	CORBA_Object_release((CORBA_Object)content_poa, ev);
	CORBA_Object_release(res._u.resProperty, ev);

	if ((CORBA_OBJECT_NIL != res_container) && use_local_proxy) {
		d("Using local network capable Brutus proxy");
		ref_key = BRUTUS_BrutusProxy_wrapObject(brutus_base->brutus_proxy,
							BRUTUS_SRestrictionContainer_tc,
							res_container,
							&res_container_wrapped,
							ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto err;
		}
		d("Local network capable Brutus proxy wrapped OK");
	}
	if (ref_key)
		res_container_active = res_container_wrapped;
	else
		res_container_active = res_container;

	// restrict the table
	br = brutus_PrepareTable(mapi_table,
				 &proptag_array,
				 res_container_active,
				 &sort_order,
				 ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto err;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto err;
	}

	br = brutus_QueryRowCount(mapi_table,
				  &row_count,
				  ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto err;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto err;
	}
	if (1 != row_count) // huh?
		goto err;

	br = brutus_QuerySomeRows(mapi_table,
				  1,
				  &row_set,
				  ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto err;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto err;
	}

	n = brutus_get_srow_index(&row_set->_buffer[0], BRUTUS_BRUTUS_PR_ENTRYID);
	if (-1 == n) {
		d("No ENTRYID");
		goto err;
	}

	brutus_sbinary_to_entryid(&row_set->_buffer[0]._buffer[n].Value._u.bin, &eid);
	if (!eid)
		goto err;

	br = BRUTUS_IABContainer_CreateEntry(priv->GAL_book,
					     eid,
					     BRUTUS_BRUTUS_CREATE_CHECK_DUP_STRICT, // better to get a dup than to overwrite accidentally
					     &mapi_prop,
					     ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto err;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto err;
	}
	goto out;

err:
	if (CORBA_OBJECT_NIL != mapi_prop)
		brutus_release_object(BRUTUS_IMAPIProp_tc, &mapi_prop, CORBA_OBJECT_NIL, ev);
out:
	if (eid)
		CORBA_free(eid);

	if (row_set)
		CORBA_free(row_set);

	if (CORBA_OBJECT_NIL != mapi_table)
		brutus_release_object(BRUTUS_IMAPITable_tc, &mapi_table, CORBA_OBJECT_NIL, ev);

	if (ref_key) {
		BRUTUS_BrutusProxy_killWrapper(brutus_base->brutus_proxy,
					       ref_key,
					       ev);
		PRINT_IF_ORBIT2_EX(ev);
		CORBA_Object_release((CORBA_Object)res_container_wrapped, ev);
	}
	brutus_free_orbit_object((CORBA_Object)res_container,
				 brutus_base->root_poa,
				 ev);
	CORBA_exception_free(ev);

	*bresult = br;

	return mapi_prop;
}

static EContact*
brutus_get_contact_from_server(EBookBackendBrutus *brutus_backend,
			       const char *uid)
{
	EContact *contact = NULL;
	EBookBackendBrutusPrivate *priv = brutus_backend->priv;
	BrutusBaseClass *brutus_base = BRUTUS_BASE_CLASS(priv);
	BRUTUS_IMAPIProp mapi_mapiprop = CORBA_OBJECT_NIL;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	mapi_mapiprop = brutus_open_message(brutus_base->mapi_session,
					    CORBA_OBJECT_NIL,
					    uid);
	cb = CORBA_Object_is_nil(mapi_mapiprop, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (cb)
		goto out;

	contact = e_contact_new();
	brutus_imapiprop_to_econtact(priv->kind,
				     mapi_mapiprop,
				     contact);
	brutus_release_object(BRUTUS_IMAPIProp_tc, &mapi_mapiprop, CORBA_OBJECT_NIL, ev);
	PRINT_IF_ORBIT2_EX(ev);

	// add to cache
	e_book_backend_db_cache_add_contact(priv->cache, contact);

	// add to summary
	e_book_backend_summary_add_contact(priv->summary, contact);

out:
	CORBA_exception_free(ev);

	return contact;
}


/******************************************************************************
 *                                                                            *
 *                            Cache methods                                   *
 *                                                                            *
 *****************************************************************************/

/*
 * Retrieves the UTC time from the remote server if 'timestamp' is zero
 * and stores it in the timestamp file. Will store 'timestamp' instead 
 * if it is non-zero.
 *
 * Return TRUE if the stamp could be made, FALSE otherwise
 */
static gboolean
brutus_set_cache_timestamp(EBookBackendBrutus *brutus_backend,
			   time_t timestamp)
{
	gboolean retv = FALSE;
	time_t server_utc_time = 0;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	static char server_utc_time_str[100] = { '\0' };

	if (!brutus_backend)
		return FALSE;
	priv = brutus_backend->priv;
	if (!priv)
		return FALSE;
	if (!priv->timestamp_file_name)
		return FALSE;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	server_utc_time = timestamp ? timestamp : brutus_get_server_utc_time(brutus_base->brutus_server,
									     brutus_base->mapi_session,
									     brutus_base->exchange_server);
	if (!server_utc_time)
		return FALSE;

	// really no need to convert it to human readable, but might be useful for debugging
	brutus_time_t_to_string(server_utc_time,
				sizeof(server_utc_time_str),
				UTC_TIME_FORMAT_STRING,
				server_utc_time_str);

	while (!g_mutex_trylock(priv->timestamp_mutex))
		g_usleep(10);
	retv = g_file_set_contents(priv->timestamp_file_name,
				   server_utc_time_str,
				   -1,
				   NULL);
	g_mutex_unlock(priv->timestamp_mutex);

	return retv;
}

static gboolean
brutus_get_cache_timestamp(EBookBackendBrutus *brutus_backend,
			   time_t *timestamp)
{
	gboolean retv = FALSE;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	gchar *server_utc_time_str = NULL;

	if (!timestamp)
		return FALSE;
	if (!brutus_backend)
		return FALSE;
	priv = brutus_backend->priv;
	if (!priv)
		return FALSE;
	if (!priv->timestamp_file_name)
		return FALSE;

	brutus_base = BRUTUS_BASE_CLASS(priv);

	while (!g_mutex_trylock(priv->timestamp_mutex))
		g_usleep(10);
	retv = g_file_get_contents(priv->timestamp_file_name,
				   &server_utc_time_str,
				   NULL,
				   NULL);
	g_mutex_unlock(priv->timestamp_mutex);
	if (!retv)
		goto out;
	if (!server_utc_time_str) {
		retv = FALSE;
		goto out;
	}

	brutus_string_to_time_t((const char*)server_utc_time_str,
				UTC_TIME_FORMAT_STRING,
				timestamp);
out:
	if (server_utc_time_str)
		g_free(server_utc_time_str);

	return retv;
}

/*
 * build (NOT[EXIST:SPROP] || (EXIST:SPROP && SPROP[VALUE OPER CONST])) SRestrictionContainer
 */
static BRUTUS_SRestrictionContainer
brutus_restrict_on_modification_time(PortableServer_POA root_poa,
				     PortableServer_POA content_poa,
				     const time_t timeval)
{
	BRUTUS_SPropValue pv;

	d("");

	// SPROP[VALUE OPER CONST]
	pv.dwAlignPad = 0;
	pv.ulPropTag = BRUTUS_BRUTUS_PR_LAST_MODIFICATION_TIME; 
	pv.Value._d = BRUTUS_BRUTUS_PT_SYSTIME;
	brutus_time_t_to_filetime(timeval, &pv.Value._u.ft);

	return brutus_build_property_restriction_time(BRUTUS_BRUTUS_RELOP_GT,
						      pv,
						      root_poa,
						      content_poa);
}

/*
 * Update everything found in the cache
 */
static gboolean
brutus_update_GAL_cache(EBookBackendBrutus *brutus_backend)
{
	time_t server_time = 0;
	time_t last_cache_update = 0;
	gboolean retv = TRUE;
	char *uid = NULL;
	GPtrArray *cached_uids = NULL; // allready seen contacts
	guint r;
	EContact *contact = NULL;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;

	d("");
	if (!brutus_backend)
		return FALSE;
	priv = brutus_backend->priv;
	if (!priv)
		return FALSE;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	// get the server time now to ensure that we don't drop changes
	// happening while we update the cache. These changes will be 
	// picked up at the next cache update
	server_time = brutus_get_server_utc_time(brutus_base->brutus_server,
						 brutus_base->mapi_session,
						 brutus_base->exchange_server);
	brutus_get_cache_timestamp(brutus_backend, &last_cache_update);
	cached_uids = e_book_backend_db_cache_search(priv->cache, SHOW_ALL_SEARCH);

	for (r = 0; r < cached_uids->len; r++) {
		uid  = g_ptr_array_index(cached_uids, r);
		contact = brutus_get_contact_from_server(brutus_backend, uid);
		if (!contact)
			goto delete_from_cache;

		e_book_backend_db_cache_add_contact(priv->cache, contact);
		goto next;

	delete_from_cache:
		retv = FALSE;
		d(uid);
		if (uid) {
			e_book_backend_db_cache_remove_contact(priv->cache, uid);
			e_book_backend_summary_remove_contact(priv->summary, uid);
		}
	next:
		if (contact) {
			g_object_unref(contact);
			contact = NULL;
		}
	}

	// all (most) is well - stamp the cache
	brutus_set_cache_timestamp(brutus_backend, server_time);

	if (cached_uids && cached_uids->len) { 
		for (r = 0; r < cached_uids->len; r++) 
			g_free(g_ptr_array_index(cached_uids, r));
		g_ptr_array_free(cached_uids, TRUE);
	}

	priv->cache->sync(priv->cache, 0);

	return retv;
}

/*
 * Update everything from the server
 */
static gboolean
brutus_update_PAB_cache(EBookBackendBrutus *brutus_backend)
{
	time_t server_time = 0;
	time_t last_cache_update = 0;
	time_t modification_time = 0;
	gboolean retv = FALSE;
	gboolean do_break = FALSE;
	char *uid = NULL;
	GPtrArray *deleted_uids = NULL; // cache and summary entries to be deleted due to missing corresponding remote message object
	guint r;
	CORBA_long n;
	EContact *contact = NULL;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	gboolean use_local_proxy = !(ORBit_proto_use("IPv4") || ORBit_proto_use("IPv6"));
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_IMAPITable contents_table = CORBA_OBJECT_NIL;
	CORBA_unsigned_long ref_key = 0;
	BRUTUS_SRowSet *row_set = NULL;
	CORBA_unsigned_long row_count = 0;
	BRUTUS_ENTRYID *eid = NULL;
	PortableServer_POA content_poa = CORBA_OBJECT_NIL;
	BRUTUS_SRestrictionContainer res_container = CORBA_OBJECT_NIL;
	BRUTUS_SRestrictionContainer res_container_wrapped = CORBA_OBJECT_NIL; // the wrapped container - wrapper
	BRUTUS_SRestrictionContainer res_container_active = CORBA_OBJECT_NIL;  // used in PrepareTable()
	CORBA_Environment ev[1];
	static BRUTUS_BDEFINE content_props[2] = {
		BRUTUS_BRUTUS_PR_ENTRYID, // short-term entryid
		BRUTUS_BRUTUS_PR_LAST_MODIFICATION_TIME // if not found by a restriction
	};
	BRUTUS_SPropTagArray proptag_array = { // for the contents table
		._maximum = 2,
		._length = 2,
		._buffer = content_props,
	};
	static BRUTUS_SSortOrderSet sort_order = {
		.cCategories = 0,
		.cExpanded = 0,
		.aSort._maximum = 0,
		.aSort._length = 0,
		.aSort._buffer = NULL,
	};

	d("");
	if (!brutus_backend)
		return FALSE;
	priv = brutus_backend->priv;
	if (!priv)
		return FALSE;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	content_poa = brutus_create_activated_poa(brutus_base->root_poa);
	if (CORBA_OBJECT_NIL == content_poa)
		return FALSE;

	CORBA_exception_init(ev);

	// get the server time now to ensure that we don't drop changes
	// happening while we update the cache. These changes will be 
	// picked up at the next cache update
	server_time = brutus_get_server_utc_time(brutus_base->brutus_server,
						 brutus_base->mapi_session,
						 brutus_base->exchange_server);
	switch (priv->kind) {
	case ADDRESS_BOOK_GAL :
		br = BRUTUS_IMAPIContainer_GetContentsTable((BRUTUS_IMAPIContainer)priv->GAL_book,
							    0,
							    &contents_table,
							    ev);
		break;
	case ADDRESS_BOOK_PAB :
		br = BRUTUS_IMAPIContainer_GetContentsTable((BRUTUS_IMAPIContainer)priv->PAB_book,
							    0,
							    &contents_table,
							    ev);
		break;
	default :
		g_assert_not_reached();
	}
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		d("Possibly this error: http://support.microsoft.com/kb/176001/en-us");
		goto out;
	}

	if (brutus_get_cache_timestamp(brutus_backend, &last_cache_update) && !brutus_base->expect_other_clients)
		res_container = brutus_restrict_on_modification_time(brutus_base->root_poa,
								     content_poa,
								     last_cache_update);
	else
		res_container = CORBA_OBJECT_NIL; // get them all...


	if ((CORBA_OBJECT_NIL != res_container) && use_local_proxy) {
		d("Using local network capable Brutus proxy");
		ref_key = BRUTUS_BrutusProxy_wrapObject(brutus_base->brutus_proxy,
							BRUTUS_SRestrictionContainer_tc,
							res_container,
							&res_container_wrapped,
							ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out;
		}
		d("Local network capable Brutus proxy wrapped OK");
	}
	if (ref_key)
		res_container_active = res_container_wrapped;
	else
		res_container_active = res_container;

	// restrict the table
	br = brutus_PrepareTable(contents_table,
				 &proptag_array,
				 res_container_active,
				 &sort_order,
				 ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}

	br = brutus_QueryRowCount(contents_table,
				  &row_count,
				  ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}
	if (!row_count) { // none found
		retv = TRUE;
		goto out;
	}

	if (brutus_base->expect_other_clients) // get the long-term "to-be-deleted" UIDs of everything in the cache
		deleted_uids = e_book_backend_db_cache_search(priv->cache, SHOW_ALL_SEARCH);

	/*
	 * query all rows in chunks to get short-term ENTRYIDs - skip some if
	 * we are low on memory
	 */
	do {
		br = brutus_QuerySomeRows(contents_table,
					  (CORBA_long)BRUTUS_IMAPITABLE_ROW_COUNT_MAX,
					  &row_set,
					  ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			goto out;
		}

		for (r = 0; r < row_set->_length; r++) {
			n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_ENTRYID);
			if (-1 == n) {
				d("No ENTRYID");
				continue;
			}

			// short-term entry id
			brutus_sbinary_to_entryid(&row_set->_buffer[r]._buffer[n].Value._u.bin, &eid);
			if (!eid)
				continue;

			uid = brutus_entryid_to_string(eid);
			if (!uid)
				goto next;

			if ((CORBA_OBJECT_NIL == res_container)) { // the rows has not been restricted
				n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_LAST_MODIFICATION_TIME);
				if (-1 != n) {
					modification_time = brutus_filetime_to_time_t(row_set->_buffer[r]._buffer[n].Value._u.ft);
					if (last_cache_update > modification_time)
						goto next;
				}
			}

			contact = brutus_get_contact_from_server(brutus_backend, uid);
			if (!contact)
				goto delete_from_cache;
			e_book_backend_db_cache_add_contact(priv->cache, contact);

			if (brutus_base->expect_other_clients) { // subtract this uid from the array of cached "to-be-deleted" UIDs
				free(uid);
				uid = (char*)e_contact_get_const(contact, E_CONTACT_UID);
				brutus_g_ptr_array_find_and_delete_string(deleted_uids, uid);
				uid = NULL;
			}
			goto next;

		delete_from_cache:
			d(uid);
			if (uid) {
				e_book_backend_db_cache_remove_contact(priv->cache, uid);
				e_book_backend_summary_remove_contact(priv->summary, uid);
			}
		next:
			if (uid) {
				free(uid);
				uid = NULL;
			}
			if (eid) {
				CORBA_free(eid);
				eid = NULL;
			}
			if (contact) {
				g_object_unref(contact);
				contact = NULL;
			}
		}

		if (!row_set->_length)
			do_break = TRUE;

		CORBA_free(row_set);
		row_set = NULL;

		if (do_break)
			break;
	} while (TRUE);

	/* These were all present in the cache but could not be found 
	 * on the remote server (for some reason). We must there assume 
	 * that they were deleted behind our back by another mailbox client. */
	if (deleted_uids && deleted_uids->len) { 
		for (r = 0; r < deleted_uids->len; r++) {
			uid = g_ptr_array_index(deleted_uids, r);
			e_book_backend_db_cache_remove_contact(priv->cache, uid);
			e_book_backend_summary_remove_contact(priv->summary, uid);
		}
		uid = NULL;
	}

	// all is well - stamp the cache
	brutus_set_cache_timestamp(brutus_backend, server_time);
	retv = TRUE;

out:
	if (deleted_uids && deleted_uids->len) { 
		for (r = 0; r < deleted_uids->len; r++) 
			g_free(g_ptr_array_index(deleted_uids, r));
		g_ptr_array_free(deleted_uids, TRUE);
	}

	priv->cache->sync(priv->cache, 0);

	if (ref_key) {
		BRUTUS_BrutusProxy_killWrapper(brutus_base->brutus_proxy,
					       ref_key,
					       ev);
		PRINT_IF_ORBIT2_EX(ev);
		CORBA_Object_release((CORBA_Object)res_container_wrapped, ev);
	}
	brutus_free_orbit_object((CORBA_Object)res_container,
				 brutus_base->root_poa,
				 ev);

	CORBA_Object_release((CORBA_Object)content_poa, ev);

	brutus_release_object(BRUTUS_IMAPITable_tc, &contents_table, CORBA_OBJECT_NIL, ev);

	if (row_set)
		CORBA_free(row_set);

	CORBA_exception_free(ev);

	return retv;
}

static void
brutus_build_summary(EBookBackendBrutus *brutus_backend)
{
	EBookBackendBrutusPrivate *priv = brutus_backend->priv;
	gchar *query_string = NULL;
	GList *contact_list = NULL;
	GList *temp_list = NULL;

	query_string = g_strdup_printf("(or (beginswith \"file_as\" \"\") "
				       "    (beginswith \"full_name\" \"\") "
				       "    (beginswith \"email\" \"\") "
				       "    (beginswith \"nickname\" \"\"))");
	contact_list = e_book_backend_db_cache_get_contacts(priv->cache, query_string);
	g_free(query_string);

	temp_list = contact_list;
	for (; contact_list != NULL; contact_list = g_list_next(contact_list)) {
		e_book_backend_summary_add_contact(priv->summary, contact_list->data);
		g_object_unref(contact_list->data);
	}
	if (temp_list)
		g_list_free(temp_list);
	priv->summary_ready = TRUE;
}

static gboolean
brutus_poll_server(gpointer data)
{
	gboolean poll_result = FALSE;
	EBookBackendBrutus *brutus_backend = NULL;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;

	d("");
	if (!data)
		return FALSE; // stop polling
	brutus_backend = (EBookBackendBrutus*)data;

	priv = brutus_backend->priv;
	if (!priv)
		return FALSE; // stop polling
	brutus_base = BRUTUS_BASE_CLASS(priv);

	// do not execute more than one instance at any one time
	if (!g_mutex_trylock(priv->poll_mutex))
		return TRUE; // skip this poll - a poll is in progress or explicitly blocked

	if (!brutus_is_likely_connected(brutus_base)) {
		struct BrutusConnectData *data = g_new0(struct BrutusConnectData, 1);

		data->brutus_backend = brutus_backend;
		data->source = e_book_backend_get_source(E_BOOK_BACKEND(brutus_backend));
		g_object_ref(G_OBJECT(data->source));
		g_timeout_add(1, (GSourceFunc)brutus_connect, (gpointer)data);
		goto out; // skip this poll - hope for better luck next time
	}

	switch (priv->kind) {
	case ADDRESS_BOOK_GAL :
		poll_result = brutus_update_GAL_cache(brutus_backend);
		break;
	case ADDRESS_BOOK_PAB :
		poll_result = brutus_update_PAB_cache(brutus_backend);
		break;
	default :
		g_assert_not_reached();
	}

	if (!poll_result)
		d("poll done - some errors was incurred");

	brutus_build_summary(brutus_backend);
out:
	g_mutex_unlock(priv->poll_mutex);
	return TRUE;
}

/******************************************************************************
 *                                                                            *
 *                           Search methods                                   *
 *                                                                            *
 *****************************************************************************/

typedef enum {
	BRUTUS_STRING_COMPARE_EXIST,
	BRUTUS_STRING_COMPARE_IS,
	BRUTUS_STRING_COMPARE_BEGIN,
	BRUTUS_STRING_COMPARE_CONTAIN,
} BRUTUS_STRING_COMPARE;

struct EBookBackendBrutusSExpData {
	EBookBackendBrutus *backend;
	GPtrArray *matching_ids;      // array of stringified PR_ENTRYIDs for matching server side contacts
};

static struct EBookBackendBrutusSExpData*
brutus_filter_new(EBookBackendBrutus *brutus_backend)
{
	struct EBookBackendBrutusSExpData *retv = NULL;

	if (!brutus_backend)
		return NULL;

	retv = g_new0(struct EBookBackendBrutusSExpData, 1);
	if (!retv)
		return NULL;
	retv->backend = brutus_backend;
	retv->matching_ids = NULL;

	return retv;
}

static void
brutus_gptrarray_free_each(gpointer data,
			   gpointer user_data)
{
	if (data)
		free(data);
}

static void
brutus_filter_free(struct EBookBackendBrutusSExpData *filter)
{
	if (!filter)
		return;

	if (filter->matching_ids) {
		g_ptr_array_foreach(filter->matching_ids, brutus_gptrarray_free_each, NULL);
		g_ptr_array_free(filter->matching_ids, TRUE);
	}

	g_free(filter);
}

static BRUTUS_SRestrictionContainer
brutus_content_restriction(PortableServer_POA root_poa,
			   PortableServer_POA content_poa,
			   size_t len,
			   BRUTUS_FUZZY_TYPE fl,
			   const char *search_string,
			   BRUTUS_BDEFINE content_tags[])
{
	guint n;
	BRUTUS_SRestrictionContainer res_container = CORBA_OBJECT_NIL;
	BRUTUS_seq_SRestriction resSeqAND;
	BRUTUS_seq_SRestriction resSeqOR;
	BRUTUS_SRestriction res;
	BRUTUS_SPropValue pv;
	CORBA_Environment ev[1];

	if (CORBA_OBJECT_NIL == content_poa)
		return CORBA_OBJECT_NIL;

	CORBA_exception_init(ev);

	/*
	 * create OR'ed list of SAndRestrictions
	 */
	resSeqOR._maximum = len;
	resSeqOR._length = resSeqOR._maximum;
	resSeqOR._buffer = CORBA_sequence_BRUTUS_SRestriction_allocbuf(resSeqOR._maximum);
	CORBA_sequence_set_release(&resSeqOR, CORBA_FALSE);

	/*
	 * create AND'ed pairs of SExistRestriction and SContentRestriction kind
	 */
	resSeqAND._maximum = 2;
	resSeqAND._length = resSeqAND._maximum;
	resSeqAND._buffer = CORBA_sequence_BRUTUS_SRestriction_allocbuf(resSeqAND._maximum);
	CORBA_sequence_set_release(&resSeqAND, CORBA_FALSE);

	// search_string and tag
	pv.dwAlignPad = 0;
	pv.Value._d = BRUTUS_BRUTUS_PT_STRING8;
	pv.Value._u.lpszA = (CORBA_char*)search_string;

	for (n = 0; n < len; n++) {

		pv.ulPropTag = content_tags[n];

		// EXIST
		resSeqAND._buffer[0]._d = BRUTUS_BRUTUS_RES_EXIST;
		resSeqAND._buffer[0]._u.resExist = impl_BRUTUS_SExistRestriction__create(0,
											 content_tags[n],
											 0,
											 content_poa,
											 ev);
		// CONTENT
		resSeqAND._buffer[1]._d = BRUTUS_BRUTUS_RES_CONTENT;
		resSeqAND._buffer[1]._u.resContent = impl_BRUTUS_SContentRestriction__create(fl,
											     content_tags[n],
											     pv,
											     content_poa,
											     ev);
		// OR'ed
		resSeqOR._buffer[n]._d = BRUTUS_BRUTUS_RES_AND;
		resSeqOR._buffer[n]._u.resAnd = impl_BRUTUS_SAndRestriction__create(&resSeqAND,
										    content_poa,
										    ev);

		// the container will free the resources properly
		CORBA_Object_release(resSeqAND._buffer[0]._u.resExist, ev);
		CORBA_Object_release(resSeqAND._buffer[1]._u.resContent, ev);
	}

	// initialize SOrRestriction
	res._d = BRUTUS_BRUTUS_RES_OR;
	res._u.resOr = impl_BRUTUS_SOrRestriction__create(&resSeqOR,
							  content_poa,
							  ev);

	res_container = impl_BRUTUS_SRestrictionContainer__create(&res,
								  content_poa,
								  root_poa,
								  ev);

	// the restriction container will properly free all contained objects
	for (n = 0; n < resSeqOR._length; n++) {
		CORBA_Object_release((CORBA_Object)resSeqOR._buffer[n]._u.resAnd, ev);
	}
	CORBA_free(resSeqOR._buffer);
	CORBA_free(resSeqAND._buffer);
	CORBA_Object_release(res._u.resOr, ev);

	CORBA_exception_free(ev);

	return res_container;
}

static ESExpResult*
brutus_func_string_search(struct _ESExp *f,
			  int argc,
			  struct _ESExpResult **argv,
			  void *data,
			  BRUTUS_STRING_COMPARE fuzzy_level)
{
	struct EBookBackendBrutusSExpData *filter = (struct EBookBackendBrutusSExpData *)data;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	const BrutusEBookSearchDef *search_mapping = NULL;
	ESExpResult *rexp = NULL;
	gboolean do_break = FALSE;
	guint r;
	guint match_id;
	CORBA_long n;
	CORBA_unsigned_long row_count = 0;
	const char *search_token = NULL;
	const char *search_value = NULL;
	BRUTUS_FUZZY_TYPE fl = 0;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	gboolean use_local_proxy = !(ORBit_proto_use("IPv4") || ORBit_proto_use("IPv6"));
	CORBA_unsigned_long ref_key = 0;
	BRUTUS_SRestrictionContainer res_container = CORBA_OBJECT_NIL;         // the "real" container object - filling
	BRUTUS_SRestrictionContainer res_container_wrapped = CORBA_OBJECT_NIL; // the wrapped container - wrapper
	BRUTUS_SRestrictionContainer res_container_active = CORBA_OBJECT_NIL;  // used in PrepareTable()
	BRUTUS_SRestriction res;
	BRUTUS_SPropValue pv;
	PortableServer_POA content_poa = CORBA_OBJECT_NIL;
	BRUTUS_IMAPITable contents_table = CORBA_OBJECT_NIL;
	BRUTUS_SRowSet *row_set = NULL;
	BRUTUS_ENTRYID *eid = NULL;
	CORBA_Environment ev[1];
	static BRUTUS_BDEFINE content_props[1] = {
		BRUTUS_BRUTUS_PR_ENTRYID // short-term entryid
	};
	BRUTUS_SPropTagArray proptag_array = { // for the contents table
		._maximum = 1,
		._length = 1,
		._buffer = content_props,
	};
	static BRUTUS_SSortOrderSet sort_order = {
		.cCategories = 0,
		.cExpanded = 0,
		.aSort._maximum = 0,
		.aSort._length = 0,
		.aSort._buffer = NULL,
	};

	d("");

	CORBA_exception_init(ev);

	if ((argc != 2)
	    || (argv[0]->type != ESEXP_RES_STRING)
	    || (argv[1]->type != ESEXP_RES_STRING))
		goto err;

	if (!filter || !filter->backend)
		goto err;
	priv = filter->backend->priv;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	search_token = argv[0]->value.string; // propert[y:ies] to search _in_
	search_value = argv[1]->value.string; // string value to search _for_
	d(search_token);
	d(search_value);

	// SContentRestriction search
	switch (fuzzy_level) { // for our content restrictions
	case BRUTUS_STRING_COMPARE_EXIST :
		break;
	case BRUTUS_STRING_COMPARE_IS :
		fl = BRUTUS_BRUTUS_FL_FULLSTRING;
		break;
	case BRUTUS_STRING_COMPARE_BEGIN :
		fl = BRUTUS_BRUTUS_FL_PREFIX;
		break;
	case BRUTUS_STRING_COMPARE_CONTAIN :
		fl = BRUTUS_BRUTUS_FL_SUBSTRING;
		break;
	default:
		g_assert_not_reached();
	}
	if (BRUTUS_STRING_COMPARE_EXIST != fuzzy_level)
		fl |= BRUTUS_BRUTUS_FL_LOOSE;

	content_poa = brutus_create_activated_poa(brutus_base->root_poa);
	if (CORBA_OBJECT_NIL == content_poa)
		goto err;

	// get the correct contents table
	switch (priv->kind) {
	case ADDRESS_BOOK_GAL :
		br = BRUTUS_IMAPIContainer_GetContentsTable((BRUTUS_IMAPIContainer)priv->GAL_book,
							    0,
							    &contents_table,
							    ev);
		break;
	case ADDRESS_BOOK_PAB :
		br = BRUTUS_IMAPIContainer_GetContentsTable((BRUTUS_IMAPIContainer)priv->PAB_book,
							    0,
							    &contents_table,
							    ev);
		break;
	default :
		g_assert_not_reached();
	}
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto err;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		d("Possibly this error: http://support.microsoft.com/kb/176001/en-us");
		goto err;
	}

	// global search
	if (!strcmp(search_token, "x-evolution-any-field")) { // this gui does (contains "x-evolution-any-field" "") for the initial query and when you hit "Clear"
		if (!search_value || !strlen(search_value)) {
			switch (priv->kind) {
			case ADDRESS_BOOK_GAL : // never attempt to return the entire GAL content!
				goto err;
			case ADDRESS_BOOK_PAB : // return the entire book content - should be of limited size
				break;
			default :
				g_assert_not_reached();
			}
			goto get_rows;
		} else { // go search for some specific string in all known non-named string properties
			res_container = brutus_content_restriction(brutus_base->root_poa,
								   content_poa,
								   G_N_ELEMENTS(priv->content_tags) - 1,
								   fl,
								   search_value,
								   &(priv->content_tags[1]));
			goto get_rows;
		}
	}

	search_mapping = brutus_find_search_mapping(search_token, priv->search_def);
	if (!search_mapping) 
		goto err;

	// special case - PR_ANR search
	if (BRUTUS_BRUTUS_PR_ANR == search_mapping->search_tags[0]) {
		pv.dwAlignPad = 0;
		pv.ulPropTag = BRUTUS_BRUTUS_PR_ANR;
		pv.Value._d = BRUTUS_BRUTUS_PT_STRING8;
		pv.Value._u.lpszA = (CORBA_char*)search_value;

		// initiate restriction union
		res._d = BRUTUS_BRUTUS_RES_PROPERTY;

		// create SPropertyRestriction
		res._u.resProperty = impl_BRUTUS_SPropertyRestriction__create(BRUTUS_BRUTUS_RELOP_EQ,
									      BRUTUS_BRUTUS_PR_ANR,
									      pv,
									      content_poa,
									      ev);

		// Create restriction container.
		// It is important that the container itself is *not* created
		// under the restriction container content poa.
		res_container = impl_BRUTUS_SRestrictionContainer__create(&res,
									  content_poa,
									  brutus_base->root_poa,
									  ev);
		CORBA_Object_release(res._u.resProperty, ev);

		goto get_rows;
	}

	res_container = brutus_content_restriction(brutus_base->root_poa,
						   content_poa,
						   search_mapping->tag_count,
						   fl,
						   search_value,
						   search_mapping->search_tags);
get_rows:
	if ((CORBA_OBJECT_NIL != res_container) && use_local_proxy) {
		d("Using local network capable Brutus proxy");
		ref_key = BRUTUS_BrutusProxy_wrapObject(brutus_base->brutus_proxy,
							BRUTUS_SRestrictionContainer_tc,
							res_container,
							&res_container_wrapped,
							ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto err;
		}
		d("Local network capable Brutus proxy wrapped OK");
	}
	if (ref_key)
		res_container_active = res_container_wrapped;
	else
		res_container_active = res_container;

	// restrict the table
	br = brutus_PrepareTable(contents_table,
				 &proptag_array,
				 res_container_active,
				 &sort_order,
				 ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto err;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto err;
	}

	br = brutus_QueryRowCount(contents_table,
				  &row_count,
				  ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto err;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto err;
	}
	if (!row_count) { // none found
		rexp = e_sexp_result_new(f, ESEXP_RES_BOOL);
		rexp->value.bool = FALSE;
		goto out;
	}

	/*
	 * extract ENTRYIDs from all found contact objects and add them to the filter
	 */
	if (filter->matching_ids) {
		g_ptr_array_foreach(filter->matching_ids, brutus_gptrarray_free_each, NULL);
		g_ptr_array_free(filter->matching_ids, TRUE);
		filter->matching_ids = NULL;
	}
	filter->matching_ids = g_ptr_array_sized_new(row_count);
	if (!filter->matching_ids)
		goto err;

	/*
	 * query all rows in chunks to get short-term ENTRYIDs - skip some if
	 * we are low on memory
	 */
	match_id = 0;
	do {
		br = brutus_QuerySomeRows(contents_table,
					  (CORBA_long)BRUTUS_IMAPITABLE_ROW_COUNT_MAX,
					  &row_set,
					  ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto err;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			goto err;
		}

		for (r = 0; r < row_set->_length; r++) {
			n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_ENTRYID);
			if (-1 == n) {
				d("No ENTRYID");
				continue;
			}

			brutus_sbinary_to_entryid(&row_set->_buffer[r]._buffer[n].Value._u.bin, &eid);
			if (!eid)
				continue;

			g_ptr_array_add(filter->matching_ids, (gpointer)brutus_entryid_to_string(eid));
			CORBA_free(eid);
		}
		if (!row_set->_length)
			do_break = TRUE;

		CORBA_free(row_set);
		row_set = NULL;

		if (do_break)
			break;
	} while (TRUE);

	goto out;

err:
	rexp = e_sexp_result_new(f, ESEXP_RES_UNDEFINED);

	if (filter && filter->matching_ids) {
		g_ptr_array_foreach(filter->matching_ids, brutus_gptrarray_free_each, NULL);
		g_ptr_array_free(filter->matching_ids, TRUE);
		filter->matching_ids = NULL;
	}

out:
	if (filter && filter->matching_ids && filter->matching_ids->len) { // some found
		rexp = e_sexp_result_new(f, ESEXP_RES_BOOL);
		rexp->value.bool = TRUE;
	}

	if (ref_key) {
		BRUTUS_BrutusProxy_killWrapper(brutus_base->brutus_proxy,
					       ref_key,
					       ev);
		PRINT_IF_ORBIT2_EX(ev);
		CORBA_Object_release((CORBA_Object)res_container_wrapped, ev);
	}
	brutus_free_orbit_object((CORBA_Object)res_container,
				 brutus_base->root_poa,
				 ev);

	CORBA_Object_release((CORBA_Object)content_poa, ev);

	brutus_release_object(BRUTUS_IMAPITable_tc, &contents_table, CORBA_OBJECT_NIL, ev);

	if (row_set)
		CORBA_free(row_set);

	CORBA_exception_free(ev);

	if (!rexp)
		rexp = e_sexp_result_new(f, ESEXP_RES_UNDEFINED);

	return rexp;
}


static ESExpResult*
brutus_func_and(ESExp *f, int argc, ESExpResult **argv, void *data)
{
	return e_sexp_result_new(f, ESEXP_RES_UNDEFINED);
}

static ESExpResult*
brutus_func_or(ESExp *f, int argc, ESExpResult **argv, void *data)
{
	return e_sexp_result_new(f, ESEXP_RES_UNDEFINED);
}

static ESExpResult*
brutus_func_not(ESExp *f, int argc, ESExpResult **argv, void *data)
{
	return e_sexp_result_new(f, ESEXP_RES_UNDEFINED);
}

static ESExpResult*
brutus_func_contains(struct _ESExp *f, int argc, struct _ESExpResult **argv, void *data)
{
	return brutus_func_string_search(f, argc, argv, data, BRUTUS_STRING_COMPARE_CONTAIN);
}

static ESExpResult*
brutus_func_is(struct _ESExp *f, int argc, struct _ESExpResult **argv, void *data)
{
	return brutus_func_string_search(f, argc, argv, data, BRUTUS_STRING_COMPARE_IS);
}

static ESExpResult*
brutus_func_beginswith(struct _ESExp *f, int argc, struct _ESExpResult **argv, void *data)
{
	return brutus_func_string_search(f, argc, argv, data, BRUTUS_STRING_COMPARE_BEGIN);
}

static ESExpResult*
brutus_func_endswith(struct _ESExp *f, int argc, struct _ESExpResult **argv, void *data)
{
	/* We don't allow endswith searches */
	return e_sexp_result_new(f, ESEXP_RES_UNDEFINED);
}

static ESExpResult*
brutus_func_exists(struct _ESExp *f, int argc, struct _ESExpResult **argv, void *data)
{
	return brutus_func_string_search(f, argc, argv, data, BRUTUS_STRING_COMPARE_EXIST);
}

// 'builtin' functions
static struct {
	char *name;
	ESExpFunc *func;
} symbols[] = {
	{ "and", brutus_func_and},
	{ "or", brutus_func_or},
	{ "not", brutus_func_not},
	{ "contains", brutus_func_contains},
	{ "is", brutus_func_is},
	{ "beginswith", brutus_func_beginswith},
	{ "endswith", brutus_func_endswith},
	{ "exists", brutus_func_exists},
};

static struct EBookBackendBrutusSExpData*
e_book_backend_brutus_build_filter(EBookBackendBrutus *brutus_backend,
				   const char *query)
{
	int i;
	ESExp *sexp = NULL;
	ESExpResult *r = NULL;
	struct EBookBackendBrutusSExpData *filter = NULL;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;

	d(query);
	if (!brutus_backend)
		return NULL;

	priv = brutus_backend->priv;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	d("");
	filter = brutus_filter_new(brutus_backend);
	if (!filter)
		return NULL;

	sexp = e_sexp_new();
	for(i = 0; i < sizeof(symbols)/sizeof(symbols[0]); i++) {
		e_sexp_add_function(sexp,
				    0,
				    symbols[i].name,
				    symbols[i].func,
				    filter);
	}
	e_sexp_input_text(sexp, query, strlen(query));
	e_sexp_parse(sexp);
	r = e_sexp_eval(sexp);
	if (!r) {
		brutus_filter_free(filter);
		filter = NULL;
		goto out;
	}

	switch (r->type) {
	case ESEXP_RES_UNDEFINED : // invalid search
		brutus_filter_free(filter);
		filter = NULL;
		break;
	case ESEXP_RES_BOOL :
		if (r->value.bool) // found
			break;

		brutus_filter_free(filter);
		filter = NULL;
		break;
	default:
		g_assert_not_reached();
	}

out:
	e_sexp_result_free(sexp, r);
	e_sexp_unref(sexp);

	return filter;
}

/******************************************************************************
 *                                                                            *
 *                           Virtual methods                                  *
 *                                                                            *
 *****************************************************************************/

static void
e_book_backend_brutus_create_contact(EBookBackend *backend,
				     EDataBook *book,
				     guint32 opid,
				     const char *vcard )
{
	EContact *contact = NULL;
	EBookBackendBrutus *brutus_backend = NULL;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	char *uid = NULL; // stringified long-term entry id
	BRUTUS_ENTRYID *eid = NULL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_IMAPITable mapi_table = CORBA_OBJECT_NIL;
	BRUTUS_IMessage mapi_message = CORBA_OBJECT_NIL;
	BRUTUS_IMailUser mapi_mailuser = CORBA_OBJECT_NIL;
	BRUTUS_IMAPIProp mapi_prop_active = CORBA_OBJECT_NIL;
	CORBA_Environment ev[1];

	d("");
	brutus_backend = E_BOOK_BACKEND_BRUTUS(backend);
	priv = brutus_backend->priv;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	switch (priv->mode) {
	case GNOME_Evolution_Addressbook_MODE_LOCAL :
		e_data_book_respond_create(book, opid, GNOME_Evolution_Addressbook_RepositoryOffline, NULL);
		return;
	case  GNOME_Evolution_Addressbook_MODE_ANY :
	case  GNOME_Evolution_Addressbook_MODE_REMOTE :
		if (!brutus_is_likely_connected(brutus_base)) {
			struct BrutusConnectData *data = g_new0(struct BrutusConnectData, 1);

			data->brutus_backend = brutus_backend;
			data->source = e_book_backend_get_source(E_BOOK_BACKEND(brutus_backend));
			g_object_ref(G_OBJECT(data->source));
			g_timeout_add(1, (GSourceFunc)brutus_connect, (gpointer)data);

			e_data_book_respond_create(book, opid, GNOME_Evolution_Addressbook_OfflineUnavailable, NULL);
			return;
		}

		contact = e_contact_new_from_vcard(vcard);
		if (e_contact_get(contact, E_CONTACT_IS_LIST))
			d("TODO: *** support contact lists ***");

		CORBA_exception_init(ev);

		// check for duplicate entry
		uid = (char*)e_contact_get_const(contact, E_CONTACT_UID);
		eid = brutus_string_to_entryid(uid);
		while (eid) {
			BRUTUS_BDEFINE obj_type;
			BRUTUS_IUnknown unk = CORBA_OBJECT_NIL;

			br = BRUTUS_IMAPISession_OpenEntry(brutus_base->mapi_session,
							   eid,
							   "",
							   BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
							   &obj_type,
							   &unk,
							   ev);
			CORBA_free(eid);
			eid = NULL;
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				break;
			}
			if (BRUTUS_BRUTUS_S_OK != br) {
				d(brutus_bresult_to_str(br));
				break;
			}

			if (CORBA_OBJECT_NIL == unk)
				break;

			// OK, damn - it went well, wrap up and exit...
			brutus_release_object(BRUTUS_IUnknown_tc, &unk, CORBA_OBJECT_NIL, ev);
			CORBA_exception_free(ev);

			e_data_book_respond_create(book, opid, GNOME_Evolution_Addressbook_ContactIdAlreadyExists, contact);
			g_object_unref(contact);

			return;
		}

		switch (priv->kind) {
		case ADDRESS_BOOK_GAL :
			mapi_mailuser = brutus_create_GAL_entry(&br,
								priv,
								brutus_base);
			if (CORBA_OBJECT_NIL == mapi_mailuser)
				goto err;

			if (BRUTUS_BRUTUS_S_OK != br) {
				d(brutus_bresult_to_str(br));
				goto err;
			}

			brutus_econtact_to_GAL_imailuser(contact, mapi_mailuser);
			mapi_prop_active = (BRUTUS_IMAPIProp)mapi_mailuser;
			break;
		case ADDRESS_BOOK_PAB :
			br = BRUTUS_IMAPIFolder_CreateMessage(priv->PAB_book,
							      "",
							      0,
							      &mapi_message,
							      ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto err;
			}
			if (BRUTUS_BRUTUS_S_OK != br) {
				d(brutus_bresult_to_str(br));
				goto err;
			}
			brutus_econtact_to_PAB_imessage(contact, mapi_message);
			mapi_prop_active = (BRUTUS_IMAPIProp)mapi_message;
			break;
		default :
			g_assert_not_reached();
		}
		break;
	default:
		break;
	}

	br = BRUTUS_IMAPIProp_SaveChanges(mapi_prop_active,
					  BRUTUS_BRUTUS_KEEP_OPEN_READONLY,
					  ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto readonly;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto readonly;
	}

	// get and set the long-term entry id
	eid = brutus_get_entryid(mapi_prop_active, BRUTUS_BRUTUS_PR_ENTRYID);
	if (!eid)
		goto err;
	uid = brutus_entryid_to_string(eid);
	if (!uid)
		goto err;
	e_contact_set(contact, E_CONTACT_UID, uid);

	e_book_backend_db_cache_add_contact(priv->cache, contact);
	priv->cache->sync(priv->cache, 0);
	e_book_backend_summary_add_contact(priv->summary, contact);
	if (!priv->is_writable) {
		e_book_backend_notify_writable(backend, TRUE);
		priv->is_writable = TRUE;
	}
	goto out;

readonly:
	e_book_backend_notify_writable(backend, FALSE); // apparently we can't write to or read the book for some reason...
	priv->is_writable = FALSE;

err:
	g_object_unref(contact);
	contact = NULL;

out:
	if (eid)
		CORBA_free(eid);
	if (uid)
		free(uid);

	if (CORBA_OBJECT_NIL != mapi_table)
		brutus_release_object(BRUTUS_IMAPITable_tc, &mapi_table, CORBA_OBJECT_NIL, ev);

	if (CORBA_OBJECT_NIL != mapi_mailuser)
		brutus_release_object(BRUTUS_IMailUser_tc, &mapi_mailuser, CORBA_OBJECT_NIL, ev);

	if (CORBA_OBJECT_NIL != mapi_message)
		brutus_release_object(BRUTUS_IMessage_tc, &mapi_message, CORBA_OBJECT_NIL, ev);

	CORBA_exception_free(ev);

	if (contact) { // all is well
		e_data_book_respond_create(book, opid, GNOME_Evolution_Addressbook_Success, contact);
		g_object_unref(contact);
	} else {       // something went wrong...
		e_data_book_respond_create(book, opid, GNOME_Evolution_Addressbook_OtherError, NULL);
	}
}

static void
e_book_backend_brutus_remove_contacts(EBookBackend *backend,
				      EDataBook *book,
				      guint32 opid,
				      GList *id_list)
{
	GList *tmp = id_list;
	EBookBackendBrutus *brutus_backend;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	CORBA_Environment ev[1];
	BRUTUS_SBinaryArray gal_ids = {
		._maximum = 0,
		._length = 0,
		._buffer = NULL,
	};
	BRUTUS_seq_ENTRYID pab_ids = {
		._maximum = 0,
		._length = 0,
		._buffer = NULL,
	};

	d("");
	if (!id_list) {
		e_data_book_respond_remove_contacts(book, opid, GNOME_Evolution_Addressbook_Success, id_list);
		return;
	}

	brutus_backend = E_BOOK_BACKEND_BRUTUS(backend);
	priv = brutus_backend->priv;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	switch (priv->mode) {
	case GNOME_Evolution_Addressbook_MODE_LOCAL :
		e_data_book_respond_remove_contacts(book, opid, GNOME_Evolution_Addressbook_RepositoryOffline, NULL);
		return;
	case GNOME_Evolution_Addressbook_MODE_ANY :
	case GNOME_Evolution_Addressbook_MODE_REMOTE :
		if (!brutus_is_likely_connected(brutus_base)) {
			struct BrutusConnectData *data = g_new0(struct BrutusConnectData, 1);

			data->brutus_backend = brutus_backend;
			data->source = e_book_backend_get_source(E_BOOK_BACKEND(brutus_backend));
			g_object_ref(G_OBJECT(data->source));
			g_timeout_add(1, (GSourceFunc)brutus_connect, (gpointer)data);

			e_data_book_respond_remove_contacts(book, opid, GNOME_Evolution_Addressbook_OfflineUnavailable, NULL);
			return;
		}

		CORBA_exception_init(ev);

		switch (priv->kind) {
		case ADDRESS_BOOK_GAL :
			if (!brutus_uid_glist_to_sbinaryarray(id_list, &gal_ids))
				goto err;

			br = BRUTUS_IABContainer_DeleteEntries(priv->GAL_book,
							       &gal_ids,
							       0,
							       ev);
			break;
		case ADDRESS_BOOK_PAB :
			if (!brutus_uid_glist_to_seq_entryid(id_list, &pab_ids))
				goto err;

			br = BRUTUS_IMAPIFolder_DeleteMessages(priv->PAB_book,
							       &pab_ids,
							       CORBA_OBJECT_NIL,
							       0,
							       ev);
			break;
		default :
			g_assert_not_reached();
		}
		break;
	default :
		d("TODO: *** Investigate unknown EBook mode ***");
		e_data_book_respond_remove_contacts(book, opid, GNOME_Evolution_Addressbook_OtherError, NULL);
		return;
	}

	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto readonly;
	}
	switch (br) {
	case BRUTUS_BRUTUS_S_OK :
	case BRUTUS_BRUTUS_MAPI_E_NOT_FOUND :
	case BRUTUS_BRUTUS_MAPI_W_PARTIAL_COMPLETION : // TODO: handle partial deletes
		break;
	default :
		d(brutus_bresult_to_str(br));
		goto readonly;
	}

	for (tmp = id_list; tmp; tmp = tmp->next) { // remove from cache and summary
		e_book_backend_db_cache_remove_contact(priv->cache, (char*)tmp->data);
		e_book_backend_summary_remove_contact(priv->summary, (char*)tmp->data);
	}
	e_data_book_respond_remove_contacts(book, opid, GNOME_Evolution_Addressbook_Success, id_list);
	if (!priv->is_writable) {
		e_book_backend_notify_writable(backend, TRUE);
		priv->is_writable = TRUE;
	}

	goto out;

readonly:
	e_book_backend_notify_writable(backend, FALSE);
	priv->is_writable = FALSE;

err:
	e_data_book_respond_remove_contacts(book, opid, GNOME_Evolution_Addressbook_OtherError, NULL);

out:
	CORBA_exception_free(ev);
	if (gal_ids._maximum)
		brutus_free_sbinaryarray_content(&gal_ids);

	if (pab_ids._maximum)
		brutus_free_seq_entryid_content(&pab_ids);
}

static void
e_book_backend_brutus_modify_contact(EBookBackend *backend,
				     EDataBook    *book,
				     guint32       opid,
				     const char   *vcard)
{
	EContact *contact = NULL;
	EBookBackendBrutus *brutus_backend = NULL;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	char *uid = NULL;
	BRUTUS_BDEFINE obj_type;
	BRUTUS_ENTRYID *eid = NULL;
	BRUTUS_IUnknown unk = CORBA_OBJECT_NIL;
	BRUTUS_IMessage mapi_message = CORBA_OBJECT_NIL;
	BRUTUS_IMailUser mapi_mailuser = CORBA_OBJECT_NIL;
	BRUTUS_IMAPIProp mapi_prop_active = CORBA_OBJECT_NIL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	CORBA_Environment ev[1];

	brutus_backend = E_BOOK_BACKEND_BRUTUS(backend);
	priv = brutus_backend->priv;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	switch (priv->mode) {
	case GNOME_Evolution_Addressbook_MODE_LOCAL :
		e_data_book_respond_modify(book, opid, GNOME_Evolution_Addressbook_RepositoryOffline, NULL);
		return;
	case GNOME_Evolution_Addressbook_MODE_ANY :
	case GNOME_Evolution_Addressbook_MODE_REMOTE :
		if (!brutus_is_likely_connected(brutus_base)) {
			struct BrutusConnectData *data = g_new0(struct BrutusConnectData, 1);

			data->brutus_backend = brutus_backend;
			data->source = e_book_backend_get_source(E_BOOK_BACKEND(brutus_backend));
			g_object_ref(G_OBJECT(data->source));
			g_timeout_add(1, (GSourceFunc)brutus_connect, (gpointer)data);

			e_data_book_respond_modify(book, opid, GNOME_Evolution_Addressbook_OfflineUnavailable, NULL);
			return;
		}

		contact = e_contact_new_from_vcard(vcard);
		if (e_contact_get(contact, E_CONTACT_IS_LIST))
			d("TODO: *** support contact lists ***");

		CORBA_exception_init(ev);

		// open Exchange object
		uid = (char*)e_contact_get_const(contact, E_CONTACT_UID);
		if (!uid || !strlen(uid))
			goto err;
		eid = brutus_string_to_entryid(uid);
		if (!eid)
			goto err;
		br = BRUTUS_IMAPISession_OpenEntry(brutus_base->mapi_session,
						   eid,
						   "",
						   BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
						   &obj_type,
						   &unk,
						   ev);
		CORBA_free(eid);
		eid = NULL;
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto err;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			goto err;
		}
		switch (priv->kind) {
		case ADDRESS_BOOK_GAL :
			mapi_mailuser = (BRUTUS_IMailUser)unk;
			mapi_prop_active = (BRUTUS_IMAPIProp)mapi_mailuser;
			brutus_econtact_to_GAL_imailuser(contact, mapi_mailuser);
			break;
		case ADDRESS_BOOK_PAB :
			mapi_message = (BRUTUS_IMessage)unk;
			mapi_prop_active = (BRUTUS_IMAPIProp)mapi_message;
			brutus_econtact_to_PAB_imessage(contact, mapi_message);
			break;
		default :
			g_assert_not_reached();
		}
		break;
	default :
		d("TODO: *** Investigate unknown EBook mode ***");
		e_data_book_respond_modify(book, opid, GNOME_Evolution_Addressbook_OtherError, NULL);
		return;
	}

	br = BRUTUS_IMAPIProp_SaveChanges(mapi_prop_active,
					  BRUTUS_BRUTUS_FORCE_SAVE,
					  ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto readonly;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto readonly;
	}
	if (!priv->is_writable) {
		e_book_backend_notify_writable(backend, TRUE);
		priv->is_writable = TRUE;
	}

	goto out;

readonly:
	e_book_backend_notify_writable(backend, FALSE);
	priv->is_writable = FALSE;

err:
	g_object_unref(contact);
	contact = NULL;

out:
	if (eid)
		CORBA_free(eid);

	brutus_release_object(BRUTUS_IUnknown_tc, &unk, CORBA_OBJECT_NIL, ev);

	CORBA_exception_free(ev);

	if (contact) {
		e_data_book_respond_modify(book, opid, GNOME_Evolution_Addressbook_Success, contact);
		g_object_unref(contact);
	} else
		e_data_book_respond_modify(book, opid, GNOME_Evolution_Addressbook_OtherError, NULL);
}

static void
e_book_backend_brutus_get_contact(EBookBackend *backend,
				  EDataBook    *book,
				  guint32       opid,
				  const char   *id)
{
	char *vcard = NULL;
	EContact *contact = NULL;
	EBookBackendBrutus *brutus_backend;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;

	d("");
	brutus_backend = E_BOOK_BACKEND_BRUTUS(backend);
	priv = brutus_backend->priv;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	switch (priv->mode) {
	case GNOME_Evolution_Addressbook_MODE_LOCAL :
		contact = e_book_backend_db_cache_get_contact(priv->cache, id);
		vcard = e_vcard_to_string(E_VCARD(contact), EVC_FORMAT_VCARD_30);
		if (contact) {
			e_data_book_respond_get_contact(book, opid, GNOME_Evolution_Addressbook_Success, vcard);
			g_free(vcard);
			g_object_unref(contact);
		} else
			e_data_book_respond_get_contact(book, opid, GNOME_Evolution_Addressbook_ContactNotFound, "");
		return;
	case GNOME_Evolution_Addressbook_MODE_ANY :
	case GNOME_Evolution_Addressbook_MODE_REMOTE :
		if (!brutus_is_likely_connected(brutus_base)) {
			struct BrutusConnectData *data = g_new0(struct BrutusConnectData, 1);

			data->brutus_backend = brutus_backend;
			data->source = e_book_backend_get_source(E_BOOK_BACKEND(brutus_backend));
			g_object_ref(G_OBJECT(data->source));
			g_timeout_add(1, (GSourceFunc)brutus_connect, (gpointer)data);

			e_data_book_respond_get_contact(book, opid, GNOME_Evolution_Addressbook_OfflineUnavailable, "");
			return;
		}

		contact = brutus_get_contact_from_server(brutus_backend, id);
		if (!contact)
			e_data_book_respond_get_contact(book, opid, GNOME_Evolution_Addressbook_ContactNotFound, "");

		vcard = e_vcard_to_string(E_VCARD(contact), EVC_FORMAT_VCARD_30);
		e_data_book_respond_get_contact(book, opid, GNOME_Evolution_Addressbook_Success, vcard);
		g_free(vcard);
		g_object_unref(contact);

		return;
	default :
		d("TODO: *** Investigate unknown EBook mode ***");
		e_data_book_respond_get_contact(book, opid, GNOME_Evolution_Addressbook_OtherError, "");
		break;
	}
}

static void
brutus_book_view_notify_update(EBookBackendBrutus *brutus_backend,
			       GPtrArray *ids,
			       EDataBookView *book_view,
			       gboolean remote_mode,
			       BrutusBackendSearchClosure *closure)
{
	guint i;
	const char *uid = NULL;
	EContact *contact = NULL;

	for (i = 0; i < ids->len; i ++) {
		if (!e_flag_is_set(closure->running))
			break;

		uid = g_ptr_array_index(ids, i);
		if (!uid)
			continue;

		contact = e_book_backend_db_cache_get_contact(brutus_backend->priv->cache, uid);
		if (!contact && remote_mode)
			contact = brutus_get_contact_from_server(brutus_backend, uid);
		if (!contact)
			continue;

		e_data_book_view_notify_update(book_view, contact);
		g_object_unref(contact);
	}
	if (e_flag_is_set(closure->running))
		e_data_book_view_notify_complete(book_view, GNOME_Evolution_Addressbook_Success);
}

static void
e_book_backend_brutus_get_contact_list(EBookBackend *backend,
				       EDataBook *book,
				       guint32 opid,
				       const char *query)
{
	guint i;
	char *uid = NULL;
	GList *contact_list = NULL;
	GList *vcard_list = NULL;
	GList *tmp_list = NULL;
	GPtrArray *ids = NULL;
	EContact *contact = NULL;
	EBookBackendBrutus *brutus_backend;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	struct EBookBackendBrutusSExpData *filter = NULL;

	d("");
	brutus_backend = E_BOOK_BACKEND_BRUTUS(backend);
	priv = brutus_backend->priv;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	switch (priv->mode) {
	case GNOME_Evolution_Addressbook_MODE_LOCAL :
		if (!priv->cache) {
			e_data_book_respond_get_contact_list(book, opid, GNOME_Evolution_Addressbook_RepositoryOffline, NULL);
			return;
		}

		if (priv->summary_ready && e_book_backend_summary_is_summary_query(priv->summary, query)) {
			int i;

			ids = e_book_backend_summary_search(priv->summary, query);
			for (i = 0; i < ids->len; i ++) {
				uid = g_ptr_array_index(ids, i);
				contact = e_book_backend_db_cache_get_contact(priv->cache, uid);
				if (!contact)
					continue;
				vcard_list = g_list_append(vcard_list, e_vcard_to_string(E_VCARD(contact), EVC_FORMAT_VCARD_30));
				g_object_unref(contact);

			}
			g_ptr_array_free(ids, TRUE);
			e_data_book_respond_get_contact_list(book, opid, GNOME_Evolution_Addressbook_Success, vcard_list);

			return;
		}

		contact_list = e_book_backend_db_cache_get_contacts(priv->cache, query);
		tmp_list = contact_list;
		for (; contact_list; contact_list = g_list_next(contact_list)) {
			vcard_list = g_list_append(vcard_list, e_vcard_to_string(E_VCARD(contact_list->data), EVC_FORMAT_VCARD_30));
			g_object_unref(contact_list->data);
		}
		e_data_book_respond_get_contact_list(book, opid, GNOME_Evolution_Addressbook_Success, vcard_list);
		if (tmp_list)
			g_list_free(tmp_list);
		return;
	case GNOME_Evolution_Addressbook_MODE_ANY :
	case GNOME_Evolution_Addressbook_MODE_REMOTE :
		if (!brutus_is_likely_connected(brutus_base)) {
			struct BrutusConnectData *data = g_new0(struct BrutusConnectData, 1);

			data->brutus_backend = brutus_backend;
			data->source = e_book_backend_get_source(E_BOOK_BACKEND(brutus_backend));
			g_object_ref(G_OBJECT(data->source));
			g_timeout_add(1, (GSourceFunc)brutus_connect, (gpointer)data);

			e_data_book_respond_get_contact_list(book, opid, GNOME_Evolution_Addressbook_OfflineUnavailable, NULL);
			return;
		}

		filter = e_book_backend_brutus_build_filter(brutus_backend, query);
		if (!filter) {
			e_data_book_respond_get_contact_list(book, opid, GNOME_Evolution_Addressbook_InvalidQuery, NULL);
			break;
		}
		if (!filter->matching_ids || !filter->matching_ids->len) {
			e_data_book_respond_get_contact_list(book, opid, GNOME_Evolution_Addressbook_ContactNotFound, NULL);
			break;
		}

		for (i = 0; i < filter->matching_ids->len; i++) {
			uid = g_ptr_array_index(filter->matching_ids, i);
			if (!uid)
				continue;

			contact = brutus_get_contact_from_server(brutus_backend, uid);
			if (!contact)
				continue;
			vcard_list = g_list_append(vcard_list, e_vcard_to_string(E_VCARD(contact), EVC_FORMAT_VCARD_30));
			g_object_unref(contact);
		}
		e_data_book_respond_get_contact_list(book, opid, GNOME_Evolution_Addressbook_Success, vcard_list);
		break;
	default :
		d("TODO: *** Investigate unknown EBook mode ***");
		e_data_book_respond_get_contact_list(book, opid, GNOME_Evolution_Addressbook_OtherError, NULL);
		break;
	}

	if (filter)
		brutus_filter_free(filter);
}

static gpointer
brutus_book_view_thread_func(gpointer data)
{
	EBookBackendBrutus *brutus_backend = NULL;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	const char *query = NULL;
	GPtrArray *ids = NULL;
	EDataBookView *book_view = NULL;
	BrutusBackendSearchClosure *closure = NULL;
	struct EBookBackendBrutusSExpData *filter = NULL;

	d("");

	if (!data)
		return NULL;

	book_view = (EDataBookView*)data;
	if (!book_view)
		return NULL;
	bonobo_object_ref(book_view);

	closure = brutus_get_closure(book_view);
	if (!closure)
		goto out;
	e_flag_set(closure->running);

	brutus_backend = closure->brutus_backend;
	priv = brutus_backend->priv;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	d(priv->book_name);

	query = e_data_book_view_get_card_query(book_view);
	dfmt("book_view query %s", query);

	switch (priv->mode) {
	case GNOME_Evolution_Addressbook_MODE_LOCAL :
		if (!brutus_backend->priv->cache) {
			e_data_book_view_notify_complete(book_view, GNOME_Evolution_Addressbook_Success);
			break;
		}

		// check if the query can be satisfied searching only the summary
		if (priv->summary_ready
		    && e_book_backend_summary_is_summary_query(priv->summary, query)) {
			ids = e_book_backend_summary_search(priv->summary, query);
			if (ids && ids->len) {
				brutus_book_view_notify_update(brutus_backend, ids, book_view, FALSE, closure); // put them into the cache
				g_ptr_array_free(ids, TRUE);
			}
			break;
		}

		// nope - we must search the cache
		ids = e_book_backend_db_cache_search(priv->cache, query);
		if (ids && ids->len) {
			brutus_book_view_notify_update(brutus_backend, ids, book_view, FALSE, closure);
			g_ptr_array_free(ids, TRUE);
		} else {
			e_data_book_view_notify_complete(book_view, GNOME_Evolution_Addressbook_ContactNotFound);
			break;
		}
		break;
	case GNOME_Evolution_Addressbook_MODE_ANY :
	case GNOME_Evolution_Addressbook_MODE_REMOTE :
		if (!brutus_is_likely_connected(brutus_base)) {
			struct BrutusConnectData *data = g_new0(struct BrutusConnectData, 1);

			data->brutus_backend = brutus_backend;
			data->source = e_book_backend_get_source(E_BOOK_BACKEND(brutus_backend));
			g_object_ref(G_OBJECT(data->source));
			g_timeout_add(1, (GSourceFunc)brutus_connect, (gpointer)data);

			e_data_book_view_notify_complete(book_view,
							 GNOME_Evolution_Addressbook_OfflineUnavailable);
			break;
		}

		e_data_book_view_notify_status_message(book_view, _("Searching..."));
		filter = e_book_backend_brutus_build_filter(brutus_backend, query);
		if (!filter) {
			e_data_book_view_notify_complete(book_view, GNOME_Evolution_Addressbook_ContactNotFound);
			break;
		}
		if (filter->matching_ids && filter->matching_ids->len)
			brutus_book_view_notify_update(brutus_backend, filter->matching_ids, book_view, TRUE, closure);
		break;
	default :
		d("TODO: *** Investigate unknown EBook mode ***");
		e_data_book_view_notify_complete(book_view, GNOME_Evolution_Addressbook_OtherError);
		break;
	}

out:
	if (filter)
		brutus_filter_free(filter);

	bonobo_object_unref(book_view);
	e_flag_clear(closure->running);

	return NULL;
}

static void
e_book_backend_brutus_start_book_view(EBookBackend  *backend,
				      EDataBookView *book_view)
{
	BrutusBackendSearchClosure *closure = brutus_init_closure(book_view,
								  E_BOOK_BACKEND_BRUTUS(backend));
	d("");
	closure->thread = g_thread_create(brutus_book_view_thread_func,
					  (gpointer)book_view,
					  FALSE,
					  NULL);
	e_flag_wait(closure->running);
	/* (Comment from Groupwise backend): at this point we know the book view thread is actually running */
}

static void
e_book_backend_brutus_stop_book_view(EBookBackend *backend,
				     EDataBookView *book_view)
{
	BrutusBackendSearchClosure *closure = brutus_get_closure(book_view);

	d("");
	e_flag_clear(closure->running);
}

static void
e_book_backend_brutus_get_changes(EBookBackend *backend,
				  EDataBook *book,
				  guint32 opid,
				  const char *change_id)
{
	d("TODO: *** And exactly what should we do here? ***");
}

static EDataBookView*
brutus_find_book_view(EBookBackendBrutus *brutus_backend)
{
	EList *views = e_book_backend_get_book_views(E_BOOK_BACKEND(brutus_backend));
	EIterator *iter;
	EDataBookView *rv = NULL;

	d("");
	if (!views)
		return NULL;

	iter = e_list_get_iterator(views);
	if (!iter) {
		g_object_unref(views);
		return NULL;
	}

	if (e_iterator_is_valid(iter)) {
		/* just always use the first book view */
		EDataBookView *v = (EDataBookView*)e_iterator_get(iter);
		if (v)
			rv = v;
	}

	g_object_unref(iter);
	g_object_unref(views);

	return rv;
}

static void
e_book_backend_brutus_authenticate_user(EBookBackend *backend,
					EDataBook *book,
					guint32 opid,
					const char *user,
					const char *passwd,
					const char *auth_method)
{
	EBookBackendBrutus *brutus_backend = NULL;
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	struct BrutusConnectData *data = NULL;

	d("");
	brutus_backend = E_BOOK_BACKEND_BRUTUS(backend);
	priv = brutus_backend->priv;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	switch (brutus_backend->priv->mode) {
	case GNOME_Evolution_Addressbook_MODE_LOCAL:
		e_book_backend_notify_connection_status(backend, FALSE);
		e_data_book_respond_authenticate_user(book, opid, GNOME_Evolution_Addressbook_RepositoryOffline);
		return;
	case GNOME_Evolution_Addressbook_MODE_ANY:
	case GNOME_Evolution_Addressbook_MODE_REMOTE:
		if (brutus_is_likely_connected(brutus_base)) {
			e_data_book_respond_authenticate_user(book, opid, GNOME_Evolution_Addressbook_Success);
			return;
		}

		data = g_new0(struct BrutusConnectData, 1);
		if (!data) {
			e_data_book_respond_authenticate_user(book, opid, GNOME_Evolution_Addressbook_OtherError);
			break;
		}
		data->brutus_backend = brutus_backend;
		data->source = e_book_backend_get_source(E_BOOK_BACKEND(brutus_backend));
		g_object_ref(G_OBJECT(data->source));
		brutus_connect((gpointer)data);

		if (brutus_is_likely_connected(brutus_base))
			e_data_book_respond_authenticate_user(book, opid, GNOME_Evolution_Addressbook_Success);
		else
			e_data_book_respond_authenticate_user(book, opid, GNOME_Evolution_Addressbook_AuthenticationFailed);
		break;
	default :
		d("TODO: *** Investigate unknown EBook mode ***");
		e_data_book_respond_authenticate_user(book, opid, GNOME_Evolution_Addressbook_OtherError);
		break;
	}

}

static void
e_book_backend_brutus_get_required_fields(EBookBackend *backend,
					  EDataBook *book,
					  guint32 opid)
{
	GList *fields = NULL;

	d("");
	fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_FILE_AS));
	e_data_book_respond_get_supported_fields(book, opid, GNOME_Evolution_Addressbook_Success, fields);
	g_list_free(fields);
}

static void
e_book_backend_brutus_get_supported_fields(EBookBackend *backend,
					   EDataBook *book,
					   guint32 opid)
{
	EBookBackendBrutus *brutus_backend = NULL;
	EBookBackendBrutusPrivate *priv = NULL;
	GList *fields = NULL;

	d("");
	if (!backend) {
		e_data_book_respond_get_supported_fields(book, opid, GNOME_Evolution_Addressbook_OtherError, NULL);
		return;
	}
	brutus_backend = E_BOOK_BACKEND_BRUTUS(backend);
	priv = brutus_backend->priv;

	switch (priv->kind) {
	case ADDRESS_BOOK_GAL :
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_FILE_AS));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_FULL_NAME));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_TITLE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_EMAIL_1));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_BUSINESS));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_PAGER));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_HOME));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_MOBILE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_BUSINESS_FAX));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ORG));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ORG_UNIT));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_OFFICE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ROLE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_NOTE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ADDRESS_HOME));
		break;
	case ADDRESS_BOOK_PAB :
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_BIRTH_DATE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ANNIVERSARY));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_FULL_NAME));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_TITLE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_NICKNAME));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ADDRESS_HOME));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ADDRESS_WORK));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ADDRESS_OTHER));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_ASSISTANT));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_BUSINESS));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_BUSINESS_2));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_BUSINESS_FAX));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_CALLBACK));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_CAR));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_COMPANY));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_HOME));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_HOME_2));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_HOME_FAX));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_ISDN));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_MOBILE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_OTHER));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_PAGER));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_PRIMARY));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_RADIO));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_TELEX));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHONE_TTYTDD));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ORG));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ORG_UNIT));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_OFFICE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_TITLE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_MANAGER));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ASSISTANT));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_HOMEPAGE_URL));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_SPOUSE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_NOTE));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_ADDRESS_WORK));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_FILE_AS));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_EMAIL_1));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_EMAIL_2));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_EMAIL_3));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_FREEBUSY_URL));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_CATEGORY_LIST));
		fields = g_list_append(fields, (gpointer)e_contact_field_name(E_CONTACT_PHOTO));
		break;
	default :
		g_assert_not_reached();
	}

	if (fields) {
		e_data_book_respond_get_supported_fields(book, opid, GNOME_Evolution_Addressbook_Success, fields);
		g_list_free(fields);
	} else
		e_data_book_respond_get_supported_fields(book, opid, GNOME_Evolution_Addressbook_OtherError, NULL);
}

static GNOME_Evolution_Addressbook_CallStatus
e_book_backend_brutus_cancel_operation(EBookBackend *backend,
				       EDataBook *book)
{
	d("");
	return GNOME_Evolution_Addressbook_CouldNotCancel;
}

static void
#if ((4 == DB_VERSION_MAJOR) && (3 <= DB_VERSION_MINOR))
file_errcall(const DB_ENV *env, const char *buf1, const char *buf2)
#else
	file_errcall (const char *buf1, char *buf2)
#endif
{
	dfmt("libdb error: %s", buf2);
}

static GNOME_Evolution_Addressbook_CallStatus
e_book_backend_brutus_load_source(EBookBackend *backend,
				  ESource *source,
				  gboolean only_if_exists)
{
	static guint count = 0;
	struct BrutusConnectData *connect_data = NULL;
	GNOME_Evolution_Addressbook_CallStatus retv = GNOME_Evolution_Addressbook_OtherError;
	EBookBackendBrutus *brutus_backend = E_BOOK_BACKEND_BRUTUS(backend);
	EBookBackendBrutusPrivate *priv = brutus_backend->priv;
	BrutusBaseClass *brutus_base = BRUTUS_BASE_CLASS(priv);
	gboolean connected = FALSE;
	char *dirname = NULL;
	char *filename = NULL;
	int db_error = 0;
	DB *db = NULL;
	DB_ENV *env = NULL;
	const char *offline = NULL;

	d("");
	g_object_ref(G_OBJECT(source));

	if (count++) {
		dfmt("load_source() invoked %d times", count);
		goto out;
	}

	offline = e_source_get_property(source, "offline_sync");
	if (offline  && g_str_equal(offline, "1"))
		priv->make_offline_cache = TRUE;
	else
		priv->make_offline_cache = FALSE;

	priv->uri =  e_source_get_uri(source);
	if (!priv->uri)
		goto out;

	if (!priv->constructed)
		priv->constructed = brutus_backend_construct(brutus_backend, source);
	if (priv->constructed) {
		connect_data = g_new0(struct BrutusConnectData, 1);
		connect_data->brutus_backend = brutus_backend;
		connect_data->source = source;
		g_object_ref(G_OBJECT(connect_data->source));
		brutus_connect((gpointer)connect_data); //
		connected = brutus_is_likely_connected(brutus_base);
	} else {
		d("could not construct the backend");
		goto out;
	}
	if (connected)
		d("the backend is likely connected");
	else {
		d("the backend might not be connected");
		goto out;
	}

	e_book_backend_set_is_loaded(E_BOOK_BACKEND (backend), priv->constructed);
	e_book_backend_set_is_writable(E_BOOK_BACKEND(backend), TRUE);
	if (GNOME_Evolution_Addressbook_MODE_LOCAL == priv->mode ) {
		e_book_backend_notify_writable(backend, FALSE);
		e_book_backend_notify_connection_status(backend, FALSE);
	} else
		e_book_backend_notify_connection_status(backend, connected);

	if (GNOME_Evolution_Addressbook_MODE_LOCAL == priv->mode) {
		if (!e_book_backend_db_cache_exists(priv->uri)) {
			retv = GNOME_Evolution_Addressbook_OfflineUnavailable;
			goto out;
		}
	}

	/*
	 * DB
	 */

	// timestamp and summary file name
	dirname = g_build_filename(brutus_base->cache_dir, priv->book_name, "/", NULL);
	filename = g_build_filename(dirname, "brutus-address-book.db", NULL);
	priv->summary_file_name = g_strdup_printf("%s%s", filename, ".summary");
	priv->timestamp_file_name = g_strdup_printf("%s%s", filename, ".timestamp");

	// load summary from file
	g_mkdir_with_parents(g_path_get_dirname(priv->summary_file_name), 0700);
	priv->summary = e_book_backend_summary_new(priv->summary_file_name, SUMMARY_FLUSH_TIMEOUT);
	e_book_backend_summary_load(priv->summary);

	db_error = e_db3_utils_maybe_recover(filename);
	if (db_error) {
		dfmt("db recovery failed with %d", db_error);
		goto out;
	}

	/*
	 * Create DB env
	 */

	db_error = db_env_create(&env, 0);
	if (db_error) {
		dfmt("db_env_create failed with %d", db_error);
		goto out;
	}

	db_error = env->open(env, NULL, DB_CREATE | DB_INIT_MPOOL | DB_PRIVATE | DB_THREAD, 0);
	if (db_error) {
		env->close(env, 0);
		dfmt("db_env_open failed with %d", db_error);
		goto out;
	}

	env->set_errcall(env, file_errcall);
	priv->cache_env = env;
	env = NULL;

	/*
	 * Create DB
	 */

	db_error = db_create(&db, env, 0);
	if (db_error) {
		dfmt("db_create failed with %d", db_error);
		goto out;
	}

	db_error = db->open(db, NULL, filename, NULL, DB_HASH, DB_THREAD, 0600);
	if (db_error == DB_OLD_VERSION) {
		db_error = e_db3_utils_upgrade_format(filename);
		if (db_error) {
			dfmt("db format upgrade failed with %d", db_error);
			goto out;
		}

		db_error = db->open (db, NULL, filename, NULL, DB_HASH, DB_THREAD, 0600);
	} else if (db_error) {
		int rv;

		dfmt("db_open failed with %d", db_error);

		rv = g_mkdir_with_parents(dirname, 0700);
		if ((-1 == rv) && (EEXIST != errno)) {
			dfmt2("failed to make directory %s: %s", dirname, strerror(errno));
			if ((EACCES == errno) || (EPERM == errno))
				retv = GNOME_Evolution_Addressbook_PermissionDenied;
			goto out;
		}

		db_error = db->open(db, NULL, filename, NULL, DB_HASH, DB_CREATE | DB_THREAD, 0600);
		if (db_error) {
			dfmt("db->open (...DB_CREATE...) failed with %d", db_error);
			goto out;
		} else
			d("database created and opened successfully");
	} else
		d("database opened successfully");

	priv->cache = db;
	e_book_backend_db_cache_set_filename(priv->cache, filename);

	// start polling
	brutus_poll_server(brutus_backend); // the first poll runs immediately
	priv->timeout_id = g_timeout_add(1000*60*priv->auto_check_timeout, // once an hour
					 (GSourceFunc)brutus_poll_server,
					 (gpointer)brutus_backend);

	retv = GNOME_Evolution_Addressbook_Success;

out:
	if (db_error || !priv->cache) {
		if (priv->cache)
			priv->cache->close(priv->cache, 0);
		priv->cache = NULL;
	}

	if (db_error || !priv->cache_env) {
		if (priv->cache_env)
			priv->cache_env->close(priv->cache_env, 0);
		priv->cache_env = NULL;
	}

	g_object_unref(G_OBJECT(source));

	if (dirname)
		g_free(dirname);

	if (filename)
		g_free(filename);

	return retv;
}

static void
e_book_backend_brutus_remove(EBookBackend *backend,
			     EDataBook *book,
			     guint32 opid)
{
	d("");
	e_data_book_respond_remove(book,  opid,  GNOME_Evolution_Addressbook_PermissionDenied);
}

static char *
e_book_backend_brutus_get_static_capabilities(EBookBackend *backend)
{
	d("");
/*	return g_strdup ("net,bulk-removes,do-initial-query,contact-lists"); */
	return g_strdup ("net,bulk-removes,do-initial-query");
}

static void
e_book_backend_brutus_get_supported_auth_methods(EBookBackend *backend, EDataBook *book, guint32 opid)
{
	GList *auth_method_list = NULL;
	char *auth_method;

	d("");
	auth_method =  g_strdup_printf("plain/password");
	auth_method_list = g_list_append(auth_method_list, auth_method);
	e_data_book_respond_get_supported_auth_methods(book, opid, GNOME_Evolution_Addressbook_Success, auth_method_list);

	g_free(auth_method);
	g_list_free(auth_method_list);
}

static void
e_book_backend_brutus_set_mode(EBookBackend *backend, int mode)
{
	EBookBackendBrutus *brutus_backend;
	EBookBackendBrutusPrivate *priv;
	BrutusBaseClass *brutus_base;

	d("");
	brutus_backend = E_BOOK_BACKEND_BRUTUS(backend);
	priv = brutus_backend->priv;
	if (!priv)
		return;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	priv->mode = mode;
	if (!e_book_backend_is_loaded(backend))
		return;

	if (mode == GNOME_Evolution_Addressbook_MODE_LOCAL) {
		e_book_backend_notify_writable(backend, FALSE);
		e_book_backend_notify_connection_status (backend, FALSE);
	} else {
		if (brutus_backend->priv->is_writable)
			e_book_backend_notify_writable(backend, TRUE);
		else
			e_book_backend_notify_writable(backend, FALSE); 

		if (brutus_is_likely_connected(brutus_base)) 
			e_book_backend_notify_connection_status(backend, TRUE); 
		else
			e_book_backend_notify_connection_status(backend, FALSE); 
	}										
}

EBookBackend *
e_book_backend_brutus_new(void)
{
	EBookBackendBrutus *brutus_backend = g_object_new(E_TYPE_BOOK_BACKEND_BRUTUS, NULL);

	d("");
	return E_BOOK_BACKEND(brutus_backend);
}

static void
e_book_backend_brutus_dispose (GObject *object)
{
	EBookBackendBrutus *brutus_backend;

	d("");
	brutus_backend = E_BOOK_BACKEND_BRUTUS(object);

	if (brutus_backend->priv) {
		g_free(brutus_backend->priv);
		brutus_backend->priv = NULL;
	}

	G_OBJECT_CLASS(e_book_backend_brutus_parent_class)->dispose(object);
}

/* Finalize handler for the backend */
static void
e_book_backend_brutus_finalize(GObject *object)
{
	EBookBackendBrutus *brutus_backend;
	EBookBackendBrutusPrivate *priv;
	BrutusBaseClass *brutus_base;

	d("");
	g_return_if_fail(object != NULL);
	g_return_if_fail(E_IS_BOOK_BACKEND_BRUTUS(object));

	brutus_backend = E_BOOK_BACKEND_BRUTUS(object);
	priv = brutus_backend->priv;

	/* Clean up */
	if (priv) {
		brutus_base = BRUTUS_BASE_CLASS(priv);

		if (priv->poll_mutex) {
			g_mutex_lock(priv->poll_mutex); // to block brutus_poll_server()
			if (priv->timeout_id) {
				g_source_remove(priv->timeout_id);
				priv->timeout_id = 0;
			}
			g_mutex_unlock(priv->poll_mutex);
			g_mutex_free(priv->poll_mutex);
		}

		if (priv->timestamp_mutex)
			g_mutex_free(priv->timestamp_mutex);

		if (brutus_is_connected(brutus_base->brutus_server, brutus_base->mapi_session))
			brutus_mapi_shutdown(brutus_backend);

		brutus_base_class_free(brutus_base);

		if (priv->cache)
			priv->cache->close(priv->cache, 0);

		if (priv->cache_env)
			priv->cache_env->close(priv->cache_env, 0);

		if (priv->uri)
			g_free(priv->uri);

		if (priv->book_name)
			g_free(priv->book_name);

		if (priv->timestamp_file_name)
			g_free(priv->timestamp_file_name);

		if (priv->summary_file_name)
			g_free(priv->summary_file_name);

		if (priv->summary) {
			e_book_backend_summary_save(priv->summary);
			g_object_unref(priv->summary);
		}

		if (priv->search_def)
			brutus_free_search_mapping(priv->search_def);

		g_free(priv);

		brutus_backend->priv = NULL;
	}

	if (G_OBJECT_CLASS(e_book_backend_brutus_parent_class)->finalize)
		(* G_OBJECT_CLASS(e_book_backend_brutus_parent_class)->finalize)(object);
}

static void
e_book_backend_brutus_class_init (EBookBackendBrutusClass *klass)
{
	GObjectClass  *object_class = G_OBJECT_CLASS (klass);
	EBookBackendClass *parent_class;

	d("");
	e_book_backend_brutus_parent_class = g_type_class_peek_parent(klass);

	parent_class = E_BOOK_BACKEND_CLASS(klass);

	/* Set the virtual methods. */
	parent_class->load_source                = e_book_backend_brutus_load_source;
	parent_class->get_static_capabilities    = e_book_backend_brutus_get_static_capabilities;
	parent_class->create_contact             = e_book_backend_brutus_create_contact;
	parent_class->remove_contacts            = e_book_backend_brutus_remove_contacts;
	parent_class->modify_contact             = e_book_backend_brutus_modify_contact;
	parent_class->get_contact                = e_book_backend_brutus_get_contact;
	parent_class->get_contact_list           = e_book_backend_brutus_get_contact_list;
	parent_class->start_book_view            = e_book_backend_brutus_start_book_view;
	parent_class->stop_book_view             = e_book_backend_brutus_stop_book_view;
	parent_class->get_changes                = e_book_backend_brutus_get_changes;
	parent_class->authenticate_user          = e_book_backend_brutus_authenticate_user;
	parent_class->get_required_fields        = e_book_backend_brutus_get_required_fields;
	parent_class->get_supported_fields       = e_book_backend_brutus_get_supported_fields;
	parent_class->get_supported_auth_methods = e_book_backend_brutus_get_supported_auth_methods;
	parent_class->cancel_operation           = e_book_backend_brutus_cancel_operation;
	parent_class->remove                     = e_book_backend_brutus_remove;
	parent_class->set_mode                   = e_book_backend_brutus_set_mode;
	object_class->dispose                    = e_book_backend_brutus_dispose;
	object_class->finalize                   = e_book_backend_brutus_finalize;
}

static void
e_book_backend_brutus_init(EBookBackendBrutus *brutus_backend)
{
	EBookBackendBrutusPrivate *priv = NULL;
	BrutusBaseClass *brutus_base = NULL;
	EDataBookView *book_view = NULL;

	d("");
	priv = g_new0(EBookBackendBrutusPrivate, 1);
	brutus_backend->priv = priv;
	if (!priv)
		goto out_no_mem;
	brutus_base = BRUTUS_BASE_CLASS(priv);

	brutus_base_class_init(brutus_base);
	if (!brutus_base->connect_mutex)
		goto out_no_mem;

	/* fixed value - the GAL/PAB is rarely updated */
	priv->auto_check_timeout = AUTO_CHECK_TIMEOUT_DEFAULT;
	priv->timeout_id = 0;

	/* create the mutex to controll poll blocking */
	priv->poll_mutex = g_mutex_new();
	if (!priv->poll_mutex)
		goto out_no_mem;

	priv->timestamp_mutex = g_mutex_new();
	if (!priv->timestamp_mutex)
		goto out_no_mem;

	priv->constructed = FALSE;

	priv->make_offline_cache = FALSE;

	priv->is_writable = TRUE; // safe default - we CAN NOT know whether it is actually writable unless we attempts a write

	priv->mode = GNOME_Evolution_Addressbook_MODE_LOCAL; // in lack of INVALID...

	priv->msg_store = CORBA_OBJECT_NIL;
	priv->PAB_book = CORBA_OBJECT_NIL; // IMAPIFolder::IMAPIContainer

	priv->addrbook = CORBA_OBJECT_NIL;
	priv->GAL_book = CORBA_OBJECT_NIL; // IABContainer::IMAPIContainer

	priv->cache = NULL;
	priv->cache_env = NULL;

	priv->cache_ready = FALSE;
	priv->summary_ready = FALSE;

	priv->uri = NULL;
	priv->book_name = NULL;
	priv->timestamp_file_name = NULL;
	priv->summary_file_name = NULL;

	priv->summary = NULL;

	priv->search_def = NULL;
	priv->content_tags = NULL;

	return;

out_no_mem:
	if (priv) {
		if (brutus_base && brutus_base->connect_mutex) {
			g_mutex_free(brutus_base->connect_mutex);
			brutus_base->connect_mutex = NULL;
		}
		if (priv->poll_mutex) {
			g_mutex_free(priv->poll_mutex);
			priv->poll_mutex = NULL;
		}
		if (priv->timestamp_mutex) {
			g_mutex_free(priv->timestamp_mutex);
			priv->timestamp_mutex = NULL;
		}
		g_free(priv);
		brutus_backend->priv = NULL;
	}

	book_view = brutus_find_book_view(brutus_backend);
	if (book_view) { // racy (in principle)
		bonobo_object_ref(book_view);
		e_data_book_view_notify_complete(book_view, GNOME_Evolution_Addressbook_OtherError);
		bonobo_object_unref(book_view);
	}

	return;
}

GType
e_book_backend_brutus_get_type(void)
{
	static GType type = 0;

	if (!type) {
		GTypeInfo info = {
			.class_size     = sizeof(EBookBackendBrutusClass),
			.base_init      = NULL,
			.base_finalize  = NULL,
			.class_init     = (GClassInitFunc)e_book_backend_brutus_class_init,
			.class_finalize = NULL,
			.class_data     = NULL,
			.instance_size  = sizeof(EBookBackendBrutus),
			.n_preallocs    = 0,
			.instance_init  = (GInstanceInitFunc)e_book_backend_brutus_init
		};

		type = g_type_register_static(E_TYPE_BOOK_BACKEND, "EBookBackendBrutus", &info, 0);
	}

	return type;
}
